# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import logging
import os
import subprocess
import time
from xen.xend import uuid
from agent.base import Agent
from agent.config import get_config
from agent.utils.xmlrpclib import ServerProxy


class Common(Agent):
    """
    Common functions.
    """

    def get_services(self, version=None):
        return {
            'echo': echo,
            'sleep': sleep,
            'dispatch': dispatch,
            'get_api_version': get_api_version}


def echo(msg):
    return msg


def sleep(secs):
    time.sleep(float(secs))


def dispatch(uri, func, *args):
    try:
        logging.debug("dispatch function %s to server %s" % (func, uri))
        sp = ServerProxy(uri)
        return getattr(sp, func)(*args)
    except Exception, e:
        logging.error("dispatch error: %s: %s" % (func, e))
        raise


def get_api_version():
    return [1, 2]


def run_cmd(cmd, success_return_code=(0,)):
    if not isinstance(cmd, list):
        raise Exception("Only accepts list!")
    cmd = [str(x) for x in cmd]
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE, close_fds=True)
    (stdoutdata, stderrdata) = proc.communicate()
    if proc.returncode not in success_return_code:
        raise RuntimeError('Command: %s failed (%s): stderr: %s stdout: %s'
                           % (cmd, proc.returncode, stderrdata, stdoutdata))
    return stdoutdata


def old_run_cmd(cmd, success_return_code=(0,), background=False):
    """Warning: incorporate unsanitized input from an untrusted source makes a
    program vulnerable to shell injection, a serious security flaw which can
    result in arbitrary command execution.
    """
    if not isinstance(cmd, basestring):
        raise Exception("Only accepts str!")
    if background:
        devnull = open(os.devnull, "rw")
        subprocess.Popen(cmd, stdin=devnull, stdout=devnull,
                         stderr=devnull, shell=True, close_fds=True)
    else:
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                   stderr=subprocess.PIPE, shell=True, close_fds=True)
        (stdoutdata, stderrdata) = proc.communicate()
        if proc.returncode not in success_return_code:
            raise RuntimeError('Command: %s failed (%s): stderr: %s stdout: %s'
                    % (cmd, proc.returncode, stderrdata, stdoutdata))
        return stdoutdata


def retry_cmd(cmd, interval=5, count=3):
    # retry if failed
    for i in range(1, count):
        try:
            output = run_cmd(cmd)
            return output
        except:
            time.sleep(interval)
    # retry for the last time
    return run_cmd(cmd)


def get_uuid():
    uuid.uuidFactory = uuid.getUuidUuidgen
    return uuid.createString()


def host_alive(ip):
    status = os.popen("ping -q -c 2 %s" % ip).close()
    return not status


def get_agt_url(host):
    config = get_config()
    if config.get("security", "ssl").lower().startswith("enable"):
        protocol = "https"
    else:
        protocol = "http"
    return "%s://%s:%s" % (protocol, host, config.get("agent", "port"))


def guess_mimetype(filename):
    ftype = run_cmd(['file', '-b', '-i', filename])
    ftype = ftype.split(',')[0] # application/x-tar, POSIX
    ftype = ftype.split(';')[0] # application/x-tar; charset=binary
    return ftype.strip()


def set_nice(increment=10, io_cls=2, io_data=7):
    os.nice(increment)
    run_cmd(['ionice', '-c', io_cls, '-n', io_data, '-p', os.getpid()])
