# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import os
import socket
import logging
from multiprocessing import Pool
from time import sleep
from agent.constants import *
from agent.config import get_config
from agent.db import dump_db, read_item, delete_item, get_cluster_db_home
from agent.common import get_agt_url
from agent.utils.xmlrpclib import ServerProxy
from agent.hypervisor.xenxm import is_vm_running, get_vm_path
from agent.serverpool import is_clustered
from agent.remaster import is_master


def f(host):
    try:
        sp = ServerProxy(get_agt_url(host))
        vms = sp.list_vms()
        dom0_uuid = "00000000-0000-0000-0000-000000000000"
        return [ vm["uuid"] for vm in vms if vm["uuid"] != dom0_uuid ]
    except Exception, e:
        logging.error("Failed to get VM list on %s: %s", host, e)


def get_live_vms():
    hosts = [host['registered_ip'] for host in dump_db("server_pool_servers", get_cluster_db_home()).values()]
    pool = Pool(processes=len(hosts))
    results = pool.map(f, hosts)
    live_vms = []
    for e in results:
        if e:
            live_vms.extend(e)
    return live_vms


def get_monitored_vms():
    return dump_db("monitored_vms", get_cluster_db_home()).keys()


def ha_start_vm(repo_id, vm_id):
    logging.debug("Trying to start VM %s", vm_id)
    hosts = [host['registered_ip'] for host in dump_db("server_pool_servers", get_cluster_db_home()).values()]
    for host in hosts:
        try:
            if is_vm_running(repo_id, vm_id):
                logging.debug("VM %s already started", vm_id)
                return
            sp = ServerProxy(get_agt_url(host))
            sp.start_vm(repo_id, vm_id)
            logging.debug("VM %s started on %s", vm_id, host)
            return
        except Exception, e:
            logging.error("Failed to start VM %s on %s: %s", vm_id, host, e)
    logging.error("Cannot start VM %s: all attempts failed", vm_id)


def apply_ha_policies(monitored_vms, live_vms):
    diff = set(monitored_vms) - set(live_vms)
    if diff:
        logging.debug("Monitored VMs: %s", monitored_vms)
        logging.debug("Live VMs: %s", live_vms)
    for uuid in diff:
        cluster_db_home = get_cluster_db_home()
        vm = read_item("monitored_vms", uuid, cluster_db_home)
        if os.path.exists(get_vm_path(vm['repo_id'], vm['vm_id'])):
            ha_start_vm(vm['repo_id'], vm['vm_id'])


def serve_forever():
    logging.info("HA SERVER STARTED")
    config = get_config()
    ha_interval = int(config.get("ha", "interval"))
    ha_timeout = int(config.get("ha", "timeout"))
    # set socket default timeout for all connections
    # which prevents xmlrpc from hanging infinitely
    socket.setdefaulttimeout(ha_timeout)
    while True:
        try:
            sleep(ha_interval)
            if is_clustered() and is_master():
                live_vms = get_live_vms()
                monitored_vms = get_monitored_vms()
                apply_ha_policies(monitored_vms, live_vms)
        except Exception, e:
            logging.exception("Error in HA process: %s", e)
            sleep(SLEEP_ON_ERR)


if __name__ == '__main__':
    serve_forever()
