# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import os
import os.path
import re
import sys
import glob
import random
import string
import logging
import tempfile
import rpm
import yum
import socket
import subprocess
import time
import stat
import errno
import signal
import threading
import platform
import agent.constants
from agent import OVSDiscoverExtension
from agent.OVSUtilityExtension import get_boot_time
from urlgrabber import urlgrab
from urlgrabber.progress import BaseMeter
from iniparse import INIConfig
from iniparse.compat import ConfigParser
from distutils.sysconfig import get_python_lib
from agent.password import update_password
from agent.daemon import reload_xmlrpc_server, reload_notification_server
from agent.utils import filelock, xmlrpclib
from agent.utils.dateBackend import dateBackend
from agent.utils.timezoneBackend import timezoneBackend
from agent.constants import (LOCAL_SERVER_DB, MONITOR_SERVER_INTERVAL,
                             LOCAL_APROC_DB, LOCAL_EXPORTS_DB)
from agent.db import read_item, write_item, delete_item, dump_db
from agent.common import run_cmd, set_nice
from agent.discover import get_server_unique_id, SimpleDiscoveryXML
from agent.serverpool import get_server_pool_uuid, get_owning_manager_uuid, get_membership_state, get_cluster_state
from agent.stats import set_statistic_interval
from agent.notification import notify_manager, async_proc_prog
from agent.host.base import Host


SERVER_RESERVED_MARKER = "##SERVER_RESERVED##\n"
DISCOVERY_MARKER = "##STORAGE_DISCOVERY##\n"
yum_conf = '/etc/yum.conf'
yum_repo_conf = '/etc/yum.repos.d/ovm.repo'
yum_lock = '/var/run/ovs-agent/yum.lock'

XML_INDENT = 2    # Indent amount for XML documents
STORAGE_DISCOVER_PROG = "storage_discover" # External storage discovery utility
SCSI_INFO_DEBUG_LVL = 0 # Default debug level for above util


class Linux(Host):
    """
    Linux host management.
    """

    def get_services(self, version=None):
        services = {
            'get_yum_config': self.get_yum_config,
            'configure_yum': self.configure_yum,
            'delete_yum_repo': self.delete_yum_repo,
            'yum_list_package': self.yum_list_package,
            'yum_update': self.yum_update,
            'list_package': self.list_package,
            'install_package': self.install_package,
            'uninstall_package': self.uninstall_package,
            'get_datetime': self.get_datetime,
            'set_datetime': self.set_datetime,
            'get_timezone': self.get_timezone,
            'set_timezone': self.set_timezone,
            'get_ntp': self.get_ntp,
            'set_ntp': self.set_ntp,
            'enable_ntp': self.enable_ntp,
            'disable_ntp': self.disable_ntp,
            'get_log': self.get_log,
            'get_support_files': self.get_support_files,
            'update_agent_password': self.update_agent_password,
            'update_core_api_bindings': self.update_core_api_bindings,
            'discover_server': self.discover_server,
            'discover_hardware': self.discover_hardware,
            'discover_physical_luns': self.discover_physical_luns,
            'discover_mounted_file_systems': self.discover_mounted_file_systems,
            'copy_file': self.copy_file,
            'ovs_download_file': self.ovs_download_file,
            'ovs_async_proc': self.ovs_async_proc,
            'ovs_async_proc_status': self.ovs_async_proc_status,
            'ovs_async_proc_signal': self.ovs_async_proc_signal,
            'get_last_boot_time': self.get_last_boot_time,
            'notify_manager': notify_manager,
            'set_statistic_interval': set_statistic_interval,
        }
        if version == 2:
            services.update({
                'export_fs': self.export_fs,
                'unexport_fs': self.unexport_fs,
                })
        return services

    def get_yum_config(self):
        lock = filelock.LockFile(yum_conf)
        lock.acquire(wait=5)
        try:
            result = {}
            yb = yum.YumBase()
            result['main'] = {}
            for (option, value) in yb.conf.iteritems():
                result['main'][option] = value
            for repo in sorted(yb.repos.findRepos('*')):
                result[repo.id] = {}
                for (option, value) in repo.iteritems():
                    result[repo.id][option] = value
            return result
        finally:
            lock.release()

    def configure_yum(self, section, params):

        def update_conf(conf):
            for (option, value) in params.items():
                if option not in conf.iterkeys():
                    raise RuntimeError('Unknown option: %s' % option)
                setattr(conf, option, value)

        lock = filelock.LockFile(yum_conf)
        lock.acquire(wait=5)
        try:
            yb = yum.YumBase()
            if section == 'main':
                ybc = yb.conf
                update_conf(ybc)
                self.writeRawConfigFile(yum_conf, 'main', ybc.cfg.options, ybc.iteritems, ybc.optionobj, params.keys())
            else:
                try:
                    repo = yb.repos.getRepo(section)
                except yum.Errors.RepoError:
                    parser = ConfigParser()
                    parser.read(yum_repo_conf)
                    parser.add_section(section)
                    conf = yum.config.RepoConf()
                    update_conf(conf)
                    for (name, value) in conf.iteritems():
                        if name not in params.keys():
                            continue
                        option = conf.optionobj(name)
                        parser.set(section, name,  option.tostring(value))
                    f = open(yum_repo_conf, "w")
                    parser.write(f)
                    f.close()
                    return
                update_conf(repo)
                self.writeRawConfigFile(repo.repofile, repo.id, repo.cfg.options, repo.iteritems, repo.optionobj, params.keys())
        finally:
            lock.release()

    def delete_yum_repo(self, repo_id):
        lock = filelock.LockFile(yum_conf)
        lock.acquire(wait=5)
        try:
            yb = yum.YumBase()
            try:
                repo = yb.repos.getRepo(repo_id)
            except yum.Errors.RepoError:
                logging.debug('No such repository: %s' % repo_id)
                return
            parser = ConfigParser()
            parser.read(repo.repofile)
            parser.remove_section(repo.id)
            f = open(repo.repofile, "w")
            parser.write(f)
            f.close()
        finally:
            lock.release()

    def yum_list_package(self, pkgnarrow='all', patterns=None, showdups=None, ignore_case=False):

        def format_package_info(pkglist):
            return [(pkg.name, pkg.arch, pkg.epoch, pkg.version, pkg.release, pkg.repo.id) for pkg in pkglist]

        lock = filelock.FileLock(yum_lock)
        lock.acquire(wait=120)
        try:
            try:
                run_cmd(['/usr/bin/yum', 'clean', 'all'])
            except Exception, err:
                logging.warn('Cannot remove yum cache data: %s', err)
            yb = yum.YumBase()
            ygh = yb.doPackageLists(pkgnarrow, patterns, showdups, ignore_case)
            result = {}
            for category in ['installed', 'available', 'reinstall_available', 'old_available',
                             'updates', 'obsoletes', 'obsoletesTuples', 'recent', 'extras']:
                result[category] = format_package_info(getattr(ygh, category))
            return result
        finally:
            lock.release()

    def yum_update(self, option=''):
        try:
            run_cmd(['/usr/bin/yum', 'clean', 'all'])
        except Exception, err:
            logging.warn('Cannot remove yum cache data: %s', err)
        cmd = ['/usr/bin/yum', 'update', '-y', '-q']
        if option:
            cmd += [option]
        run_cmd(cmd)

    def list_package(self, name='*'):
        ts = rpm.TransactionSet()
        if re.search('[?*]|\[.+\]', name):
            mi = ts.dbMatch()
            mi.pattern("name", rpm.RPMMIRE_GLOB, name)
        else:
            mi = ts.dbMatch("name", name)
        pkgs = []
        for hdr in mi:
            info = {'name': hdr[rpm.RPMTAG_NAME],
                    'arch': hdr[rpm.RPMTAG_ARCH],
                    'version': hdr[rpm.RPMTAG_VERSION],
                    'release': hdr[rpm.RPMTAG_RELEASE],
                    'epoch': hdr[rpm.RPMTAG_EPOCH]}
            pkgs.append(info)
        return pkgs

    def install_package(self, pkg_data, option=''):
        (fd, fn) = tempfile.mkstemp(suffix='.rpm', prefix="rpm-package-")
        try:
            os.write(fd, pkg_data.data)
            os.close(fd)
            cmd = ['/bin/rpm', '--upgrade']
            if option:
                cmd += [option]
            cmd += [fn]
            run_cmd(cmd)
        finally:
            os.unlink(fn)

    def uninstall_package(self, pkg_list, option=''):
        cmd = ['/bin/rpm', '--erase']
        if option:
            cmd += [option]
        cmd += pkg_list
        run_cmd(cmd)

    def get_datetime(self):
        date_backend = dateBackend()
        return tuple(date_backend.getDate())

    def set_datetime(self, year, month, day, hour, min, sec):
        date_backend = dateBackend()
        date_backend.writeDateConfig((year, month - 1, day), (hour, min, sec))
        date_backend.syncHardwareClock()

    def get_timezone(self):
        timezone_backend = timezoneBackend()
        timezone_info = timezone_backend.getTimezoneInfo()
        if timezone_info[1] == 'false':
            return timezone_info[0], False
        else:
            return timezone_info[0], True

    def set_timezone(self, timezone, utc=False):
        if utc:
            utc = 1
        else:
            utc = 0
        timezone_backend = timezoneBackend()
        timezone_backend.writeConfig(timezone, utc)

    def get_ntp(self):
        date_backend = dateBackend()
        (ntp_servers, local_time_source) = date_backend.getNtpServers()
        is_ntp_running = bool(date_backend.isNtpRunning())
        return (ntp_servers, local_time_source, is_ntp_running)

    def set_ntp(self, servers, local_time_source=False, allow_query=False):
        if len(servers) == 0 and local_time_source == False:
            raise Exception('Please specify an NTP server to use or use local time source.')
        logging.debug('substitute manager src ip %s for %s', agent.constants.MANAGER_SRC_IP, servers)
        servers = [re.sub('0\.0\.0\.0', agent.constants.MANAGER_SRC_IP, s) for s in servers]
        date_backend = dateBackend()
        broadcast_client = date_backend.getNtpBroadcastClient()
        iburst = date_backend.getIburst()
        if date_backend.writeNtpConfig(servers, broadcast_client, local_time_source, iburst) == None:
            raise Exception('Could not write NTP configurations.')
        fd = open('/etc/sysconfig/ntpd', 'r+')
        lines = fd.readlines()
        fd.seek(0)
        fd.truncate(0)
        for line in lines:
            if re.match(r'^OPTIONS=.*', line) and not re.search(r'("| )-x( |")', line):
                line = re.sub('"$', ' -x"', line)
            fd.write(line)
        fd.close()
        fd = open('/etc/ntp.conf', 'w')
        for line in date_backend.ntpFile:
            if allow_query:
                line = re.sub('^restrict default kod nomodify notrap nopeer noquery$',
                              'restrict default kod nomodify notrap nopeer', line)
                line = re.sub('^restrict -6 default kod nomodify notrap nopeer noquery$',
                              'restrict -6 default kod nomodify notrap nopeer', line)
            else:
                line = re.sub('^restrict default kod nomodify notrap nopeer$',
                              'restrict default kod nomodify notrap nopeer noquery', line)
                line = re.sub('^restrict -6 default kod nomodify notrap nopeer$',
                              'restrict -6 default kod nomodify notrap nopeer noquery', line)
            fd.write(line)
        fd.close()

    def enable_ntp(self):
        date_backend = dateBackend()
        date_backend.startNtpService(None)
        date_backend.chkconfigOn()

    def disable_ntp(self):
        date_backend = dateBackend()
        date_backend.stopNtpService()
        date_backend.chkconfigOff()

    def get_log(self, loglist=None):
        logs = {'messages': '/var/log/messages',
                'dmesg': '/var/log/dmesg',
                'xend': '/var/log/xen/xend.log',
                'ovs-agent': '/var/log/ovs-agent.log'}
        if loglist:
            if not set(loglist) <= set(logs):
                raise Exception('log list contains items not in %s: %s' % (logs.keys(), loglist))
            logs = dict([(k, v) for (k, v) in logs.items() if k in loglist])
        result = {}
        for (logname, filename) in logs.items():
            if os.path.exists(filename):
                result[logname] = ''
                for i in (5, 4, 3, 2, 1):
                    rotate_filename = '%s.%s' % (filename, i)
                    if os.path.exists(rotate_filename):
                        fd = open(rotate_filename)
                        try:
                            result[logname] += fd.read()
                        finally:
                            fd.close()
                fd = open(filename)
                try:
                    result[logname] += fd.read()
                finally:
                    fd.close()
                result[logname] = xmlrpclib.Binary(result[logname])
        return result

    def get_support_files(self):
        name = ''.join(random.sample(string.ascii_lowercase, 6))
        for filename in glob.glob('/tmp/sosreport-%s-*.tar.bz2' % name):
            os.unlink(filename)
        cmd = ['/usr/sbin/sosreport', '--batch', '--name=%s' % name]
        run_cmd(cmd)
        filenames = glob.glob('/tmp/sosreport-%s-*.tar.bz2' % name)
        if not filenames:
            raise RuntimeError('No sosreport is generated.')
        filename = filenames[0]
        fd = open(filename)
        data = xmlrpclib.Binary(fd.read())
        fd.close()
        os.unlink(filename)
        os.unlink('%s.md5' % filename)
        return data

    def update_agent_password(self, username, password):
        update_password(username, password)
        reload_xmlrpc_server()

    def update_core_api_bindings(self, url, option=None):
        logging.debug('url: %s option: %s', url, option)
        new_url = re.sub("0\.0\.0\.0(?=(:|/))", agent.constants.MANAGER_SRC_IP, url, 1)
        logging.debug('new url: %s', new_url)
        filename = '/tmp/tmp-%s' % ''.join(random.sample(string.ascii_letters, 6))
        filename = self.ovs_download_file(new_url, filename, option)
        try:
            cmd = ['unzip', '-o', filename, '-d', get_python_lib()]
            run_cmd(cmd)
            reload_notification_server()
            logging.debug('new core api version: %s', get_core_api_version())
        finally:
            os.unlink(filename)

    def discover_server(self):
        """
        Discover the identity, version and other properties of the server.

        @param      None

        @return     Discovery data as an XML document.
                    Raises an exception on failure.
        """
        lock = filelock.FileLock('/var/run/ovs-agent/discover_server.lock')
        lock.acquire(wait=120)
        try:
            doc = SimpleDiscoveryXML("Discover_Server_Result")
            doc.push("Server")
            response = get_server_unique_id()
            if not response:
                raise Exception("Unable to obtain server unique id")
            doc.add("Unique_Id", response)

            doc.add("Boot_Time", str(self.get_last_boot_time()['last_boot_time']))

            (timezone, utc) = self.get_timezone()
            doc.push("Date_Time")
            doc.add("Time_Zone", str(timezone))
            doc.add("UTC", str(utc))
            doc.pop()

            (ntp_servers, local_time_source, is_ntp_running) = self.get_ntp()
            doc.push("NTP")
            for ntp_server in ntp_servers:
                doc.add("Server", ntp_server)
            doc.add("Local_Time_Source", str(local_time_source))
            doc.add("Is_NTP_Running", str(is_ntp_running))
            doc.pop()

            pkgs = self.list_package('ovs-agent')
            if not pkgs:
                raise Exception("unable to find package ovs-agent.")
            agent_version = pkgs[0]['version']
            agent_release = pkgs[0]['release']

            # RPM and agent versions are currently the same
            doc.add("Agent_Version", '-'.join([agent_version, agent_release]))
            doc.add("RPM_Version", '-'.join([agent_version, agent_release]))

            ovs_info = get_ovs_info()
            release = ovs_info['release']
            build = ovs_info['build']
            doc.add("OVM_Version", '-'.join([release, build]))

            doc.add("CPU_Type", platform.processor())
            doc.add("OS_Type", platform.system())

            d = {'OS_Name': '',
                 'OS_Major_Version': '',
                 'OS_Minor_Version': ''}
            try:
                d = parse_oracle_release()
            except:
                logging.exception('Error parsing /etc/oracle-release')

            doc.add("OS_Name", d["OS_Name"])
            doc.add("OS_Major_Version", d["OS_Major_Version"])
            doc.add("OS_Minor_Version", d["OS_Minor_Version"])

            doc.add("Hypervisor_Type", "xen")
            doc.add("Hypervisor_Name", "Xen")

            # FIXME: WE NEED TO SEPARATE HOST FROM HYPERVISOR
            self.discover_xen(doc)

            uuid = get_server_pool_uuid()
            if not uuid:
                uuid = "none"
            doc.add("Pool_Unique_Id", uuid)

            uuid = get_owning_manager_uuid()
            if not uuid:
                uuid = "none"
            doc.add("Manager_Unique_Id", uuid)
            doc.add("Hostname", socket.gethostname())

            d = dump_db(LOCAL_SERVER_DB)
            registered_ip = d.get("registered_ip")
            node_number = d.get("node_number")
            server_roles = d.get("roles")
            is_master = d.get("is_master")
            virtual_ip = d.get("pool_virtual_ip")
            stat_interval = d.get("stat_interval")

            if registered_ip:
                doc.add("Registered_IP", registered_ip)

            if node_number != None:
                doc.add("Node_Number", str(node_number))

            if not server_roles:
                server_roles = "unassociated"
            else:
                server_roles = ",".join(list(server_roles))

            doc.add("Server_Roles", server_roles)

            if is_master == True:
                master_str = "true"
            else:
                master_str = "false"
            doc.add("Is_Current_Master", master_str)

            if virtual_ip:
                doc.add("Master_Virtual_Ip", virtual_ip)

            doc.add("Manager_Core_API_Version", get_core_api_version() or '0.0.0.0')

            doc.add("Membership_State", get_membership_state())
            doc.add("Cluster_State", get_cluster_state())

            # Statistic
            stat_interval = stat_interval or MONITOR_SERVER_INTERVAL
            doc.push("Statistic")
            doc.add("Interval", str(stat_interval))
            doc.pop()

            # Exports
            discover_exports(doc)

            doc.pop() # Server

            return doc.finish()
        finally:
            lock.release()

    def discover_xen(self, doc):
        kernel_items = [ ("release", "Host_Kernel_Release"),
                         ("version", "Host_Kernel_Version"), ]

        version_items = [ ("xen_major", "Major"),
                          ("xen_minor", "Minor"),
                          ("xen_extra", "Extra"), ]

        xen_items = [ ("xen_caps", "Capabilities"), ]

        compiler_items = [ ("cc_compiler", "Compiler"),
                           ("cc_compile_by", "By"),
                           ("cc_compile_domain", "Domain"),
                           ("cc_compile_date", "Date"), ]

        from agent.hypervisor.xenxm import get_xen_info
        xeninfo = get_xen_info()

        for (xeninfo_name, discovery_name) in kernel_items:
            doc.add(discovery_name, xeninfo[xeninfo_name])

        doc.push("VMM")
        doc.push("Version")
        for (xeninfo_name, discovery_name) in version_items:
            doc.add(discovery_name, xeninfo[xeninfo_name])
        doc.pop()

        doc.push("Compile_Information")
        for (xeninfo_name, discovery_name) in compiler_items:
            doc.add(discovery_name, xeninfo[xeninfo_name])
        doc.pop()

        for (xeninfo_name, discovery_name) in xen_items:
            doc.add(discovery_name, xeninfo[xeninfo_name])

        doc.pop()

    def discover_hardware(self):
        """
        Discover information about the hardware present on this server.
        This includes CPU, I/O Adapters, Memory, and BIOS info.
        No information about storage devices is returned.

        Discovery of storage adapters is handled from python by forking and running
        the storage discovery utility.  Other discovery is handled by
        invoking a python extension that calls legacy C code from the VI agent.

        @param None

        @return    Returns the discovery data as an XML document.
                   Raises an exceptionon failure.
        """
        lock = filelock.FileLock('/var/run/ovs-agent/discover_hardware.lock')
        lock.acquire(wait=120)
        try:
            # Get the non-storage hardware, complete with headers
            (status, response, logmsg) = OVSDiscoverExtension.discover_hardware("")
            if status != 0:
                raise Exception(logmsg)

            # Find the marker for where to insert storage data
            idx = response.find(DISCOVERY_MARKER)
            marker_begin = response.rfind("\n", 0, idx)

            if idx < 0 or marker_begin < 0 or marker_begin >= idx:
                raise Exception("Internal Error: No storage discovery marker")

            indent = idx - marker_begin -1
            whole_marker = " " * indent + DISCOVERY_MARKER
            # Get the storage data
            (scsi_status, scsi_response, scsi_logmsg) = self.do_scsi_no_detail(indent)
            if scsi_status != 0:
                raise Exception(scsi_logmsg)

            response = response.replace(whole_marker, scsi_response)

            # Find the marker for where to insert server reserved data
            idx = response.find(SERVER_RESERVED_MARKER)
            marker_begin = response.rfind("\n", 0, idx)

            if idx < 0 or marker_begin < 0 or marker_begin >= idx:
                raise Exception("Internal Error: No server reserved")

            indent = idx - marker_begin -1
            whole_marker = " " * indent + SERVER_RESERVED_MARKER

            reserve_response = self.do_server_reserve(indent)

            return response.replace(whole_marker, reserve_response)
        finally:
            lock.release()

    def discover_physical_luns(self, args):
        """
        Discover information about one or more LUNs visible on this server.
        If an empty argument string is specified, return information about all LUNs
    
        @param      Space-separated list of SCSI IDs to report on
    
        @return    If successful, returns the discovery data in XML
                   Raises an exception on failure.
        """
        lock = filelock.FileLock('/var/run/ovs-agent/discover_physical_luns.lock')
        lock.acquire(wait=120)
        try:
            disk_list = args
            indent = 0
            (indent, header) = XML_header(indent)
            prog_args = [STORAGE_DISCOVER_PROG, scsi_debug_arg(),
                         "--indent", "%d" % indent,
                         "--infoDumpType", "disk_XML" ]
            if disk_list != "":
                # List is sent via stdin
                disk_list_string = disk_list.replace(" ", "\n") + "\n"
                prog_args.extend(["--diskList", "<<STDIN>>"])
            else:
                disk_list_string = ""
            proc = subprocess.Popen(prog_args, stdin=subprocess.PIPE,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE, close_fds=True)
            (stdout_string, stderr_string) = proc.communicate(disk_list_string)
            (indent, trailer) = XML_trailer(indent)
            if proc.returncode != 0 or stderr_string:
                raise Exception("discover_physical_luns: %s", stderr_string)
            return header + stdout_string + trailer
        finally:
            lock.release()

    def discover_mounted_file_systems(self, args):
        """
        Get mounted file systems of the specified type and NFS exports managed
        by the agent.

        @return     Returns the discovery data as an XML document.
                    Raises an exception on failure
        """

        class MountInfo:
            def __init__(self, fs_file, fs_spec, fs_mntops):
                self.fs_file = fs_file
                self.fs_spec = fs_spec
                self.fs_mntops = fs_mntops

        lock = filelock.FileLock('/var/run/ovs-agent/discover_mounted_file_systems.lock')
        lock.acquire(wait=120)
        try:
            f = open("/proc/mounts")
            all_mounts = None
            mounts = {}
            doc = SimpleDiscoveryXML("Discover_Mounted_File_Systems_Result")
            fstypes = args.split()
            for fstype in fstypes:
                if fstype == 'all':
                    all_mounts = 'all'
                    break
            for line in f:
                fs_spec, fs_file, fs_vfstype, fs_mntops, fs_freq, fs_passno = line.split()
                if not mounts.has_key(fs_vfstype):
                    mounts[fs_vfstype] = list()
                mounts[fs_vfstype].append(MountInfo(fs_file, fs_spec, fs_mntops))
            # return all of the mounted files systems or just the specified ones
            if all_mounts == 'all':
                fstypes = mounts.keys()
            for fstype in fstypes:
                mount_list = mounts.get(fstype)
                if mount_list != None:
                    doc.push("Filesystem", "Type", fstype)
                    for mount_entry in mount_list:
                        #doc.push("Mount", "Dir", mount_entry.fs_file)
                        doc.push("Mount", "Dir", mount_entry.fs_file.rstrip('/'))
                        doc.add("Device", mount_entry.fs_spec)
                        doc.add("Mount_Options", mount_entry.fs_mntops)
                        doc.pop()
                    doc.pop()
            return doc.finish()
        finally:
            lock.release()

    def do_scsi_no_detail(self, indent):
        # The output here will get inserted into already formatted discover output
        args = [STORAGE_DISCOVER_PROG, scsi_debug_arg(),
                "--indent", "%d" % indent,
                "--infoDumpType", "io_host_XML"]

        proc = subprocess.Popen(args, stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE, close_fds=True)
        (stdout_string, stderr_string) = proc.communicate()
        return (proc.returncode, stdout_string, stderr_string)


    def do_server_reserve(self, indent):

        exclusions = ''
        exclusions_list = get_lun_exclusions()
        if exclusions_list:
            exclusions = XML_tag("ServerReserved", indent, True)
            lun = 0
            for entry in exclusions_list:
                exclusions += XML_tag('Disk Num="'+ ("%d" % lun) + '"', indent + 2, True)
                exclusions += XML_tag_with_value("WWID", indent + 4, entry)
                exclusions += XML_tag("Disk", indent + 2, False)
                lun += 1
            exclusions += XML_tag("ServerReserved", indent, False)

        return exclusions


    # Begin of copying from yum-config-manager.py
    def writeRawConfigFile(self, filename, sectionname, cfgoptions, items, optionobj, only=None):
        """
        From writeRawRepoFile, but so we can alter [main] too.
        """
        ini = INIConfig(open(filename))
        # Updated the ConfigParser with the changed values
        cfgOptions = cfgoptions(sectionname)
        for name,value in items():
            if value is None: # Proxy
                continue
            option = optionobj(name)
            if option.default != value or name in cfgOptions :
                if only is None or name in only:
                    ini[sectionname][name] = option.tostring(value)
        fp =file(filename, "w")
        fp.write(str(ini))
        fp.close()
    # End of copy

    def copy_file(self, src, dst, sparse=True, update_period=20):
        set_nice()
        args = ['/bin/dd', 'if=%s' % src, 'of=%s' % dst, 'bs=1M', 'iflag=direct', 'oflag=direct']
        if sparse:
            args.append('conv=sparse')

        src_stat = os.stat(src)
        src_size = 0
        if stat.S_ISREG(src_stat[stat.ST_MODE]):
            src_size = src_stat[stat.ST_SIZE]
        elif stat.S_ISBLK(src_stat[stat.ST_MODE]):
            src_size = run_cmd(['/sbin/blockdev', '--getsize64', src])

        proc = subprocess.Popen(args, stderr=subprocess.PIPE, close_fds=True)

        def progress_reporter():
            try:
                logging.debug('progress reporter for (%s) started', ' '.join(args))
                dd_stat_re = re.compile(r'^(?P<copied>\d+) bytes? (.*) copied.*$')
                time.sleep(update_period)
                proc.poll()
                while proc.returncode is None:
                    try:
                        os.kill(proc.pid, signal.SIGUSR1)
                    except OSError, err:
                        if err[0] == errno.ESRCH:
                            break
                        else:
                            raise
                    while True:
                        dd_stat = proc.stderr.readline()
                        m = dd_stat_re.match(dd_stat)
                        if m:
                            copied = m.group('copied')
                            status = {'OPERATION': 'copy',
                                      'CURRENT_BYTES': long(copied),
                                      'TOTAL_BYTES': long(src_size),
                                      'OBJECT': 'file',
                                      'CURRENT_OBJECT_COUNT': 1,
                                      'TOTAL_OBJECT_COUNT': 1}
                            async_proc_prog(status)
                            break
                    time.sleep(update_period)
                    proc.poll()
                logging.debug('progress reporter for (%s) finished', ' '.join(args))
            except Exception, err:
                logging.debug('progress reporter for (%s) error: %s', ' '.join(args), err)

        threading.Thread(target=progress_reporter).start()

        proc.wait()
        if proc.returncode != 0:
            raise RuntimeError('Command: (%s) failed: (%s)' % (' '.join(args), proc.returncode))

    def ovs_download_file(self, url, filename, option=None, obj='file', obj_current=1,
                          obj_total=1, update_period=20):
        return ovs_download_file(url, filename, option, obj, obj_current, obj_total, update_period)

    def clean_aproc_db(self):
        for (key, pid) in dump_db(LOCAL_APROC_DB).items():
            if not os.path.exists("/proc/%s" % pid):
                delete_item(LOCAL_APROC_DB, key)

    def get_proc_id(self, pid):
        return read_item(LOCAL_APROC_DB, str(pid))

    def remove_proc_id(self, pid):
        delete_item(LOCAL_APROC_DB, str(pid))

    def ovs_async_proc(self, func, *args):
        """
        Runs func as a daemon, add it PID to the process DB table (aproc)

        @param func    The function to run
        @param args    A list of arguments to pass to the function

        @return        If successful, returns the process ID.
                       Raises an exception on failure.

        Restriction:
            The function must be known to OVSServices, otherwise this call will fail.
        """
        if func == None or func == '':
            raise Exception ("ovs_async_proc: NULL function name")
        self.clean_aproc_db()

        from agent.service import get_api
        api = get_api()
        function = api.get(func)
        if function == None:
            raise Exception('ovs_async_proc: function: %s not found ' % func)
        logging.debug('async call function: %s', func)

        pid = os.fork()
        if pid > 0:
            os._exit(0)

        os.chdir("/")
        os.setsid()
        os.umask(022)

        pid = os.fork()
        if pid > 0:
            write_item(LOCAL_APROC_DB, str(pid), pid)
            return pid

        pid = os.getpid()
        status = 0
        try:
            im_self = getattr(function, 'im_self', None)
            if im_self:
                output = function(im_self, *args)
            else:
                output = function(*args)
        except Exception, err:
            status = 1
            output = str(err)

        delete_item(LOCAL_APROC_DB, str(pid))

        if output is None:
            output = ''

        logging.debug('ovs_async_proc: pid %s status %s output %s', pid, status, output)
        try:
            notify_manager('{ASYNC_PROC} exit PID %s' % pid, {'status': status, 'output': output})
        except Exception, err:
            logging.warn('Error sending notification: pid(%s) status(%s) output(%s) error(%s)' % (pid, status, output, err))

    def ovs_async_proc_status(self, pid):
        """
        Gets the running status of a daemon as defined by the Linux kernel

        @param pid     Process ID

        @return        If successful, returns procesess running status
                       Raises an exception on failure.
        Restriction:
           - PID must exists in the process DB table (aproc)
        """
        if pid == None or pid == '':
            raise Exception('ovs_async_proc_status: PID is NULL')

        id = self.get_proc_id(pid)
        if id == None:
            raise Exception('ovs_async_proc_status: PID %d does not exist in DB' % int(pid))

        if not os.path.exists("/proc/%s"  % pid):
            raise Exception('ovs_async_proc_status: Process %d does not exist' % int(pid))

        f = open('/proc/%d/status' % int(pid),'r')
        try:
            lines = f.readlines()
        finally:
            f.close()
        for ln in lines:
            if ln.startswith( 'State:' ):
                junk, state, text = ln.split( None, 2 )
                return "Process %d  state: %s %s" % (int(pid), state, text)

    def ovs_async_proc_signal(self, sig, pid):
        """
        Sends SIGKILL, SIGUSR1 to a running process

        @param sig     (kill | stop)
                       kill: sends a SIGKILL to the process
                       stop: sends a SIGUSR1 to the process
        @param pid     Process ID to signal

        @return        If successful, returns procesess ID, and signal delivered
                       Raises an exception on failure.

        Restriction:
           - For a process to receive SIGUSR1, a processes must register a
             signal handler. The purpose of this signal is to tell the process
             to clean up and exit vs just killing it.
           - PID must exists in the process DB table (aproc)
        """
        if sig != "stop" and sig != "kill":
            raise Exception("ovs_async_proc_signa: signal %s is not supported" % sig)

        id = self.get_proc_id(pid)
        if id == None:
            raise Exception('ovs_async_proc_signal: PID %d does not exist in DB' % int(pid))

        if not os.path.exists("/proc/%s"  % pid):
            raise Exception('ovs_async_proc_status: Process %d does not exist' % int(pid))

        if sig == "kill":
            try:
                os.kill(int(pid), signal.SIGKILL)
                self.remove_proc_id(pid)
                return "Process %d is killed" % int(pid)
            except OSError, e:
                raise Exception('ovs_async_proc_signal: failed to send sig %s to proc %d => %s' % (sig, int(pid), str(e)))
        else:
            try:
                os.kill(int(pid), signal.SIGUSR1)
                #In order to prevent the DB from getting out of sync, do not remove the PID
                #at this point, as the function may not respond to SIGUSR1 and continue to run.
                #the PID will get cleaned up the next ovs_async_proc() gets called.
                #remove_proc_id(pid)
                return "Process %d is stopped" % int(pid)
            except OSError, e:
                raise Exception('ovs_async_proc_signal: failed to send sig %s to proc %d => %s' % (sig, int(pid), str(e)))

    def get_last_boot_time(self):
        # both are seconds since the epoch, in UTC.
        return {'local_time': long(time.time()),
                'last_boot_time': long(get_boot_time()[0])}

    def export_fs(self, export_uuid, export_type, client, path, options=None):
        """
        Export the given path.

        @param      export_uuid - uuid of the export
        @param      export_type - type of the export
        @param      client - client IP address or hostname
        @param      path - mount point to be exported
        @param      options - exportfs options

        @return     None
                    Raise an exception on failure.
        """
        if export_type == 'nfs':
            do_export_nfs(client, path, options)
            d = {'export_uuid': export_uuid,
                 'export_type': export_type,
                 'client': client,
                 'path': path,
                 'options': options}
            write_item(LOCAL_EXPORTS_DB, export_uuid, d)
        else:
            raise NotImplementedError()

    def unexport_fs(self, export_uuid):
        """
        Unexport the given path.

        @param      export_uuid - uuid of the export

        @return     None
                    Raise an exception on failure.
        """
        d = read_item(LOCAL_EXPORTS_DB, export_uuid)
        if d:
            if d['export_type'] == 'nfs':
                do_unexport_nfs(d['client'], d['path'])
                delete_item(LOCAL_EXPORTS_DB, export_uuid)
            else:
                raise NotImplementedError()

def get_ovs_info():
    ovs_info = {}
    f = open("/etc/ovs-info")
    for line in f:
        m = re.match(r'^release:(.*)$', line)
        if m:
            ovs_info['release'] = m.group(1).strip()
        m = re.match(r'^date:(.*)$', line)
        if m:
            ovs_info['date'] = m.group(1).strip()
        m = re.match(r'^build:(.*)$', line)
        if m:
            ovs_info['build'] = m.group(1).strip()
    f.close()
    return ovs_info


def get_core_api_version():
    try:
        f = open(os.path.join(get_python_lib(), 'version.properties'))
        d = {}
        for line in f:
            if '=' not in line:
                continue
            (k, v) = line.split('=', 1)
            k = k.strip()
            v = v.strip()
            d[k] = v
        f.close()
        return '.'.join([d['major.version'], d['minor.version'],
                         d['patch.version'], d['bld.version']])
    except Exception, err:
        logging.warn('cannot get core api version: %s', err)


class OVSMeter(BaseMeter):
    def __init__(self, update_period, obj, obj_current, obj_total):
        BaseMeter.__init__(self)
        self.update_period = update_period
        self.obj = obj
        self.obj_current = obj_current
        self.obj_total = obj_total

    def send_notification(self, amount_read, now=None):
        if self.size is not None:
            status = {'OPERATION': 'download',
                      'CURRENT_BYTES': long(self.last_amount_read),
                      'TOTAL_BYTES': long(self.size),
                      'OBJECT': self.obj,
                      'CURRENT_OBJECT_COUNT': self.obj_current,
                      'TOTAL_OBJECT_COUNT': self.obj_total}
            async_proc_prog(status)

    def _do_update(self, amount_read, now=None):
        self.send_notification(amount_read, now)

    def _do_end(self, amount_read, now=None):
        self.send_notification(amount_read, now)


def ovs_download_file(url, filename, option=None, obj='file', obj_current=1,
                      obj_total=1, update_period=20):
    meter = OVSMeter(update_period, obj, obj_current, obj_total)
    defaults = {'progress_obj': meter,
                'copy_local': 1,
                'ssl_verify_peer': False,
                'ssl_verify_host': False}
    if option is None:
        option = defaults
    else:
        option.update(defaults)
    try:
        return urlgrab(url, filename, **option)
    except Exception, err:
        try:
            os.unlink(filename)
        except:
            pass
        raise err

def get_lun_exclusions():
    luns = []
    if os.path.exists("/etc/boot_sysdisk_wwid"):
        f = open("/etc/boot_sysdisk_wwid", 'r')
        try:
            lines = f.readlines()
        finally:
            f.close()
        for line in lines:
            if line.startswith('/ = '):
                line = line.replace('/ = ',"").strip()
                luns.append(line)
    return luns

def do_export_nfs(client, path, options):
    if not os.path.ismount(path):
        logging.warn('%s is not a mount point', path)
    if not options:
        options = 'rw,async,no_root_squash'
    cmd = ['service', 'nfs', 'start']
    run_cmd(cmd)
    client_path = '%s:%s' % (client, path)
    cmd = ['exportfs', '-o', options, client_path]
    run_cmd(cmd)

def do_unexport_nfs(client, path):
    client_path = '%s:%s' % (client, path.rstrip('/'))
    cmd = ['exportfs', '-u', client_path]
    run_cmd(cmd)
    if not run_cmd(['exportfs']):
        cmd = ['service', 'nfs', 'stop']
        run_cmd(cmd)

def init_nfs_exports():
    for v in dump_db(LOCAL_EXPORTS_DB).values():
        try:
            if v['export_type'] == 'nfs':
                do_export_nfs(v['client'], v['path'], v['options'])
        except Exception, e:
            logging.exception('Error exporting %s: %s', v, e)

def discover_exports(doc):
    doc.push('Exports')
    for v in dump_db(LOCAL_EXPORTS_DB).values():
        doc.push('Export', 'UUID', v['export_uuid'])
        doc.add('Type', v['export_type'])
        doc.add('Client', v['client'])
        doc.add('Path', v['path'])
        doc.add('Options', v['options'])
        doc.pop() # Export
    doc.pop() # Exports


def XML_header(indent):
    header = '<?xml version="1.0" encoding="UTF-8"?>\n'
    indent += XML_INDENT

    header = header + XML_tag("Discover_Physical_LUNS_Result", indent, True)
    indent += XML_INDENT
    return (indent, header)


def XML_trailer(indent):
    indent -= XML_INDENT
    trailer = XML_tag("Discover_Physical_LUNS_Result", indent, False)
    return (indent, trailer)


def scsi_debug_arg():
    if SCSI_INFO_DEBUG_LVL > 0:
        return "--debugLvl %d" % SCSI_INFO_DEBUG_LVL
    else:
        return ""


def XML_tag_with_value(element, indent, value):
    template = "%-*s<%s>%s</%s>\n"
    if element == "" or indent < 0:
        return "? invalid arguments"

    return template % (indent, "", element, value, element)


def XML_tag(element, indent, start):
    template = "%-*s<%s%s>\n"
    if element == "" or indent < 0:
        return "? invalid arguments"

    if start:
        close = ""
    else:
        close = "/"
    return template % (indent, "", close, element)


def parse_oracle_release():
    f = open('/etc/oracle-release')
    try:
        s = f.read()
    finally:
        f.close()
    l = map(string.strip, s.split('release'))
    d = {}
    d['OS_Name'] = l[0]
    d['OS_Major_Version'], d['OS_Minor_Version'] = l[1].split('.')
    return d
