# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import re
import os
import time
import string
import random
import shutil
import logging
import fnmatch
import tarfile
import zipfile
import commands
import threading
import subprocess
import rpm
import uuid
from xen.xend import sxp
from xen.util.blkif import blkdev_name_to_number
from xen.xend.xenstore.xstransact import xstransact
from xen.xend.xenstore.xsutil import GetDomainPath
from agent import ovfxen
from agent.utils import filelock, xmlrpclib
from agent.common import guess_mimetype, set_nice
from agent.constants import OVS_REPO_HEAD
from agent.common import run_cmd, get_uuid
from agent.db import write_item, delete_item, get_cluster_db_home
from agent.remaster import dlm_acquire_lock, dlm_release_lock
from agent.notification import async_proc_prog
from agent.host.linux import ovs_download_file
from agent.hypervisor.base import Hypervisor


MAX_KEYNAME_LEN = 256
MAX_KEYVALUE_LEN = 8192
MAX_ENTRY_SIZE = 2048
MAX_MESSAGE_NUMBER = 256


def check_cfg(cfg):
    # FIXME: Are we going to support arbitrary options? Yes for now.
    for (option, value) in cfg.items():
        if option in VM_CFG:
            checker = VM_CFG[option][1]
            try:
                checker(value)
            except Exception, err:
                raise RuntimeError('%s:type_error:%s' % (option, err))


def fix_cfg(cfg):
    for (option, value) in cfg.items():
        if option in VM_CFG:
            checker = VM_CFG[option][1]
            if checker == check_int:
                try:
                    cfg[option] = int(value)
                except ValueError, err:
                    raise RuntimeError('Cannot fix option (%s) value (%s): %s' % (option, value, err))


def create(path, params):
    check_cfg(params)
    if not os.path.exists(os.path.dirname(path)):
        os.makedirs(os.path.dirname(path))
    try:
        write_cfg(path, params)
    except Exception:
        delete(path)
        raise


def delete(path):
    shutil.rmtree(os.path.dirname(path))


def parse_domain_info(info_sxp):
    dom_info = {}
    dom_info['device'] = {}
    dom_info['image'] = {}
    for (k, v) in info_sxp:
        if k == 'device':
            if v[0] not in dom_info['device']:
                dom_info['device'][v[0]] = []
            d = {}
            for item in v[1:]:
                if len(item) == 2:
                    d[item[0]] = item[1]
                elif len(item) > 2:
                    d[item[0]] = item[1:]
            dom_info['device'][v[0]].append(d)
        elif k == 'image':
            dom_info['builder'] = v[0]
            for item in v[1:]:
                if len(item) == 2:
                    dom_info['image'][item[0]] = item[1]
                elif len(item) > 2:
                    dom_info['image'][item[0]] = item[1:]
        else:
            dom_info[k] = v
    return dom_info


def configure(path, params):
    check_cfg(params)
    for option in sorted(params.keys()):
        value = params[option]
        if option in VM_CFG:
            setter = VM_CFG[option][0]
            if is_running(get_vm_uuid(path)) and setter:
                try:
                    setter(get_name(path), value)
                except Exception, err:
                    raise RuntimeError('%s:setup_error:%s' % (option, err))
        try:
            update_cfg(path, option, value)
        except Exception, err:
            raise RuntimeError('%s:config_error:%s' % (option, err))


def unconfigure(path, params):
    cfg = read_cfg(path)
    changed = False
    for option in params:
        if option in cfg:
            del cfg[option]
            changed = True
    if changed:
        write_cfg(path, cfg)


def find_ovf_descriptor(unpacked_path):
    ovf_filenames = find_file('*.ovf', unpacked_path)
    if not ovf_filenames:
        raise RuntimeError('Cannot find OVF descriptor file.')
    if len(ovf_filenames) > 1:
        raise RuntimeError('Multiple OVF descriptor files found.')
    return ovf_filenames[0]


def check_ova(filename):
    if not tarfile.is_tarfile(filename):
        raise RuntimeError('Invalid OVA package: not a tarfile.')


def format_uuid(uuid):
    _uuid = uuid.replace("-", "").lower()
    if len(_uuid) != 32:
        raise RuntimeError('Invalid UUID: %s' % uuid)
    return "-".join([_uuid[0:8], _uuid[8:12], _uuid[12:16], _uuid[16:20], _uuid[20:]])


def get_lock_file(repo_id, lock_id):
    return '/var/run/ovs-agent/vm-%s-%s.lock' % (repo_id, lock_id)


def get_vm_path(repo_id, vm_id):
    return os.path.join(OVS_REPO_HEAD, repo_id, 'VirtualMachines', vm_id, 'vm.cfg')


def get_template_path(repo_id, template_id):
    return os.path.join(OVS_REPO_HEAD, repo_id, 'Templates', template_id, 'vm.cfg')


def get_assembly_path(repo_id, assembly_id):
    return os.path.join(OVS_REPO_HEAD, repo_id, 'Assemblies', assembly_id)


def get_virtualdisk_path(repo_id, virtualdisk_id):
    return os.path.join(OVS_REPO_HEAD, repo_id, 'VirtualDisks', virtualdisk_id)


def get_iso_path(repo_id, iso_id):
    return os.path.join(OVS_REPO_HEAD, repo_id, 'ISOs', iso_id)


def find_file(pattern, root=os.curdir):
    result = []
    abs_root = os.path.abspath(root)
    real_root = os.path.realpath(abs_root)
    for (path, dirs, files) in os.walk(real_root):
        for filename in files:
            if fnmatch.fnmatch(filename, pattern):
                abs_path = os.path.join(path, filename)
                real_path = os.path.realpath(abs_path)
                if real_path.startswith(real_root):
                    result.append(real_path)
    return result


def get_directory_size(directory):
    du_re = re.compile(r'^(?P<size>\d+)\s.*$')
    output = run_cmd(['du', '-s', '-b', directory])
    match = du_re.match(output)
    if not match:
        raise Exception('Error getting directory size.')
    return int(match.group('size'))


def run_cmd_with_progress(cmd, directory, obj, obj_current, obj_total, update_period, filesize):
    if not os.path.exists(directory):
        os.makedirs(directory)
    proc = subprocess.Popen(cmd, stderr=subprocess.PIPE, close_fds=True, shell=True)

    def progress_reporter():
        try:
            logging.debug('progress reporter for (%s) started', cmd)
            directory_size_orig = get_directory_size(directory)
            time.sleep(update_period)
            proc.poll()
            while proc.returncode is None:
                directory_size_new = get_directory_size(directory)
                unpacked_size = directory_size_new - directory_size_orig
                status = {'OPERATION': 'unpack',
                          'CURRENT_BYTES': long(unpacked_size),
                          'TOTAL_BYTES': long(filesize),
                          'OBJECT': obj,
                          'CURRENT_OBJECT_COUNT': obj_current,
                          'TOTAL_OBJECT_COUNT': obj_total}
                async_proc_prog(status)
                time.sleep(update_period)
                proc.poll()
            logging.debug('progress reporter for (%s) finished', cmd)
        except Exception, err:
            logging.debug('progress reporter for (%s) error: %s', cmd, err)

    threading.Thread(target=progress_reporter).start()
    proc.wait()
    if proc.returncode != 0:
        raise RuntimeError('Command: (%s) failed: (%s)' % (cmd, proc.returncode))


def unzip_file(filename, directory, obj, obj_current, obj_total, update_period):
    filesize_re = re.compile(r'^.*\b(?P<size>\d+)\s+\d+\s+files?$', re.DOTALL)
    zipinfo = run_cmd(['unzip', '-l', filename])
    match = filesize_re.match(zipinfo)
    if not match:
        raise RuntimeError('Unrecognized zip file: %s' % filename)
    filesize = int(match.group('size'))
    cmd = 'unzip -o %s -d %s' % (filename, directory)
    run_cmd_with_progress(cmd, directory, obj, obj_current, obj_total, update_period, filesize)


def unpack_rpm(filename, directory, obj, obj_current, obj_total, update_period):
    ts = rpm.TransactionSet()
    fd = os.open(filename, os.O_RDONLY)
    hdr = ts.hdrFromFdno(fd)
    filesize = hdr[rpm.RPMTAG_SIZE]
    os.close(fd)
    cmd = 'rpm2cpio %s | (cd %s; cpio -iumd)' % (filename, directory)
    run_cmd_with_progress(cmd, directory, obj, obj_current, obj_total, update_period, filesize)


def untar_file(filename, directory, obj, obj_current, obj_total, update_period):
    ftype = guess_mimetype(filename)
    if ftype == 'application/x-tar':
        cmd = 'tar xf %s --no-same-owner -C %s' % (filename, directory)
    elif ftype == 'application/x-gzip':
        cmd = 'tar xzf %s --no-same-owner -C %s' % (filename, directory)
    elif ftype == 'application/x-bzip2':
        cmd = 'tar xjf %s --no-same-owner -C %s' % (filename, directory)
    else:
        logging.debug('unsupported tarball format: %s', ftype)
        return
    run_cmd_with_progress(cmd, directory, obj, obj_current, obj_total, update_period, 0)


def unpack_file(filename, directory, obj='file', obj_current=1, obj_total=1, update_period=20):
    if tarfile.is_tarfile(filename):
        untar_file(filename, directory, obj, obj_current, obj_total, update_period)
    elif zipfile.is_zipfile(filename):
        unzip_file(filename, directory, obj, obj_current, obj_total, update_period)
    elif filename.endswith('.rpm'):
        unpack_rpm(filename, directory, obj, obj_current, obj_total, update_period)


def unpack_files(root, unpacked, obj, obj_current, obj_total):
    need_unpack = True
    while need_unpack:
        need_unpack = False
        for (path, dirs, files) in os.walk(root):
            for filename in files:
                filepath = os.path.join(path, filename)
                if filepath in unpacked:
                    continue
                unpack_file(filepath, path, obj, obj_current, obj_total)
                unpacked.append(filepath)
                need_unpack = True


def move_files(src, dst):
    for (path, dirs, files) in os.walk(src):
        for filename in files:
            src_path = os.path.join(path, filename)
            dst_path = os.path.join(dst, filename)
            shutil.move(src_path, dst_path)


def make_filename(filename=''):
    valid_chars = '_-.' + string.digits + string.ascii_letters
    filename = os.path.basename(filename)
    filename = ''.join([x for x in filename if x in valid_chars])
    filename = filename.lstrip('.')
    if not filename:
        filename = ''.join(random.sample(string.ascii_letters, 6))
    return filename


def read_cfg(path):
    cfg = {}
    execfile(path, {}, cfg)
    return cfg


def write_cfg(path, cfg):
    f = open(path, 'w')
    for (k, v) in cfg.items():
        f.write('%s = %s\n' % (k, repr(v)))
    f.close()


def update_cfg(path, option, value):
    cfg = read_cfg(path)
    if option not in cfg or cfg[option] != value:
        cfg[option] = value
        write_cfg(path, cfg)


def get_name(path):
    return read_cfg(path)['name']


def get_vm_uuid(path):
    return read_cfg(path)['uuid']


def get_domid(domain):
    return int(run_cmd(['xm', 'domid', domain]))


def is_running(vm_uuid):
    if xstransact.List('/vm/%s' % uuid.UUID(vm_uuid)):
        return True
    else:
        return False


def is_vm_running(repo_id, vm_id):
    try:
        path = get_vm_path(repo_id, vm_id)
        cfg = read_cfg(path)
        # is it running on this server?
        if is_running(cfg["uuid"]):
            return True
        # is it running somewhere else?
        uuid = format_uuid(cfg["uuid"])
        try:
            dlm_acquire_lock(uuid)
            dlm_release_lock(uuid)
            return False
        except:
            return True
    except Exception, e:
        logging.debug("Error checking if VM is running: %s", e)
        # assume it is not running
        return False


def get_xen_info():
    xen_info = {}
    output = run_cmd(['xm', 'info'])
    for line in output.splitlines():
        (attr, value) = line.split(':', 1)
        xen_info[attr.strip()] = value.strip()
    return xen_info


def set_name(domain, value):
    run_cmd(['xm', 'rename', domain, value])


def set_memory(domain, value):
    run_cmd(['xm', 'mem-set', domain, value])


def set_maxmem(domain, value):
    run_cmd(['xm', 'mem-max', domain, value])


def set_cpus(domain, value):
    if isinstance(value, str):
        if value:
            run_cmd(['xm', 'vcpu-pin', domain, 'all', value])
        else:
            run_cmd(['xm', 'vcpu-pin', domain, 'all', 'all'])
    else:
        for (i, v) in enumerate(value):
            if v:
                run_cmd(['xm', 'vcpu-pin', domain, i, v])
            else:
                run_cmd(['xm', 'vcpu-pin', domain, i, 'all'])


def set_vcpus(domain, value):
    run_cmd(['xm', 'vcpu-set', domain, value])


def set_cpu_cap(domain, value):
    run_cmd(['xm', 'sched-credit', '-d', domain, '-c', value])


def set_cpu_weight(domain, value):
    run_cmd(['xm', 'sched-credit', '-d', domain, '-w', value])


def set_disk(domain, value):
    vbds_old = set()
    cdroms_old = set()
    cdroms_info_old = {}
    info = run_cmd(['xm', 'list', '--long', domain])
    info_sxp = sxp.from_string(info)
    domain_info = parse_domain_info(info_sxp[1:])
    for vbd in domain_info['device']['vbd']:
        front_dev = re.sub(r'^/dev/(.*)', r'\1', vbd['dev'])
        if front_dev.endswith(':cdrom'):
            cdroms_old.add(front_dev)
            cdroms_info_old[front_dev] = [vbd.get('uname', ''), front_dev, vbd['mode']]
        else:
            vbds_old.add((vbd['uname'], front_dev.split(':')[0], vbd['mode']))

    vbds_new = set()
    cdroms_new = set()
    cdroms_info_new = {}
    for entry in value:
        backend, front_dev, mode = entry.split(',')
        front_dev = re.sub(r'^/dev/(.*)', r'\1', front_dev)
        if front_dev.endswith(':cdrom'):
            cdroms_new.add(front_dev)
            cdroms_info_new[front_dev] = [backend, front_dev, mode]
        else:
            vbds_new.add((backend, front_dev, mode))

    vbds_add = vbds_new - vbds_old
    vbds_remove = vbds_old - vbds_new
    cdroms_add = cdroms_new - cdroms_old
    cdroms_remove = cdroms_old - cdroms_new
    cdroms_changed = cdroms_old & cdroms_new

    for entry in vbds_remove:
        devid = blkdev_name_to_number(entry[1].split(':')[0])[1]
        run_cmd(['xm', 'block-detach', domain, devid])
    for entry in cdroms_remove:
        devid = blkdev_name_to_number(entry.split(':')[0])[1]
        run_cmd(['xm', 'block-detach', domain, devid])
    for entry in vbds_add:
        run_cmd(['xm', 'block-attach', domain] + list(entry))
    for entry in cdroms_add:
        run_cmd(['xm', 'block-attach', domain] + cdroms_info_new[entry])
    for entry in cdroms_changed:
        run_cmd(['xm', 'block-configure', domain] + cdroms_info_new[entry])


def set_vif(domain, value):
    output = run_cmd(['xm', 'network-list', '--long', domain])
    vifs_sxpr = sxp.all_from_string(output)
    vifs_old = set()
    for (vif_id, vif_desc) in vifs_sxpr:
        for (k, v) in vif_desc:
            if k == 'mac':
                vifs_old.add(v)

    vifs_new = set()
    vifs_new_info = {}
    for entry in value:
        # FIXME: when there's no 'mac' in the entry? Just continue for now.
        m = re.search(r'mac\s*=\s*([a-fA-F0-9:]+)', entry)
        if not m:
            continue
        mac = m.group(1)
        vifs_new.add(mac)
        vifs_new_info[mac] = entry

    add_to = list(vifs_new - vifs_old)
    remove_from = list(vifs_old - vifs_new)
    for entry in remove_from:
        run_cmd(['xm', 'network-detach', domain, entry])
    for entry in add_to:
        run_cmd(['xm', 'network-attach', domain] + vifs_new_info[entry].split(','))


def set_disk_other_config(domain, value):
    output = run_cmd(['xm', 'block-list', '--long', domain])
    vbds_sxpr = sxp.all_from_string(output)
    vbds_info = {}
    for (vbd_id, vbd_desc) in vbds_sxpr:
        vbds_info[vbd_id] = {}
        for (k, v) in vbd_desc:
            vbds_info[vbd_id][k] = v
    for (front_end, algorithm, params) in value:
        front_id = blkdev_name_to_number(front_end)[1]
        if front_id in vbds_info and 'backend' in vbds_info[front_id]:
            run_cmd(['xenstore-write', '%s/qos_algorithm_type' % vbds_info[front_id]['backend'], algorithm])
            run_cmd(['xenstore-write', '%s/qos_algorithm_params' % vbds_info[front_id]['backend'], params])


def set_vif_other_config(domain, value):
    output = run_cmd(['xm', 'network-list', '--long', domain])
    vifs_sxpr = sxp.all_from_string(output)
    vifs_info = {}
    for (vif_id, vif_desc) in vifs_sxpr:
        vifs_info[vif_id] = {}
        for (k, v) in vif_desc:
            vifs_info[vif_id][k] = v
    for (mac, algorithm, params) in value:
        for vif in vifs_info.values():
            if 'mac' in vif and 'backend' in vif and vif['mac'] == mac.lower():
                run_cmd(['xenstore-write', '%s/qos_algorithm_type' % vif['backend'], algorithm])
                run_cmd(['xenstore-write', '%s/qos_algorithm_params' % vif['backend'], params])


def check_str(value):
    if not isinstance(value, str):
        raise RuntimeError('A string is required, but %s is provided.' % type(value))


def check_int(value):
    try:
        int(value)
    except ValueError, err:
        raise RuntimeError('An integer is required: %s.' % err)


def check_list(value):
    if not isinstance(value, list):
        raise RuntimeError('A list is required, but %s is provided.' % type(value))


def check_str_or_list(value):
    if not isinstance(value, (str, list)):
        raise RuntimeError('A string or list is required, but %s is provided.' % type(value))


VM_CFG =  {
    # option: (setter, checker)
    'access_control': (None, check_list),
    'acpi': (None, check_int),
    'apic': (None, check_int),
    'blkif': (None, check_int),
    'bootargs': (None, check_str),
    'bootloader': (None, check_str),
    'boot': (None, check_str),
    'builder': (None, check_str),
    'console_autoconnect': (None, check_int),
    'cpu_cap': (set_cpu_cap, check_int),
    'cpuid_check': (None, check_list),
    'cpuid': (None, check_list),
    'cpus': (set_cpus, check_str_or_list),
    'cpu': (None, check_int),
    'cpu_weight': (set_cpu_weight, check_int),
    'device_model': (None, check_str),
    'dhcp': (None, check_str),
    'disk_other_config': (set_disk_other_config, check_list),
    'disk': (set_disk, check_list),
    'display': (None, check_int),
    'expose_host_uuid': (None, check_int),
    'extra': (None, check_str),
    'fda': (None, check_str),
    'fdb': (None, check_str),
    'features': (None, check_str),
    'gateway': (None, check_str),
    'guest_os_type': (None, check_str),
    'hap': (None, check_int),
    'hostname': (None, check_str),
    'hpet': (None, check_int),
    'interface': (None, check_str),
    'ioports': (None, check_list),
    'ip': (None, check_str),
    'irq': (None, check_list),
    'isa': (None, check_int),
    'kernel': (None, check_str),
    'keymap': (None, check_str),
    'loader': (None, check_str),
    'localtime': (None, check_int),
    'machine_address_size': (None, check_int),
    'maxmem': (set_maxmem, check_int),
    'maxvcpus': (None, check_int),
    'memory': (set_memory, check_int),
    'monitor_path': (None, check_str),
    'monitor': (None, check_int),
    'name': (set_name, check_str),
    'netif': (None, check_int),
    'netmask': (None, check_str),
    'nfs_root': (None, check_str),
    'nfs_server': (None, check_str),
    'nographic': (None, check_int),
    'nomigrate': (None, check_int),
    'on_crash': (None, check_str),
    'on_poweroff': (None, check_str),
    'on_reboot': (None, check_str),
    'on_xend_start': (None, check_str),
    'on_xend_stop': (None, check_str),
    'opengl': (None, check_int),
    'pae': (None, check_int),
    'pasued': (None, check_int),
    'pci_msitranslate': (None, check_int),
    'pci_power_mgmt': (None, check_int),
    'pci': (None, check_list),
    'ramdisk': (None, check_str),
    'root': (None, check_str),
    'rtc_timeoffset': (None, check_int),
    's3_integrity': (None, check_int),
    'sdl': (None, check_int),
    'serial': (None, check_str),
    'shadow_memory': (None, check_int),
    'soundhw': (None, check_str),
    'stdvga': (None, check_int),
    'superpages': (None, check_int),
    'suppress_spurious_page_faults': (None, check_int),
    'target': (None, check_int),
    'timer_mode': (None, check_int),
    'tpmif': (None, check_int),
    'tsc_mode': (None, check_int),
    'usbdevice': (None, check_str),
    'usb': (None, check_int),
    'uuid': (None, check_str),
    'vcpus': (set_vcpus, check_int),
    'vfb': (None, check_list),
    'vhpt': (None, check_int),
    'videoram': (None, check_int),
    'vif2': (None, check_list),
    'vif_other_config': (set_vif_other_config, check_list),
    'vif': (set_vif, check_list),
    'viridian': (None, check_int),
    'vncconsole': (None, check_int),
    'vncdisplay': (None, check_int),
    'vnclisten': (None, check_str),
    'vncpasswd': (None, check_str),
    'vncunused': (None, check_int),
    'vnc': (None, check_int),
    'vpt_align': (None, check_int),
    'vscsi': (None, check_list),
    'vtpm': (None, check_list),
    'xauthority': (None, check_str),
    'xen_platform_pci': (None, check_int),
}


class Xen(Hypervisor):
    """
    Xen hypervisor management.
    """

    def get_services(self, version=None):
        return {
            'create_vm': self.create_vm,
            'delete_vm': self.delete_vm,
            'start_vm': self.start_vm,
            'stop_vm': self.stop_vm,
            'reboot_vm': self.reboot_vm,
            'pause_vm': self.pause_vm,
            'unpause_vm': self.unpause_vm,
            'save_vm': self.save_vm,
            'restore_vm': self.restore_vm,
            'migrate_vm': self.migrate_vm,
            'trigger_vm': self.trigger_vm,
            'sysrq_vm': self.sysrq_vm,
            'dump_vm_core': self.dump_vm_core,
            'list_vm_core': self.list_vm_core,
            'delete_vm_core': self.delete_vm_core,
            'configure_vm': self.configure_vm,
            'unconfigure_vm': self.unconfigure_vm,
            'get_vm_config': self.get_vm_config,
            'set_vm_config': self.set_vm_config,
            'list_vm': self.list_vm,
            'list_vms': self.list_vms,
            'send_to_guest': self.send_to_guest,
            'create_template': self.create_template,
            'delete_template': self.delete_template,
            'configure_template': self.configure_template,
            'unconfigure_template': self.unconfigure_template,
            'get_template_config': self.get_template_config,
            'set_template_config': self.set_template_config,
            'import_template': self.import_template,
            'delete_assembly': self.delete_assembly,
            'get_assembly_config': self.get_assembly_config,
            'get_assembly_config_xml': self.get_assembly_config_xml,
            'import_assembly': self.import_assembly,
            'deploy_assembly': self.deploy_assembly,
            'configure_vm_ha': self.configure_vm_ha,
        }

    def create_vm(self, repo_id, vm_id, params):
        lock = filelock.FileLock(get_lock_file(repo_id, vm_id))
        lock.acquire(wait=120)
        try:
            return create(get_vm_path(repo_id, vm_id), params)
        finally:
            lock.release()

    def delete_vm(self, repo_id, vm_id):
        lock = filelock.FileLock(get_lock_file(repo_id, vm_id))
        lock.acquire(wait=120)
        try:
            return delete(get_vm_path(repo_id, vm_id))
        finally:
            lock.release()

    def start_vm(self, repo_id, vm_id):
        path = get_vm_path(repo_id, vm_id)
        lock = filelock.FileLock('/var/run/ovs-agent/vm-start.lock')
        lock.acquire(wait=300)
        try:
            state_file = os.path.join(os.path.dirname(path), 'state')
            if os.path.exists(state_file):
                logging.debug('Starting saved VM will remove the state file.')
                os.unlink(state_file)
            run_cmd(['xm', 'create', path])
        finally:
            lock.release()

    def stop_vm(self, repo_id, vm_id, force=False):
        if repo_id:
            domain = get_name(get_vm_path(repo_id, vm_id))
        else:
            domain = vm_id
        if not force:
            run_cmd(['xm', 'shutdown', domain])
        else:
            (status, output) = commands.getstatusoutput('xm destroy %s' % domain)
            if status != 0:
                logging.info('Destroy VM failed: %s', output)

    def reboot_vm(self, repo_id, vm_id, wait=True):
        path = get_vm_path(repo_id, vm_id)
        # Leave at least 60 seconds between two reboots, otherwise xend will
        # destroy the VM to prevent loops.
        try:
            domid = get_domid(get_name(path))
            domain_path = GetDomainPath(domid)
            vm_path = xstransact.Read('%s/vm' % domain_path)
            start_time = xstransact.Read('%s/start_time' % vm_path)
            elapse = time.time() - float(start_time)
            if elapse < 60:
                time.sleep(60 - elapse + 1)
        except Exception, err:
            logging.warn('Error getting start time for VM %s: %s', domain, err)
        if wait:
            run_cmd(['xm', 'reboot', '-w', get_name(path)])
        else:
            run_cmd(['xm', 'reboot', get_name(path)])

    def pause_vm(self, repo_id, vm_id):
        path = get_vm_path(repo_id, vm_id)
        run_cmd(['xm', 'pause', get_name(path)])

    def unpause_vm(self, repo_id, vm_id):
        path = get_vm_path(repo_id, vm_id)
        run_cmd(['xm', 'unpause', get_name(path)])

    def save_vm(self, repo_id, vm_id, checkpoint=False):
        path = get_vm_path(repo_id, vm_id)
        state_file = os.path.join(os.path.dirname(path), 'state')
        if checkpoint:
            run_cmd(['xm', 'save', '--checkpoint', get_name(path), state_file])
        else:
            run_cmd(['xm', 'save', get_name(path), state_file])

    def restore_vm(self, repo_id, vm_id, paused=False):
        path = get_vm_path(repo_id, vm_id)
        state_file = os.path.join(os.path.dirname(path), 'state')
        if paused:
            run_cmd(['xm', 'restore', '--paused', state_file])
        else:
            run_cmd(['xm', 'restore', state_file])
        os.unlink(state_file)

    def migrate_vm(self, repo_id, vm_id, dest, live=True, ssl=True):
        path = get_vm_path(repo_id, vm_id)
        cmd = ['xm', 'migrate']
        if live:
            cmd += ['--live']
        if ssl:
            cmd += ['--ssl']
        cmd += [get_name(path), dest]
        run_cmd(cmd)

    def trigger_vm(self, repo_id, vm_id, name, vcpu=None):
        path = get_vm_path(repo_id, vm_id)
        cmd = ['xm', 'trigger', get_name(path), name]
        if vcpu:
            cmd += [vcpu]
        run_cmd(cmd)

    def sysrq_vm(self, repo_id, vm_id, letter):
        path = get_vm_path(repo_id, vm_id)
        run_cmd(['xm', 'sysrq', get_name(path), letter])

    def dump_vm_core(self, repo_id, vm_id, live=False, crash=False, reset=False):
        path = get_vm_path(repo_id, vm_id)
        now = time.gmtime()
        core_filename = 'core-%s' % time.strftime('%Y-%m%d-%H%M.%S', now)
        core_file = os.path.join(os.path.dirname(path), core_filename)
        cmd = ['xm', 'dump-core']
        if live:
            cmd += ['--live']
        if crash:
            cmd += ['--crash']
        if reset:
            cmd += ['--reset']
        cmd += [get_name(path), core_file]
        run_cmd(cmd)
        return xmlrpclib.DateTime(now)

    def list_vm_core(self, repo_id, vm_id):
        path = get_vm_path(repo_id, vm_id)
        result = []
        for filename in os.listdir(os.path.dirname(path)):
            match = re.match(r'^core-(?P<date>.+)$', filename)
            if match:
                date = time.strptime(match.group('date'), '%Y-%m%d-%H%M.%S')
                result.append(xmlrpclib.DateTime(date))
        return result

    def delete_vm_core(self, repo_id, vm_id, core_date):
        path = get_vm_path(repo_id, vm_id)
        date = time.strptime(core_date.value, '%Y%m%dT%H:%M:%S')
        core_filename = 'core-%s' % time.strftime('%Y-%m%d-%H%M.%S', date)
        core_file = os.path.join(os.path.dirname(path), core_filename)
        os.unlink(core_file)

    def configure_vm(self, repo_id, vm_id, params):
        lock = filelock.FileLock(get_lock_file(repo_id, vm_id))
        lock.acquire(wait=120)
        try:
            return configure(get_vm_path(repo_id, vm_id), params)
        finally:
            lock.release()

    def unconfigure_vm(self, repo_id, vm_id, params):
        lock = filelock.FileLock(get_lock_file(repo_id, vm_id))
        lock.acquire(wait=120)
        try:
            return unconfigure(get_vm_path(repo_id, vm_id), params)
        finally:
            lock.release()

    def get_vm_config(self, repo_id, vm_id):
        lock = filelock.FileLock(get_lock_file(repo_id, vm_id))
        lock.acquire(wait=120)
        try:
            return read_cfg(get_vm_path(repo_id, vm_id))
        finally:
            lock.release()

    def set_vm_config(self, repo_id, vm_id, params):
        check_cfg(params)
        lock = filelock.FileLock(get_lock_file(repo_id, vm_id))
        lock.acquire(wait=120)
        try:
            return write_cfg(get_vm_path(repo_id, vm_id), params)
        finally:
            lock.release()

    def list_vm(self, repo_id, vm_id):
        path = get_vm_path(repo_id, vm_id)
        lock = filelock.FileLock('/var/run/ovs-agent/vm-list.lock')
        lock.acquire(wait=120)
        try:
            info = run_cmd(['xm', 'list', '--long', get_name(path)])
            info_sxp = sxp.from_string(info)
            return parse_domain_info(info_sxp[1:])
        finally:
            lock.release()

    def list_vms(self):
        lock = filelock.FileLock('/var/run/ovs-agent/vm-list.lock')
        lock.acquire(wait=120)
        try:
            info = run_cmd(['xm', 'list', '--long'])
            info_sxp = sxp.all_from_string(info)
            result = []
            for dom_info_sxp in info_sxp:
                dom_info = parse_domain_info(dom_info_sxp[1:])
                result.append(dom_info)
            return result
        finally:
            lock.release()

    def send_to_guest(self, repo_id, vm_id, params):
        path = get_vm_path(repo_id, vm_id)
        domid = get_domid(get_name(path))
        dompath = GetDomainPath(domid)
        lock = filelock.FileLock(get_lock_file(repo_id, vm_id))
        lock.acquire(wait=120)
        try:
            if not xstransact.List(dompath):
                raise RuntimeError('VMAPI is not ready.')
            last_write_str = xstransact.Read(dompath, 'control/oracle-vmapi/to-guest/last-write')
            if not last_write_str:
                last_write_str = '0'
                xstransact.Write(dompath, 'control/oracle-vmapi/to-guest/last-write', last_write_str)
            last_write = int(last_write_str)
            for (key, value) in params.items():
                entries = xstransact.List('%s/control/oracle-vmapi/to-guest' % dompath)
                if (len(entries) - 1) > MAX_MESSAGE_NUMBER:
                    raise RuntimeError('No space to write message.')
                if len(key) > MAX_KEYNAME_LEN:
                    raise RuntimeError('Message key name too long (should be below or equal to %s): %s' % (MAX_KEYNAME_LEN, key))
                if len(value) > MAX_KEYVALUE_LEN:
                    raise RuntimeError('Message key value too long (should be below or equal to %s): %s' % (MAX_KEYVALUE_LEN, value))
                to_store = {}
                last_write += 1
                iteration = 0
                while len(value) > 0:
                    to_store['control/oracle-vmapi/to-guest/%s/%s' % (last_write, iteration)] =  value[:MAX_ENTRY_SIZE]
                    value = value[MAX_ENTRY_SIZE:]
                    iteration += 1
                to_store['control/oracle-vmapi/to-guest/%s' % last_write] = key
                to_store['control/oracle-vmapi/to-guest/last-write'] = str(last_write)
                xstransact.Write(dompath, to_store)
        finally:
            lock.release()

    def create_template(self, repo_id, template_id, params):
        lock = filelock.FileLock(get_lock_file(repo_id, template_id))
        lock.acquire(wait=120)
        try:
            return create(get_template_path(repo_id, template_id), params)
        finally:
            lock.release()

    def delete_template(self, repo_id, template_id):
        lock = filelock.FileLock(get_lock_file(repo_id, template_id))
        lock.acquire(wait=120)
        try:
            return delete(get_template_path(repo_id, template_id))
        finally:
            lock.release()

    def configure_template(self, repo_id, template_id, params):
        lock = filelock.FileLock(get_lock_file(repo_id, template_id))
        lock.acquire(wait=120)
        try:
            return configure(get_template_path(repo_id, template_id), params)
        finally:
            lock.release()

    def unconfigure_template(self, repo_id, template_id, params):
        lock = filelock.FileLock(get_lock_file(repo_id, template_id))
        lock.acquire(wait=120)
        try:
            return unconfigure(get_template_path(repo_id, template_id), params)
        finally:
            lock.release()

    def get_template_config(self, repo_id, template_id):
        lock = filelock.FileLock(get_lock_file(repo_id, template_id))
        lock.acquire(wait=120)
        try:
            return read_cfg(get_template_path(repo_id, template_id))
        finally:
            lock.release()

    def set_template_config(self, repo_id, template_id, params):
        check_cfg(params)
        lock = filelock.FileLock(get_lock_file(repo_id, template_id))
        lock.acquire(wait=120)
        try:
            return write_cfg(get_template_path(repo_id, template_id), params)
        finally:
            lock.release()

    def import_template(self, repo_id, template_id, url_list, option=None):
        set_nice()
        if not url_list:
            raise RuntimeError('No URL is provided.')
        template_path = get_template_path(repo_id, template_id)
        create(template_path, {})
        virtualdisk_deployed = {}
        iso_deployed = {}
        try:
            tmp_path = os.path.join(os.path.dirname(template_path), 'tmp')
            if not os.path.exists(tmp_path):
                os.makedirs(tmp_path)
            unpacked = []
            obj_total = len(url_list)
            obj_current = 1
            for url in url_list:
                tmp_name = make_filename(url)
                filename = os.path.join(tmp_path, tmp_name)
                filename = ovs_download_file(url, filename, option, 'template', obj_current, obj_total)
                unpack_files(tmp_path, unpacked, 'template', obj_current, obj_total)
                obj_current += 1
            vm_cfg_files = find_file('vm.cfg', tmp_path)
            if not vm_cfg_files:
                raise RuntimeError('No VM configure file found.')
            if len(vm_cfg_files) > 1:
                raise RuntimeError('Multiple VM configure files found.')
            vm_cfg = read_cfg(vm_cfg_files[0])
            fix_cfg(vm_cfg)
            check_cfg(vm_cfg)
            new_disk = []
            for disk in vm_cfg['disk']:
                old_disk_filepath = disk.split(',')[0].split(':')[-1]
                if old_disk_filepath == '':
                    # ',hdc:cdrom,r'
                    new_disk.append(disk)
                    continue
                old_disk_filename = os.path.basename(old_disk_filepath)
                old_disk_files = find_file(old_disk_filename, tmp_path)
                if not old_disk_files:
                    raise RuntimeError('Cannot find disk file: %s' % old_disk_filename)
                is_cdrom = disk.split(',')[1].endswith(':cdrom')
                if is_cdrom:
                    disk_id = '%s.iso' % get_uuid()
                    disk_path = get_iso_path(repo_id, disk_id)
                    shutil.move(old_disk_files[0], disk_path)
                    iso_deployed[disk_id] = old_disk_filename
                else:
                    disk_id = '%s.img' % get_uuid()
                    disk_path = get_virtualdisk_path(repo_id, disk_id)
                    shutil.move(old_disk_files[0], disk_path)
                    virtualdisk_deployed[disk_id] = old_disk_filename
                new_disk.append(disk.replace(old_disk_filepath, disk_path))
            vm_cfg['disk'] = new_disk
            vm_cfg['name'] = template_id
            vm_cfg['uuid'] = template_id
            write_cfg(template_path, vm_cfg)
            try:
                shutil.rmtree(tmp_path)
            except Exception, e:
                # FIXME: a temporary fix for NFS issue: orabug 12694949
                logging.warn('Error removing directory %s: %s', tmp_path, e)
            return {'VirtualDisk': virtualdisk_deployed,
                    'ISO': iso_deployed}
        except Exception, err:
            for virtualdisk_id in virtualdisk_deployed:
                try:
                    os.unlink(get_virtualdisk_path(repo_id, virtualdisk_id))
                except:
                    pass
            for iso_id in iso_deployed:
                try:
                    os.unlink(get_iso_path(repo_id, iso_id))
                except:
                    pass
            try:
                delete(template_path)
            except:
                pass
            raise RuntimeError('Template import error: %s' % err)

    def delete_assembly(self, repo_id, assembly_id):
        shutil.rmtree(get_assembly_path(repo_id, assembly_id))

    def get_assembly_config(self, repo_id, assembly_id):
        assembly_path = get_assembly_path(repo_id, assembly_id)
        unpacked_path = os.path.join(assembly_path, 'unpacked')
        ovf_filename = find_ovf_descriptor(unpacked_path)
        return ovfxen.parseOvfDescriptor(ovf_filename)

    def get_assembly_config_xml(self, repo_id, assembly_id):
        assembly_path = get_assembly_path(repo_id, assembly_id)
        unpacked_path = os.path.join(assembly_path, 'unpacked')
        ovf_filename = find_ovf_descriptor(unpacked_path)
        ovf_file = open(ovf_filename)
        ovf_cfg = ovf_file.read()
        ovf_file.close()
        return ovf_cfg

    def import_assembly(self, repo_id, assembly_id, url, option=None):
        set_nice()
        assembly_path = get_assembly_path(repo_id, assembly_id)
        unpacked_path = os.path.join(assembly_path, 'unpacked')
        imports_path = os.path.join(assembly_path, 'imports')
        if not os.path.exists(assembly_path):
            os.makedirs(assembly_path)

        try:
            if not os.path.exists(unpacked_path):
                os.makedirs(unpacked_path)
            if not os.path.exists(imports_path):
                os.makedirs(imports_path)
            basename = make_filename(url)
            if basename[-4:] not in ['.ovf', '.ova', '.rpm']:
                raise RuntimeError('.ovf, .ova or .rpm file is needed.')
            if basename.endswith('.rpm'):
                rpm_filename = os.path.join(assembly_path, basename)
                rpm_filename = ovs_download_file(url, rpm_filename, option, 'assembly')
                tmp_path = os.path.join(assembly_path, 'tmp')
                if not os.path.exists(tmp_path):
                    os.makedirs(tmp_path)
                unpack_file(rpm_filename, tmp_path, 'assembly')
                ovf_files = find_file('*.ovf', tmp_path)
                if ovf_files:
                    move_files(tmp_path, unpacked_path)
                else:
                    ova_files = find_file('*.ova', tmp_path)
                    if not ova_files:
                        raise RuntimeError('Cannot find OVF or OVA file.')
                    ova_filename = ova_files[0]
                    check_ova(ova_filename)
                    unpack_file(ova_filename, unpacked_path, 'assembly')
                shutil.rmtree(tmp_path)
            elif basename.endswith('.ova'):
                ova_filename = os.path.join(assembly_path, basename)
                ova_filename = ovs_download_file(url, ova_filename, option, 'assembly')
                check_ova(ova_filename)
                unpack_file(ova_filename, unpacked_path, 'assembly')
            else:
                ovf_filename = os.path.join(unpacked_path, basename)
                ovf_filename = ovs_download_file(url, ovf_filename, option)
                base_url = url[:-4] # remove trailing '.ovf'
                mf_url = base_url + '.mf'
                cert_url = base_url + '.cert'
                mf_filename = os.path.join(unpacked_path, make_filename(mf_url))
                cert_filename = os.path.join(unpacked_path, make_filename(cert_url))
                try:
                    ovs_download_file(mf_url, mf_filename, option)
                except Exception, err:
                    logging.warn('Error geting %s: %s', mf_url, err)
                try:
                    ovs_download_file(cert_url, cert_filename, option)
                except Exception, err:
                    logging.warn('Error geting %s: %s', cert_url, err)

            ovf_filename = find_ovf_descriptor(unpacked_path)
            ovf_cfg = ovfxen.parseOvfDescriptor(ovf_filename)
            file_numbers = len(ovf_cfg['File'])
            obj_total = len(ovf_cfg['File'])
            obj_current = 1
            for file_info in ovf_cfg['File']:

                def download_file(file_href):
                    # FIXME: should we download external resource to imports/ here? Yes for now.
                    if '://' in file_href:
                        dst_filename = os.path.join(imports_path, os.path.basename(file_href))
                    else:
                        dst_filename = os.path.join(unpacked_path, os.path.basename(file_href))
                        base_url = os.path.dirname(url)
                        file_href = os.path.join(base_url, file_href)
                    if not os.path.exists(dst_filename):
                        ovs_download_file(file_href, dst_filename, option, 'assembly_disk', obj_current, obj_total)

                if file_info['size'] and file_info['chunkSize']:
                    (chunk_number, remainder) = divmod(int(file_info['size']), int(file_info['chunkSize']))
                    if remainder:
                        chunk_number += 1
                    for chunk in range(chunk_number):
                        download_file('%s.%09d' % (file_info['href'], chunk))
                else:
                    download_file(file_info['href'])
                obj_current += 1

            ovfxen.validateAppliance(ovf_filename)
        except Exception, err:
            shutil.rmtree(assembly_path)
            raise RuntimeError('Assembly import error: %s' % err)

    def deploy_assembly(self, repo_id, assembly_id, to_deploy=None, target_repo_id=None):
        set_nice()
        if target_repo_id is None:
            target_repo_id = repo_id
        assembly_path = get_assembly_path(repo_id, assembly_id)
        unpacked_path = os.path.join(assembly_path, 'unpacked')
        ovf_filename = find_ovf_descriptor(unpacked_path)
        ovf_cfg =  ovfxen.parseOvfDescriptor(ovf_filename)
        if to_deploy is None:
            to_deploy = {}
            for vs_info in ovf_cfg['VirtualSystem']:
                to_deploy[vs_info['id']] = get_uuid()

        virtualdisk_to_deploy = {}
        iso_to_deploy = {}
        for (vs_id, template_id) in to_deploy.items():
            vs_info = None
            for vs in ovf_cfg['VirtualSystem']:
                if vs['id'] == vs_id:
                    vs_info = vs
                    break
            if not vs_info:
                raise RuntimeError('Cannot find virtual system with id: %s' % vs_id)
            for disk in vs_info['Disk']:
                (resource_type, resource_id) = ovfxen.parseHostResource(disk['HostResource'])
                if resource_type in ['file', 'disk']:
                    virtualdisk_to_deploy[resource_id] = ('%s.img' % get_uuid(), resource_type)
            for disk in vs_info['CD']:
                (resource_type, resource_id) = ovfxen.parseHostResource(disk['HostResource'])
                if resource_type in ['file', 'disk']:
                    iso_to_deploy[resource_id] = ('%s.iso' % get_uuid(), resource_type)

        try:
            virtualdisk_deployed = {}
            iso_deployed = {}
            template_deployed = {}
            for (resource_id, (virtualdisk_id, resource_type)) in virtualdisk_to_deploy.items():
                virtualdisk_path = get_virtualdisk_path(target_repo_id, virtualdisk_id)
                if resource_type == 'file':
                    ovfxen.deployFile(ovf_filename, resource_id, virtualdisk_path)
                elif resource_type == 'disk':
                    ovfxen.deployDisk(ovf_filename, resource_id, virtualdisk_path)
                virtualdisk_deployed[resource_id] = virtualdisk_id
            for (resource_id, (iso_id, resource_type)) in iso_to_deploy.items():
                iso_path = get_iso_path(target_repo_id, iso_id)
                if resource_type == 'file':
                    ovfxen.deployFile(ovf_filename, resource_id, iso_path)
                elif resource_type == 'disk':
                    ovfxen.deployDisk(ovf_filename, resource_id, iso_path)
                iso_deployed[resource_id] = iso_id
            for (vs_id, template_id) in to_deploy.items():
                vm_cfg = ovfxen.getDomainConfig(ovf_filename, vs_id)
                vm_cfg['name'] = template_id
                vm_cfg['uuid'] = template_id
                new_disk = []
                for disk in vm_cfg['disk']:
                    (backend, frontend, mode) = disk.split(',')
                    (resource_type, resource_id) = ovfxen.parseHostResource(backend)
                    is_cdrom = frontend.endswith(':cdrom')
                    if is_cdrom:
                        if resource_id:
                            iso_id = iso_deployed[resource_id]
                            iso_path = get_iso_path(target_repo_id, iso_id)
                            new_disk.append('file:%s,%s,%s' % (iso_path, frontend, mode))
                        else:
                            # ',hdc:cdrom,r'
                            new_disk.append(disk)
                    else:
                        if not resource_id:
                            raise RuntimeError('No backend is found for %s' % frontend)
                        virtualdisk_id = virtualdisk_deployed[resource_id]
                        virtualdisk_path = get_virtualdisk_path(target_repo_id, virtualdisk_id)
                        new_disk.append('file:%s,%s,%s' % (virtualdisk_path, frontend, mode))
                vm_cfg['disk'] = new_disk
                self.create_template(target_repo_id, template_id, vm_cfg)
                template_deployed[vs_id] = template_id
            # revert map: Template virtual disk/iso ID as key and original assembly disk ID as value
            return {'VirtualDisk': dict((v, k) for k, v in virtualdisk_deployed.iteritems()),
                    'ISO': dict((v, k) for k, v in iso_deployed.iteritems())}
        except Exception, err:
            for virtualdisk_id in virtualdisk_deployed.values():
                virtualdisk_path = get_virtualdisk_path(target_repo_id, virtualdisk_id)
                os.unlink(virtualdisk_path)
            for iso_id in iso_deployed.values():
                iso_path = get_iso_path(target_repo_id, iso_id)
                os.unlink(iso_path)
            for template_id in template_deployed.values():
                self.delete_template(target_repo_id, template_id)
            raise RuntimeError('Assembly deploy failed: %s' % err)

    def configure_vm_ha(self, repo_id, vm_id, enable_ha=True):
        path = get_vm_path(repo_id, vm_id)
        uuid = format_uuid(read_cfg(path)['uuid'])
        db_home = get_cluster_db_home()
        if enable_ha:
            d = {'repo_id': repo_id, 'vm_id': vm_id}
            write_item('monitored_vms', uuid, d, db_home)
        else:
            delete_item('monitored_vms', uuid, db_home)
