# Copyright (C) 2006-2012 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

from agent.base import Agent


class Network(Agent):

    """
    Class for network management.
    
    """

    def configure_virtual_ip(self, virtual_ip, base_ip):
        """
        Configure the virtual IP address of the server pool.

        @param virtual_ip: the virtual IP address of the server pool
        @type  virtual_ip: C{str}
        
        @param base_ip: the IP addresss that the manager uses to talk
        to this server; this indicates the network interface on to
        which the virtual IP address should be configured.
        @type  base_ip: C{str}

        @raise IModuleException: on failure

        """
        raise NotImplementedError()


    def deconfigure_virtual_ip(self, virtual_ip):
        """
        Deconfigure the virtual IP address of the server pool.

        @param virtual_ip: the virtual IP address of the server pool
        @type  virtual_ip: C{str}

        @raise IModuleException: on failure

        """
        raise NotImplementedError()


    def ovs_br_config(self, action, br_name, net_dev):
        """
        Configure a bridge connected to a network interface.

        @param action: either C{'start'} to create a bridge, or C{'stop'}
        to remove a bridge.
        @type  action: C{str}

        @param br_name: the name of bridge to act upon
        @type  br_name: C{str}
        
        @param net_dev: the network interface to add to the bridge, it
        should be either a single physical network interface or an
        aggregation of multiple physical network interfaces; it should
        not be a bridge, or a physical network interface which is part of
        an aggregation
        @type  net_dev: C{str}

        @return: the name of the bridge and the name of the interface
        it is connected to.
        @rtype: C{str}

        @raise IModuleException: on failure

        """
        raise NotImplementedError()


    def ovs_async_bridge(self, action, br_name, net_dev):
        """
        Configure a bridge connected to a network interface. This command
        can be used in asynchronous mode.

        @param action: either C{'start'} to create a bridge, or C{'stop'}
        to remove a bridge.
        @type  action: C{str}

        @param br_name: a name for the bridge
        @type  br_name: C{str}
        
        @param net_dev: the network interface to add to the bridge, it
        should be either a single physical network interface or an
        aggregation of multiple physical network interfaces; it should
        not be a bridge, or a physical network interface which is part of
        an aggregation
        @type  net_dev: C{str}

        @return: the name of the bridge and the name of the interface
        it is connected to.
        @rtype: C{str}

        @raise IModuleException: on failure

        """
        raise NotImplementedError()


    def ovs_vlan_bridge(self, action, br_name, net_dev, vlan_id):
        """
        Configure a bridge connected to a network interface and associate
        it with a VLAN.

        @param action: either C{'start'} to create a bridge and VLAN, or
        C{'stop'} to remove a bridge and VLAN.
        @type  action: C{str}

        @param br_name: a name for the bridge
        @type  br_name: C{str}
        
        @param net_dev: the network interface to add to the bridge and
        on which to create the VLAN, it should be either a single physical
        network interface or an aggregation of multiple physical network
        interfaces; it should not be a bridge, or a physical network
        interface which is part of an aggregation
        @type  net_dev: C{str}

        @param vlan_id: a VLAN ID (1-4095); VLAN ID of 1 is the
        untagged VLAN
        @type  vlan_id: C{str}

        @return: the name of the bridge, the name of the interface it is
        connected to, and the VLAN id.
        @rtype: C{str}

        @raise IModuleException: on failure

        """
        raise NotImplementedError()


    def ovs_bond_config(self, action, bond, *args):
        """
        Configure a bond, that is  a logical network interface for
        bonding or aggregating physical network interfaces.

        @param action: either C{'start'} to create a bond, or C{'stop'}
        to remove a bond.
        @type  action: C{str}

        @param bond: a name for the bond.
        @type  bond: C{str}

        @param args: this parameter is required only when creating a
        bond; this is an array containing the following parameters:
           - I{bondmode} (C{str}): the bonding or aggregation modes:
              - C{'1'} for active-backup load balancing
              - C{'4'} for 802.3ad link aggregation
              - C{'6'} for adaptative load balancing
           - I{primary} (C{str}): for active-backup load balancing, the
           primary physical network interface to use; this parameter is
           unused for other modes.
           - I{backup} (C{str}): for active-backup load balancing, the
           backup physical network interface; this parameter is unused
           for other modes.
        @type  args: C{array}
        
        @return: the name of the bond.
        @rtype: C{str}

        @raise IModuleException: on failure

        """
        raise NotImplementedError()


    def ovs_async_bond(self, action, bond, *args):
        """
        Configure a bond, that is  a logical network interface for
        bonding or aggregating physical network interfaces.

        @param action: either C{'start'} to create a bond, or C{'stop'}
        to remove a bond.
        @type  action: C{str}

        @param bond: a name for the bond.
        @type  bond: C{str}

        @param args: this parameter is required only when creating a
        bond; this is an array containing the following parameters:
           - I{bondmode} (C{str}): the bonding or aggregation modes:
              - C{'1'} for active-backup load balancing
              - C{'4'} for 802.3ad link aggregation
              - C{'6'} for adaptative load balancing
           - I{primary} (C{str}): for active-backup load balancing, the
           primary physical network interface to use; this parameter is
           unused for other modes.
           - I{backup} (C{str}): for active-backup load balancing, the
           backup physical network interface; this parameter is unused
           for other modes.
        @type  args: C{array}
        
        @return: the name of the bond.
        @rtype: C{str}

        @raise IModuleException: on failure

        """
        raise NotImplementedError()


    def ovs_bond_op(self, action, bond, backup):
        """
        Manage an existing bond.

        @param action: either C{'add'} to add, or C{'remove'} to remove
        a physical network interface to/from a bond.

        @param bond: name of the bond.
        @type  bond: C{str}

        @param backup: the physical interface to add or remove.
        @type  backup: C{str}

        @return: the name of the physical network interfaces and the
        name of the bond.
        @rtype: C{str}

        @raise IModuleException: on failure

        """
        raise NotImplementedError()


    def ovs_bond_mode(self, bond, mode):
        """
        Change bonding or aggregation mode of a bond.

        @param bond: name of the bond
        @type  bond: C{str}
        
        @param mode: the bonding or aggregation mode:
           - C{'1'} for active-backup load balancing
           - C{'4'} for 802.3ad link aggregation
           - C{'6'} for adaptative load balancing

        @return: the name of the bond and its new mode
        @rtype: C{str}

        @raise IModuleException: on failure

        """
        raise NotImplementedError()


    def ovs_local_config(self, action, br_name):
        """
        Configure a local bridge, that is a bridge which is not
        connected to a physical network interface.

        @param action: either C{'start'} to create a bridge, or C{'stop'}
        to remove a bridge.
        @type  action: C{str}

        @param br_name: the name of bridge to act upon
        @type  br_name: C{str}
        
        @return: the name of the bridge
        @rtype: C{str}

        @raise IModuleException: on failure

        """
        raise NotImplementedError()


    def ovs_vlan_config(self, action, interface, vlanid):
        """
        Configure a VLAN on a physical or a bond interface.

        @param action: either C{'start'} to create a VLAN, or C{'stop'}
        to remove a VLAN.

        @param interface: the interface on which to create a VLAN
        @type  interface: C{str}
        
        @param vlanid: the id of the VLAN (2-4095)
        @type  vlanid: C{str}

        @return: the name of the interface and the VLAN id.
        @rtype: C{str}

        @raise IModuleException: on failure

        """
        raise NotImplementedError()


    def ovs_ip_config(self, interface, optype, *args):
        """
        Configure the IP address and netmask of a physical, VLAN or
        bond network interface.

        @param interface: the network interface to configure.
        @type  interface: C{str}
        
        @param optype: the IP configuration to setup on the network
        interface, either:
            - C{'static'}: assigns the given IP and netmask to the
            interface
            - C{'dynamic'}: configures the interface with DHCP
            - C{'flush'}: unconfigures the interfaces: flushes the IP
            address and routes
        @type  optype: C{str}

        @param args: additional parameters, required with the C{'static'}
        option, otherwise this parameter is ignore. For the C{'static'}
        option, it should provide:
            - IP address (C{str}): an IPv4 address in decimal notation;
            for example C{'101.230.112'}
            - netmask (C{str}): a network netmask in a decimal notation;
            for example C{'255.255.255.0'}
        @type  args: C{list}

        @return: the name of the interface and addresses added or
        flushed.
        @rtype: C{str}

        @raise IModuleException: on failure

        """
        raise NotImplementedError()


    def ovs_if_meta(self, interface, data):
        """
        Associate meta-data with a network interface.

        This functions associate meta-data with a network interface,
        and store them persistently. Meta-data are used by the manager
        to identify networks that interfaces belong to. If an interface
        already has meta-data, then they are replaced with the new one.
        If an meta-data is provided then any existing meta-data is
        removed.

        @param interface: the network interface with which meta-data
        show be associated
        @type  interface: C{str}

        @param data: the meta-data
        @type  data: C{str}

        @return: the name of the interface and associated meta-data
        @rtype: C{str}

        @raise IModuleException: on failure

        """
        raise NotImplementedError()


    def ovs_change_mtu(self, interface, mtu):
        """
        Change the MTU of a network interface.

        The network interface can be a physical NIC, a vlan, a bond
        or a bridge. Changing a bond MTU will also change the MTU of
        its slaves. Changing the MTU of an interface that is part of
        a bridge will cause the MTU of the bridge and of all connected
        interfaces to change.

        @param interface: the network interface for which the MTU should
        be changed.
        @type  interface: C{str}
        
        @param mtu: the new MTU to use
        @type  mtu: C{str}

        @return: the name of the interface and the MTU
        @rtype: C{str}

        @raise IModuleException: on failure

        """
        raise NotImplementedError()


    def discover_network(self):
        """
        Discover the current network configuration.

        This includes the state of physical NICs, bonds, and bridges.
        Also return information stored for this server that is needed
        to configure the network when the manager is not available.

        @return: discovery data as an XML document
        @rtype: C{str}

        @raise IModuleException: on failure

        """
        raise NotImplementedError()
