# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import os
import re
import string
import shutil
import tempfile
import xml.dom
from ovf import OvfSet
from agent.common import run_cmd, old_run_cmd


def convertFile(src_filename, src_format, dst_filename, dst_format,
                src_compression=None, src_chunk_size=None, dst_size=None,
                dst_populated_size=None, dst_variant=None):
    # FIXME: add dst_size, dst_populated_size and parent support
    if src_chunk_size is not None:
        cmd = 'cat %s.*' % src_filename
    else:
        cmd = 'cat %s' % src_filename
    if src_compression:
        cmd += ' | gzip -dc'

    if src_format == 'vmdk':
        cmd += ' | vbox-img convert --stdin --srcformat %s --dstfilename %s --dstformat %s' \
               % (src_format, dst_filename, dst_format)
        if dst_variant:
            cmd += ' --variant %s' % dst_variant
        old_run_cmd(cmd)
    elif src_format in ['vdi', 'vhd']:
        (_, tmp_filename) = tempfile.mkstemp(dir=os.path.dirname(dst_filename))
        cmd += ' >%s' % tmp_filename
        old_run_cmd(cmd, (0, 2))
        cmd = 'vbox-img convert --srcfilename %s --srcformat %s --dstfilename %s --dstformat %s' \
              % (tmp_filename, src_format, dst_filename, dst_format)
        try:
            old_run_cmd(cmd)
        finally:
            os.unlink(tmp_filename)
    elif src_format == 'raw':
        # FIXME: when input is in raw format, only support output to raw.
        cmd += ' | dd of=%s bs=1M' % dst_filename
        old_run_cmd(cmd)
    else:
        raise RuntimeError('Unsupported source file format: %s' % src_format)


def createFile(filename, fmt, size, populated_size=None, sparse=True):
    # FIXME: add format, populated_size and parent support
    dirpath = os.path.dirname(filename)
    if not os.path.exists(dirpath):
        os.makedirs(dirpath)
    fd = None
    try:
        try:
            fd = os.open(filename, os.O_WRONLY | os.O_CREAT | os.O_DSYNC)
            if sparse:
                os.ftruncate(fd, size)
            else:
                bufsize = 1024 * 1024
                for i in range(0, size / bufsize):
                    os.write(fd, '\x00' * bufsize)
                os.write(fd, '\x00' * size % bufsize)
        except OSError, err:
            raise RuntimeError('Error creating file %s: %s' % (filename, err))
    finally:
        if fd is not None:
            os.close(fd)


def blkname(i):
    if i < 26:
        return string.ascii_lowercase[i]
    return blkname(i/26 - 1) + string.ascii_lowercase[i%26]


def getStrings(ovf):
    # FIXME: cannot handle external resource bundles.
    result = []
    for rootchild in ovf.envelope.childNodes:
        if rootchild.localName == 'Strings':
            for node in rootchild.getElementsByTagNameNS('*', 'Msg'):
                msg_info = {'lang': rootchild.getAttribute('xml:lang'),
                            'msgid': node.getAttribute('ovf:msgid'),
                            'msg': node.firstChild.data}
                result.append(msg_info)
    return result


def getFile(ovf):
    result = []
    for item in ovf.files:
        file_info = {'id': item.file_id,
                     'href': item.href,
                     'size': item.size,
                     'chunkSize': item.chunksize,
                     'compression': item.compression}
        result.append(file_info)
    return result


def getDisk(ovf):
    result = []
    for node in ovf.document.getElementsByTagNameNS('*', 'Disk'):
        disk_info = {'diskId': node.getAttribute('ovf:diskId'),
                     'fileRef': node.getAttribute('ovf:fileRef'),
                     'capacity': node.getAttribute('ovf:capacity'),
                     'capacityAllocationUnits': node.getAttribute('ovf:capacityAllocationUnits'),
                     'format': node.getAttribute('ovf:format'),
                     'populatedSize': node.getAttribute('ovf:populatedSize'),
                     'parentRef': node.getAttribute('ovf:parentRef')}
        result.append(disk_info)
    return result


def getNetwork(ovf):
    result = []
    for node in ovf.document.getElementsByTagNameNS('*', 'Network'):
        network_info = {'id': node.getAttribute('ovf:id'),
                        'name': node.getAttribute('ovf:name')}
        for child in node.childNodes:
            if child.localName == 'Description' and child.firstChild:
                network_info['Description'] = child.firstChild.data
        result.append(network_info)
    return result


def getVirtualSystemCollection(ovf):
    result = []
    for node in ovf.document.getElementsByTagNameNS('*', 'VirtualSystemCollection'):
        vsc_info = {'id': node.getAttribute('ovf:id'),
                    'Info': node.getElementsByTagNameNS('*', 'Info')[0].firstChild.data}
        for child in node.childNodes:
            if child.localName == 'Name':
                vsc_info['Name'] = child.nodeValue
        result.append(vsc_info)
    return result


def getAllocationUnits(units):
    """DSP0004 v2.5.0 outlines the Programmatic Unit forms for OVF. This
    pertains specifically to rasd:AllocationUnits, which accepts both the
    current and deprecated forms. New implementations should not use Unit
    Qualifiers as this form is deprecated.
      - PUnit form, as in "byte * 2^20"
      - PUnit form w/ Units Qualifier(deprecated), as in "MegaBytes"
    """
    if units.startswith('byte') or units.startswith('bit'):
        # FIXME: support hertz * 10^6 ?
        units = units.replace('^', '**')
        units = units.replace('byte', '2 ** 0')
        units = units.replace('bit', '2 ** -3')
        factor = int(eval(units))
    else:
        if units.startswith('Kilo'):
            factor = 2 ** 10
        elif units.startswith('Mega'):
            factor = 2 ** 20
        elif units.startswith('Giga'):
            factor = 2 ** 30
        else:
            raise ValueError("Incompatible PUnit quantifier.")
        if units.endswith('Bytes'):
            factor *= 1
        elif units.endswith('Bits'):
            factor *= 2 ** -3
        else:
            raise ValueError("Incompatible PUnit quantifier.")
    return factor


def getRasd(node):
    result = []
    for child in node.childNodes:
        child_dict = {}
        if child.nodeType != xml.dom.Node.ELEMENT_NODE:
            continue
        if child.localName != 'Item':
            continue
        for index in range(child.attributes.length):
            attr_name = child.attributes.item(index).name
            attr_value = child.attributes.item(index).value
            child_dict[attr_name] = attr_value
        for grandchild in child.childNodes:
            if grandchild.nodeType == xml.dom.Node.ELEMENT_NODE and grandchild.firstChild:
                child_dict[grandchild.tagName] = grandchild.firstChild.data
        result.append(child_dict)
    return result


def getVirtualSystemMemory(node):
    for resource in getRasd(node):
        if resource.get('rasd:ResourceType') == '4':
            return int(resource['rasd:VirtualQuantity']) * \
                   getAllocationUnits(resource['rasd:AllocationUnits']) / 1024 / 1024


def getVirtualSystemProcessor(node):
    for resource in getRasd(node):
        if resource.get('rasd:ResourceType') == '3':
            return int(resource['rasd:VirtualQuantity'])


def getVirtualSystemBusType(node, instance_id):
    bus = 'ide'
    if instance_id:
        for resource in getRasd(node):
            if resource['rasd:InstanceID'] == instance_id and \
               resource['rasd:ResourceType'] in ['6', '7', '8', '9', '20']:
                bus = 'scsi'
    return bus


def getVirtualSystemDisk(node):
    result = []
    for resource in getRasd(node):
        disk_info = {}
        if not resource.get('rasd:ResourceType') in ['17', '19']:
            continue
        disk_info['HostResource'] = resource.get('rasd:HostResource')
        disk_info['bus'] = getVirtualSystemBusType(node, resource.get('rasd:Parent'))
        result.append(disk_info)
    return result


def getVirtualSystemCD(node):
    result = []
    for resource in getRasd(node):
        cd_info = {}
        if not resource.get('rasd:ResourceType') in ['15', '16']:
            continue
        cd_info['HostResource'] = resource.get('rasd:HostResource')
        cd_info['bus'] = getVirtualSystemBusType(node, resource.get('rasd:Parent'))
        result.append(cd_info)
    return result


def getVirtualSystemFloppy(node):
    result = []
    for resource in getRasd(node):
        floppy_info = {}
        if not resource.get('rasd:ResourceType') == '14':
            continue
        floppy_info['HostResource'] = resource.get('rasd:HostResource')
        floppy_info['bus'] = getVirtualSystemBusType(node, resource.get('rasd:Parent'))
        result.append(floppy_info)
    return result


def getVirtualSystemEthernet(node):
    result = []
    for resource in getRasd(node):
        # FIXME: what info can be extracted from OVF?
        ethernet_info = {}
        if resource['rasd:ResourceType'] == '10':
            result.append(ethernet_info)
    return result


def getVirtualSystemType(node):
    vstypes = node.getElementsByTagName('vssd:VirtualSystemType')
    if vstypes:
        return vstypes[0].firstChild.data


def getVirtualSystemInfo(node):
    vs_info = {'VirtualSystemCollection': '',
               'id': node.getAttribute('ovf:id'),
               'Info': node.getElementsByTagNameNS('*', 'Info')[0].firstChild.data}
    vs_info['VirtualSystemType'] = getVirtualSystemType(node)
    for child in node.childNodes:
        if child.localName == 'Name':
            vs_info['Name'] = child.nodeValue
        if child.localName == 'VirtualHardwareSection':
            vs_info['Memory'] = getVirtualSystemMemory(child)
            vs_info['Processor'] = getVirtualSystemProcessor(child)
            vs_info['Disk'] = getVirtualSystemDisk(child)
            vs_info['CD'] = getVirtualSystemCD(child)
            vs_info['Floppy'] = getVirtualSystemFloppy(child)
            vs_info['Ethernet'] = getVirtualSystemEthernet(child)
    return vs_info


def getVirtualSystem(ovf):
    result = []
    for rootchild in ovf.envelope.childNodes:
        if rootchild.localName == 'VirtualSystem':
            vs_info = getVirtualSystemInfo(rootchild)
            result.append(vs_info)
    for vsc in ovf.document.getElementsByTagNameNS('*', 'VirtualSystemCollection'):
        for vs in vsc.childNodes:
            if vs.localName == 'VirtualSystem':
                vs_info = getVirtualSystemInfo(vs)
                vs_info['VirtualSystemCollection'] = vsc.getAttribute('ovf:id')
                result.append(vs_info)
    return result


def parseOvfDescriptor(filename):
    result = {}
    ovfset = OvfSet.OvfSet(filename)
    ovf = ovfset.ovfFile
    result['Name'] = ovfset.name
    result['Strings'] = getStrings(ovf)
    result['File'] = getFile(ovf)
    result['Disk'] = getDisk(ovf)
    result['Network'] = getNetwork(ovf)
    result['VirtualSystemCollection'] = getVirtualSystemCollection(ovf)
    result['VirtualSystem'] = getVirtualSystem(ovf)
    return result


def isPvGuest(vstype):
    return vstype in ['DMTF:Oracle:X86:OracleVM_Xen:PVM_Linux', 'DMTF:Oracle:OracleVM:PVM', 'xenpv']


def getDomainConfig(filename, vs_id):
    result = {}
    ovf_cfg = parseOvfDescriptor(filename)
    vs_info = None
    for vs in ovf_cfg['VirtualSystem']:
        if vs['id'] == vs_id:
            vs_info = vs
            break
    if not vs_info:
        raise RuntimeError('Cannot find virtual system with id: %s' % vs_id)

    result['name'] = str(vs_info['id'])
    result['memory'] = int(vs_info['Memory'])
    if vs_info['Processor']:
        result['vcpus'] = vs_info['Processor']

    # FIXME: we convert all disks to 'xvd'.
    result['disk'] = []
    xen_disk_count = 0
    for disk_info in vs_info['Disk']:
        if disk_info['HostResource']:
            backend = disk_info['HostResource']
        else:
            raise RuntimeError('HostResource is required for disk: %s' % disk_info)
        frontend = 'xvd%s' % blkname(xen_disk_count)
        xen_disk_count += 1
        mode = 'w'
        result['disk'].append(str('%s,%s,%s' % (backend, frontend, mode)))

    for disk_info in vs_info['CD']:
        if disk_info['HostResource']:
            backend = disk_info['HostResource']
        elif not isPvGuest(vs_info['VirtualSystemType']):
            backend = ''
        else:
            raise RuntimeError('HostResource is required for disk: %s' % disk_info)
        frontend = 'xvd%s' % blkname(xen_disk_count)
        xen_disk_count += 1
        frontend += ':cdrom'
        mode = 'r'
        result['disk'].append(str('%s,%s,%s' % (backend, frontend, mode)))

    result['vif'] = []
    for netif in vs_info['Ethernet']:
        result['vif'].append('')

    if isPvGuest(vs_info['VirtualSystemType']):
        result['builder'] = 'linux'
        result['bootloader'] = '/usr/bin/pygrub'
        result['vfb'] = ['type=vnc']
    else:
        result['builder'] = 'hvm'
        result['serial'] = 'pty'
        result['vnc'] = 1

    return result


def validateAppliance(filename):
    run_cmd(['ova', '--validate', '-f', filename])


def unpackAppliance(filename, dst_dir):
    if not os.path.exists(dst_dir):
        os.makedirs(dst_dir)
    run_cmd(['tar', 'xf', filename, '-C', dst_dir])


def parseHostResource(resource):
    if resource is None:
        return (None, None)
    m = re.match(r'^(ovf:)?/(file|disk)/(.+)$', resource)
    if m:
        return (m.group(2), m.group(3))
    else:
        return (None, None)


def deployFile(ovf_filename, file_id, dst_filename):
    ovfset = OvfSet.OvfSet(ovf_filename)
    ovf = ovfset.ovfFile
    files = getFile(ovf)
    file_info = None
    for f in files:
        if f['id'] == file_id:
            file_info = f
            break
    if not file_info:
        raise RuntimeError('Cannot find file with id: %s' % file_id)

    unpacked_path = os.path.dirname(ovf_filename)
    imports_path = os.path.join(unpacked_path, '../imports')
    if '://' in file_info['href']:
        src_filename = os.path.join(imports_path, os.path.basename(file_info['href']))
    else:
        src_filename = os.path.join(unpacked_path, os.path.basename(file_info['href']))

    shutil.copyfile(src_filename, dst_filename)


def deployDisk(ovf_filename, disk_id, dst_filename):
    ovfset = OvfSet.OvfSet(ovf_filename)
    ovf = ovfset.ovfFile
    files = getFile(ovf)
    disks = getDisk(ovf)
    disk_info = None
    for disk in disks:
        if disk['diskId'] == disk_id:
            disk_info = disk
            break
    if not disk_info:
        raise RuntimeError('Cannot find disk with id: %s' % disk_id)

    # FIXME: how to convert ovf:capacityAllocationUnits="byte * 2^20" to integer?
    dst_size = int(disk_info['capacity'])
    if disk_info['capacityAllocationUnits']:
        dst_size *= getAllocationUnits(disk_info['capacityAllocationUnits'])

    # FIXME: how to get disk format? No definiton on OVF spec. Always use 'raw' for now.
    dst_format = 'raw'

    if not disk_info['fileRef']:
        createFile(dst_filename, dst_format, dst_size, disk_info['populatedSize'])
    else:
        file_info = None
        for item in files:
            if item['id'] == disk_info['fileRef']:
                file_info = item
                break
        if not file_info:
            raise RuntimeError('Disk %s refrenced file %s is not exist' % (disk_id, disk_info['fileRef']))
        unpacked_path = os.path.dirname(ovf_filename)
        imports_path = os.path.join(unpacked_path, '../imports')
        if '://' in file_info['href']:
            src_filename = os.path.join(imports_path, os.path.basename(file_info['href']))
        else:
            src_filename = os.path.join(unpacked_path, os.path.basename(file_info['href']))

        if file_info['href'].lower().endswith('vmdk'):
            src_format = 'vmdk'
        elif file_info['href'].lower().endswith('vdi'):
            src_format = 'vdi'
        elif file_info['href'].lower().endswith('vhd'):
            src_format = 'vhd'
        else:
            src_format = 'raw'
        convertFile(src_filename, src_format, dst_filename, dst_format, file_info['compression'],
                    file_info['chunkSize'], dst_size, disk_info['populatedSize'])
