# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import os
import re
import sha
import crypt
import shutil
import string
import random
import tempfile
from agent.constants import CONFIG_DIR
from agent.utils import filelock


def random_salt():
    valid_chars = string.ascii_letters + string.digits + './'
    return '$6$%s$' % ''.join(random.SystemRandom().sample(valid_chars, 16))


def encrypt_password(password, encryption_method='CRYPT', salt=None):
    if encryption_method == 'SHA':
        return sha.new(password).digest().encode("base64").replace("\n", "")
    elif encryption_method == 'CRYPT':
        if salt is None:
            salt = random_salt()
        return crypt.crypt(password, salt)
    else:
        raise Exception('Unknown encryption method.')


def read_password():
    lock = filelock.FileLock('/var/run/ovs-agent/passwdfile.lock')
    lock.acquire(wait=120)
    try:
        pwd_dict = {}
        pwd_file = open(os.path.join(CONFIG_DIR, 'passwdfile'))
        try:
            for line in pwd_file:
                pwd_match = re.match(r'^(.+?):{(.+?)}(.+)$', line)
                if not pwd_match:
                    continue
                user, method, enc_pwd = pwd_match.groups()
                pwd_dict[user] = (method, enc_pwd)
            return pwd_dict
        finally:
            pwd_file.close()
    finally:
        lock.release()


def write_password(pwd_dict):
    lock = filelock.FileLock('/var/run/ovs-agent/passwdfile.lock')
    lock.acquire(wait=120)
    try:
        tmpfd, tmpfn = tempfile.mkstemp(dir=CONFIG_DIR)
        try:
            for user, (method, pwd) in pwd_dict.items():
                os.write(tmpfd, "%s:{%s}%s\n" % (user, method, pwd))
            os.fsync(tmpfd)
        finally:
            os.close(tmpfd)
        shutil.move(tmpfn, os.path.join(CONFIG_DIR, 'passwdfile'))
    finally:
        lock.release()


def update_password(username, password):
    password_dict = read_password()
    if username not in password_dict:
        raise Exception('User %s does not exists.' % username)
    password_dict[username] = ('CRYPT', encrypt_password(password))
    write_password(password_dict)


def add_user(username, encrypted_password='!'):
    password_dict = read_password()
    if username in password_dict:
        raise Exception('User %s already exists.' % username)
    password_dict[username] = ('CRYPT', encrypted_password)
    write_password(password_dict)


def delete_user(username):
    password_dict = read_password()
    try:
        del password_dict[username]
    except KeyError:
        raise Exception("User %s doesn't exist." % username)
    write_password(password_dict)
