# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import os
import string
import logging
import shutil
from os.path import join, basename, exists, dirname
from agent.constants import (CLUSTER_STATE_DLM_READY, LOCAL_SERVER_DB,
                             LUN_PREFIX, MEMBERSHIP_STATE_CLUSTERED, NFS_MOUNT,
                             POOLFS_DMNFS_DEVICE, POOLFS_IMG, POOLFS_IMG_SIZE,
                             POOLFS_LABEL, POOLFS_META, POOLFS_MOUNT,
                             POOLFS_NODE_SLOTS, POOLFS_VERSION)
from agent.poolfs.base import PoolFilesystem
from agent.common import run_cmd, retry_cmd
from agent.ocfs2 import ocfs2_file_lookup, ocfs2_file_read, ocfs2_file_write
from agent.db import dump_db
from agent.stats import reload_monitor
from agent.discover import SimpleDiscoveryXML
from agent.serverpool import get_membership_state, get_cluster_state
from agent.utils import filelock


class PoolOCFS2(PoolFilesystem):
    """
    Class for managing the shared metadata in a pool using bdb on a small clustered ocfs2 fs.
    """

    def get_services(self, version=None):
        return {
            'discover_pool_filesystem': self.discover_pool_filesystem,
            'create_pool_filesystem': self.create_pool_filesystem,
            'destroy_pool_filesystem': self.destroy_pool_filesystem,
            'ocfs2_get_meta_data': self.ocfs2_get_meta_data,
        }

    def discover_pool_filesystem(self):
        """
        Discover information about the pool filesystem.
        Returns an empty document if no pool filesystem found.

        @param      None
        @return     Discovery data as an XML document.
                    Raises an exception on failure.
        """
        lock = filelock.FileLock('/var/run/ovs-agent/discover_pool_filesystem.lock')
        lock.acquire(wait=120)
        try:
            doc = SimpleDiscoveryXML("Discover_Pool_Filesystem_Result")
            add_nfs_poolfs_info(doc)
            add_lun_poolfs_info(doc)
            return doc.finish()
        finally:
            lock.release()


    def create_pool_filesystem(self, poolfs_type, poolfs_target, cluster_name,
            poolfs_uuid, poolfs_nfsbase_uuid, manager_uuid, pool_uuid,
            block_size="4K", cluster_size="4K", journal_size="4M"):
        """
        Create a pool filesystem.

        @param      poolfs_type - "nfs" or "lun"
        @param      poolfs_target -
                    physical location of the pool filesystem
        @param      cluster_name - cluster to use the pool filesystem
        @param      poolfs_uuid -
                    a unique identifier for the pool filesystem
        @param      poolfs_nfsbase_uuid -
                    a unique identifier for the base NFS filesystem
        @param      manager_uuid - a unique identifier for the manager
        @param      pool_uuid - a unique identifier for the server pool
        @param      block_size - mkfs.ocfs2 option
        @param      cluster_size - mkfs.ocfs2 option
        @param      journal_size - mkfs.ocfs2 option

        @return     None. Raises an exception on failure.
        """
        if poolfs_type == "nfs":
            nfs_create_pool_filesystem(poolfs_target, cluster_name, poolfs_uuid,
                    poolfs_nfsbase_uuid, manager_uuid, pool_uuid)
        elif poolfs_type == "lun":
            lun_create_pool_filesystem(poolfs_target, cluster_name, poolfs_uuid,
                    manager_uuid, pool_uuid)
        else:
            raise ValueError("Invalid pool filesystem type: %s" % poolfs_type)


    def destroy_pool_filesystem(self, poolfs_type, poolfs_target, poolfs_uuid, poolfs_nfsbase_uuid):
        """
        Destroy a pool filesystem.

        @param      poolfs_type - "nfs" or "lun"
        @param      poolfs_target -
                    physical location of the pool filesystem
        @param      poolfs_uuid -
                    a unique identifier for the pool filesystem
        @param      poolfs_nfsbase_uuid -
                    a unique identifier for the base NFS filesystem

        @return     None. Raises an exception on failure.
        """
        if poolfs_type == "nfs":
            nfs_destroy_pool_filesystem(poolfs_target, poolfs_nfsbase_uuid)
        elif poolfs_type == "lun":
            lun_destroy_pool_filesystem(poolfs_target)
        else:
            raise ValueError("Invalid pool filesystem type: %s" % poolfs_type)

    def ocfs2_get_meta_data(self, device, filename):
        """
        @return     Meta data as a dict.
                    Raises an exception on failure.
        """
        # flush the cache
        run_cmd(['sync'])
        if ocfs2_file_lookup(device, filename) == 0:
            contents = ocfs2_file_read(device, filename)
            return parse_metafile(contents)
        else:
            return {}


def nfs_create_pool_filesystem(poolfs_target, cluster_name, poolfs_uuid,
        poolfs_nfsbase_uuid, manager_uuid, pool_uuid):
    # mount NFS base
    mount_point = join(NFS_MOUNT, poolfs_nfsbase_uuid)
    mount_device(poolfs_target, mount_point)
    # make pool filesystem
    poolfs_img = join(mount_point, POOLFS_IMG)
    create_poolfs_image(poolfs_img)
    make_pool_filesystem(poolfs_img, poolfs_uuid, cluster_name)
    # generate poolfs metafile contents
    poolfs_meta = join(mount_point, POOLFS_META)
    contents = """OVS_POOLFS_UUID=%s
OVS_POOLFS_NFSBASE_UUID=%s
OVS_POOLFS_MGR_UUID=%s
OVS_POOLFS_VERSION=%s
OVS_POOLFS_POOL_UUID=%s
""" % (poolfs_uuid,
poolfs_nfsbase_uuid,
manager_uuid,
POOLFS_VERSION,
pool_uuid,
)
    try:
        # write poolfs metafile
        f = file(poolfs_meta, "w")
        try:
            f.write(contents)
        finally:
            f.close()
        # umount poolfs nfsbase
        umount_device(poolfs_target)
    except:
        if exists(poolfs_meta):
            os.remove(poolfs_meta)


def lun_create_pool_filesystem(poolfs_target, cluster_name, poolfs_uuid,
        manager_uuid, pool_uuid):
    # destroy pool filesystem
    lun_destroy_pool_filesystem(poolfs_target)
    # make pool filesystem
    make_pool_filesystem(poolfs_target, poolfs_uuid, cluster_name)
    # generate poolfs metafile contents
    contents = """OVS_POOLFS_UUID=%s
OVS_POOLFS_MGR_UUID=%s
OVS_POOLFS_VERSION=%s
OVS_POOLFS_POOL_UUID=%s
OVS_POOLFS_LUN_UUID=%s
""" % (poolfs_uuid,
manager_uuid,
POOLFS_VERSION,
pool_uuid,
basename(poolfs_target),
)
    try:
        # write poolfs metafile
        ocfs2_file_write(poolfs_target, POOLFS_META, contents)
    except:
        lun_destroy_pool_filesystem(poolfs_target)


def nfs_destroy_pool_filesystem(poolfs_target, poolfs_nfsbase_uuid):
    mount_point = join(NFS_MOUNT, poolfs_nfsbase_uuid)
    mount_device(poolfs_target, mount_point)
    poolfs_meta = join(mount_point, POOLFS_META)
    if exists(poolfs_meta):
        os.remove(poolfs_meta)
    umount_device(poolfs_target)


def lun_destroy_pool_filesystem(poolfs_target):
    cmd = [ "dd", "if=/dev/zero", "of=%s" % poolfs_target,
            "bs=1M", "count=4" ]
    run_cmd(cmd)


def add_nfs_poolfs_info(doc):
    """
    Add XML for NFS based pool filesystem.
    """
    output = run_cmd(["cat", "/proc/mounts"])
    lines = output.splitlines()

    for line in lines:
        parts = line.split()
        if parts[1].startswith("/nfsmnt/") and parts[2] == "nfs":
            poolfs_meta = join(parts[1], POOLFS_META)
            if exists(poolfs_meta):
                f = file(poolfs_meta)
                try:
                    contents = f.read()
                finally:
                    f.close()
                d = parse_metafile(contents)

                doc.push("Pool_Filesystem")
                doc.add("Pool_Filesystem_Type", "nfs")
                doc.add("Pool_Filesystem_Target", parts[0])
                doc.add("Pool_Filesystem_Uuid", d.get("OVS_POOLFS_UUID"))
                doc.add("Pool_Filesystem_Nfsbase_Uuid",
                        d.get("OVS_POOLFS_NFSBASE_UUID"))
                doc.add("Pool_Filesystem_Manager_Uuid",
                        d.get("OVS_POOLFS_MGR_UUID"))
                doc.add("Pool_Filesystem_Version",
                        d.get("OVS_POOLFS_VERSION"))
                doc.add("Pool_Filesystem_Pool_Uuid",
                        d.get("OVS_POOLFS_POOL_UUID"))
                doc.pop()


def add_lun_poolfs_info(doc):
    """
    Add XML for LUN based pool filesystem.
    """
    output = run_cmd(["mounted.ocfs2", "-d"])
    lines = output.splitlines()

    for line in lines:
        parts = line.split()
        if parts[0].startswith(LUN_PREFIX) and parts[-1] == POOLFS_LABEL:
            try:
                if ocfs2_file_lookup(parts[0], POOLFS_META) != 0:
                    continue
                contents = ocfs2_file_read(parts[0], POOLFS_META)
                d = parse_metafile(contents)

                doc.push("Pool_Filesystem")
                doc.add("Pool_Filesystem_Type", "lun")
                doc.add("Pool_Filesystem_Target", parts[0])
                doc.add("Pool_Filesystem_Uuid", d.get("OVS_POOLFS_UUID"))
                doc.add("Pool_Filesystem_Lun_Uuid",
                        d.get("OVS_POOLFS_LUN_UUID"))
                doc.add("Pool_Filesystem_Manager_Uuid",
                        d.get("OVS_POOLFS_MGR_UUID"))
                doc.add("Pool_Filesystem_Version",
                        d.get("OVS_POOLFS_VERSION"))
                doc.add("Pool_Filesystem_Pool_Uuid",
                        d.get("OVS_POOLFS_POOL_UUID"))
                doc.pop()
            except Exception, e:
                logging.warn('Cannot get poolfs info: %s', e)


def rename_file(src, dst):
    if exists(dst):
        os.remove(dst)
    if exists(src):
        shutil.move(src, dst)


def create_poolfs_image(poolfs_img):
    """
    Create a pool filesystem image.

    @param      poolfs_img - pool filesystem image

    @return     None. Raises an exception on failure.
    """
    if exists(poolfs_img):
        os.remove(poolfs_img)

    stat = os.statvfs(dirname(poolfs_img))
    free_space = stat.f_bsize * stat.f_bfree / 1024 / 1024
    if free_space < POOLFS_IMG_SIZE + 1:
        raise Exception("Not enough space for pool filesystem.")

    cmd = [ "dd", "if=/dev/zero", "of=%s" % poolfs_img,
            "bs=1M", "seek=%s" % POOLFS_IMG_SIZE, "count=0" ]
    run_cmd(cmd)


def make_pool_filesystem(device, poolfs_uuid, cluster_name,
        block_size="4K", cluster_size="4K", journal_size="4M"):
    """
    Make an ocfs2 filesystem for global heartbeat.

    @param      device - target device
    @param      poolfs_uuid - pool filesystem uuid
    @param      cluster_name - cluster name
    @param      block_size - block size
    @param      cluster_size - cluster size
    @param      journal size - journal size

    @return     None. Raises an exception on failure.
    """
    cmd = [ "mkfs.ocfs2",
            "-b", block_size,
            "-C", cluster_size,
            "-J", "size=%s" % journal_size,
            "-N", POOLFS_NODE_SLOTS,
            "-L", POOLFS_LABEL,
            "-U", poolfs_uuid,
            "--global-heartbeat",
            "--cluster-stack=o2cb",
            "--cluster-name=%s" % cluster_name,
            "--force",
            device ]
    run_cmd(cmd)


def create_dmnfs_device(poolfs_img, device_name):
    """
    Create a logical device for the pool filesystem image.

    @param      poolfs_img - pool filesystem image
    @device     device_name - device to create

    @return     None. Raises an exception on failure.
    """
    cmd = [ "dmsetup", "table", "--target", "nfs" ]
    output = run_cmd(cmd)
    logging.debug("dmsetup output: %s" % output)

    # dmnfs0: 0 65536 nfs /test/x.img 0
    for ln in output.splitlines():
        if len(ln.split(":")) == 2:
            device, table = ln.split(":")
            if device == device_name:
                if table and table.split()[3] == poolfs_img:
                    # already created
                    return
                else:
                    raise Exception("Device busy: %s" % ln)

    # 512 bytes sectors
    sectors = POOLFS_IMG_SIZE * 1024 * 2
    cmd = [ "dmsetup", "create", device_name,
            "--table", "0 %s nfs %s 0" % (sectors, poolfs_img) ]
    run_cmd(cmd)


def remove_dmnfs_device(device_name):
    """
    Remove a dmnfs device.

    @param      device_name - device to remove

    @return     None. Raise an exception on failure.
    """
    cmd = [ "dmsetup", "table", "--target", "nfs" ]
    output = run_cmd(cmd)
    logging.debug("dmsetup output: %s" % output)

    # dmnfs0: 0 65536 nfs /test/x.img 0
    for ln in output.splitlines():
        device, table = ln.split(":")
        if device == device_name:
            cmd = [ "dmsetup", "remove", device_name ]
            run_cmd(cmd)


def mount_device(device, mount_point):
    """
    Mount a device.

    @param      device - device to mount
    @param      mount_point - mount point

    @return     None. Raise an exception on failure.
    """
    logging.debug("Trying to mount %s to %s", device, mount_point)

    if not exists(mount_point):
        os.makedirs(mount_point)

    device = device.strip().rstrip("/")
    mount_point = mount_point.strip().rstrip("/")
    mounted = False

    for ln in open("/proc/mounts", "r"):
        dev, mnt = map(string.rstrip, ln.split()[0:2], ["/", "/"])
        if device == dev or mount_point == mnt:
            if device == dev and mount_point == mnt:
                mounted = True
                break
            else:
                raise Exception("%s already mounted on %s" % (dev, mnt))

    if mounted:
        cmd = [ "mount", "-o", "remount", device ]
        retry_cmd(cmd)
        logging.debug("%s re-mounted to %s", device, mount_point)
    else:
        cmd = [ "mount", device, mount_point ]
        retry_cmd(cmd)
        logging.debug("%s mounted to %s", device, mount_point)


def umount_device(device):
    """
    Umount a device.

    @param      device - device to umount

    @return     None. Raise an exception on failure.
    """
    logging.debug("Trying to umount %s", device)

    device = device.strip().rstrip("/")
    mount_point = None

    for ln in open("/proc/mounts", "r"):
        dev, mnt = map(string.rstrip, ln.split()[0:2], ["/", "/"])
        if device == dev or device == mnt:
            mount_point = mnt
            break

    if mount_point:
        # force unmount, in case of an unreachable NFS system
        cmd = [ "umount", "-f", device ]
        try:
            retry_cmd(cmd)
            try:
                # remove mount point if empty
                os.rmdir(mount_point)
            except Exception, e:
                logging.warn("Error removing mount point: %s")
            logging.debug("%s umounted", device)
        except Exception, e:
            # list all processes accessing files on the filesystem
            cmd = [ "lsof", mount_point ]
            try:
                output = run_cmd(cmd).strip()
                logging.error("Device or resource busy: \n%s", output)
            except:
                # unexpected dmnfs device?
                cmd = [ "dmsetup", "table", "--target", "nfs" ]
                try:
                    output = run_cmd(cmd)
                    for line in output.splitlines():
                        if mount_point in line:
                            logging.error("Unexpected dmnfs device: %s", line)
                except:
                    logging.exception("Error getting dmnfs devices")
            # raise umount error
            raise e
    else:
        logging.debug("%s not mounted, skip umounting", device)


def parse_metafile(contents):
    """
    Parse metafile contents to a dict.

    @param      contents - metafile contents as a string

    @return     Metafile contents as a dict.
    """
    d = {}
    for line in contents.splitlines():
        # deal with comments
        ln = line.split("#")[0].strip()
        # parse key/value pairs
        if ln == "":
            continue
        elif "=" in ln and len(ln.split("=")) == 2:
            key, value = map(string.strip, ln.split("="))
            d[key] = value
        else:
            raise Exception("Invalid line in metafile: %s" % line)
    return d


def init_pool_filesystem():

    from agent.cluster.o2cb import server_cluster_online, server_cluster_offline

    d = dump_db(LOCAL_SERVER_DB)
    if get_membership_state() == MEMBERSHIP_STATE_CLUSTERED \
        and get_cluster_state() != CLUSTER_STATE_DLM_READY:
        poolfs_type = d["poolfs_type"]
        poolfs_nfsbase_uuid = d["poolfs_nfsbase_uuid"]
        poolfs_target = d["poolfs_target"]
        poolfs_uuid = d["poolfs_uuid"]
        poolfs_mount_point = os.path.join(POOLFS_MOUNT, poolfs_uuid)

        if poolfs_type == "nfs":
            nfsbase_mount_point = os.path.join(NFS_MOUNT, poolfs_nfsbase_uuid)
            poolfs_img = os.path.join(nfsbase_mount_point, POOLFS_IMG)
            poolfs_device = os.path.join("/dev/mapper/", POOLFS_DMNFS_DEVICE)
        else:
            poolfs_device = poolfs_target

        try:
            if poolfs_type == "nfs":
                mount_device(poolfs_target, nfsbase_mount_point)
                create_dmnfs_device(poolfs_img, POOLFS_DMNFS_DEVICE)

            server_cluster_online()
            mount_device(poolfs_device, poolfs_mount_point)

        except Exception, e:
            try:
                # cleanup broken symlink
                reload_monitor()
                server_cluster_offline()
                if poolfs_type == "nfs":
                    remove_dmnfs_device(POOLFS_DMNFS_DEVICE)
                    umount_device(poolfs_target)
            finally:
                raise e
