# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import os
import re
import logging
import time
from os.path import join, basename, isdir, exists
from agent.constants import *
from agent.utils import filelock
from agent.config import get_config
from agent.common import old_run_cmd, host_alive
from agent.db import read_item, write_item, dump_db, get_cluster_db_home
from agent.notification import notify_manager
from agent.serverpool import get_cluster_state, is_clustered


def dlm_join_domain(domain_name="ovm"):
    _dir = join(DLM, basename(domain_name))
    if not isdir(_dir):
        os.mkdir(_dir)
    # else: already joined


def dlm_acquire_lock(lock):
    dlm_join_domain(DLM_DOMAIN)

    # limit dlm lock name to 31 chars
    lock_path = join(DLM_DOMAIN, re.sub(r"[^a-zA-Z0-9]+", "", lock)[-31:])

    try:
        if not exists(lock_path):
            fd = os.open(lock_path, os.O_CREAT | os.O_NONBLOCK)
            os.close(fd)
        # create the EX lock
        fd = os.open(lock_path, os.O_RDWR | os.O_NONBLOCK)
        # close it and hold it
        os.close(fd)
    except Exception, e:
        # to get rid of the side effect of touch
        if exists(lock_path):
            os.remove(lock_path)
        raise e


def dlm_release_lock(lock):
    # limit dlm lock name to 31 chars
    lock_path = join(DLM_DOMAIN, re.sub(r"[^a-zA-Z0-9]+", "", lock)[-31:])
    if exists(lock_path):
        os.remove(lock_path)


def acquire_master_dlm_lock():

    ETXTBUSY = 26

    if get_cluster_state() != CLUSTER_STATE_DLM_READY:
        raise Exception("Pool filesystem not ready.")

    dlm_join_domain(DLM_DOMAIN)

    # close() does not downconvert the lock level nor does it drop the lock. The
    # holder still owns the lock at that level. All close() does is allows any
    # downconvert request to succeed.
    # However, a downconvert request is only generated for queued requests. And
    # O_NONBLOCK is specifically a noqueue dlm request.

    # 1) O_CREAT | O_NONBLOCK will create a lock file if it does not exist, whether
    #    we are the lock holder or not.
    # 2) if we hold O_RDWR lock, and we close but not delete it, we still hold it.
    #    afterward, O_RDWR will succeed, but O_RDWR | O_NONBLOCK will not.
    # 3) but if we donnot hold the lock, O_RDWR will hang there waiting,
    #    which is not desirable -- any uninterruptable hang is undesirable.
    # 4) if noboday else hold the lock either, but the lock file exists as side effect
    #    of 1), with O_NONBLOCK, it may result in ETXTBUSY

    # a) we need O_NONBLOCK to avoid 3)
    # b) we need to delete it ourself to avoid 2) (for e.g. agtctl restart), but when?
    #   *) if not succeed in 1), remove the lock file (side product), avoid 4)
    #   *) if everything is good, we drop it and we clean out.
    #   *) if stopped with agtctl (killed), agtctl should remove the lock.
    #   *) if crashed, agtctl stop --force should clean out the env.

    try:
        if not exists(DLM_LOCK_MASTER):
            fd = os.open(DLM_LOCK_MASTER, os.O_CREAT | os.O_NONBLOCK)
            os.close(fd)

        master_lock = os.open(DLM_LOCK_MASTER, os.O_RDWR | os.O_NONBLOCK)

        # close it and hold it
        os.close(master_lock)
        return True

    except OSError, e:
        if e.errno == ETXTBUSY:
            # if we donnot remove the lock file (the side effect of touch),
            # we may never succeed to hold the exclusive lock, i.e. 4)
            if exists(DLM_LOCK_MASTER):
                os.remove(DLM_LOCK_MASTER)
        else:
            raise e


def release_master_dlm_lock():
    if exists(DLM_LOCK_MASTER):
        os.remove(DLM_LOCK_MASTER)


def _get_vip_dev(vip):
    """Get dev the vip is bound to."""
    cmd = "ip addr show to %s | awk '/scope global/ {print $NF;}'" % vip
    output = old_run_cmd(cmd)
    dev = output.strip()
    return dev


def ifconfig_vip_up(vip, ip):
    """Bind vip to dev that ip is assigned to."""
    dev = _get_vip_dev(vip)
    if dev: # already bound
        return
    cmd = "ip addr show to %s" % ip
    output = old_run_cmd(cmd)
    lines = output.strip().splitlines()
    if len(lines) < 2:
        raise Exception("IP %s is not up." % ip)
    parts = lines[-1].split()
    dev = parts[-1]
    netmask = parts[1].split("/")[-1]

    dev_ip = dev
    if ":" in dev:
        dev_ip = dev.split(":")[0]

    cmd = "ifconfig | awk '/Link encap/ {print $1}'"
    output = old_run_cmd(cmd)
    dev_list = output.splitlines()

    idx = 0
    while True:
        dev_vip = "%s:%s" % (dev_ip, idx)
        if dev_vip not in dev_list:
            break
        else:
            idx += 1

    cmd = "ifconfig %s %s/%s" % (dev_vip, vip, netmask)
    old_run_cmd(cmd)

    #refresh switch arp table
    #http://wiki.wireshark.org/Gratuitous_ARP
    cmd = "arping -q -U -c 2 -I %s %s" % (dev_ip, vip)
    old_run_cmd(cmd)
    cmd = "arping -q -A -c 2 -I %s %s" % (dev_ip, vip)
    old_run_cmd(cmd)


def ifconfig_vip_down(vip):
    """Release vip."""
    dev = _get_vip_dev(vip)
    if dev:
        cmd = "ifconfig %s down" % dev
        old_run_cmd(cmd)


def is_holding_vip(vip):
    """Check whether the current server is holding the vip."""
    dev = _get_vip_dev(vip)
    return dev and True or False


def is_master():
    """Check whether the current server is the server pool master."""
    my_hostname = read_item(LOCAL_SERVER_DB, "registered_hostname")
    if not my_hostname:
        return False
    master_hostname = read_item("server_pool", "pool_master_hostname", get_cluster_db_home())
    return my_hostname == master_hostname


def setup_master_env():
    """Setup master environment:
        1) setup master vip.
        2) store master info to master db.
    """
    d = dump_db(LOCAL_SERVER_DB)
    db_home = get_cluster_db_home()

    vip = read_item("server_pool", "pool_virtual_ip", db_home)
    ip = d.get("registered_ip")
    ifconfig_vip_up(vip, ip)

    my_hostname = d.get("registered_hostname")
    write_item("server_pool", "pool_master_hostname", my_hostname, db_home)


def teardown_master_env():
    """Teardown master environment:
        1) release the master_vip
        2) release master dlm lock
    """
    vip = read_item("server_pool", "pool_virtual_ip", get_cluster_db_home())
    if vip:
        ifconfig_vip_down(vip)

    release_master_dlm_lock()


def send_remaster_event():
    hostname = read_item(LOCAL_SERVER_DB, "registered_hostname")
    msg = "{CLUSTER} {REMASTER} master is [%s]" % hostname
    try:
        notify_manager(msg)
    except Exception, e:
        logging.debug("Error sending notification: %s", e)


take_crown = True
def cluster_remaster():
    global take_crown
    db_home = get_cluster_db_home()
    pool_vip = read_item("server_pool", "pool_virtual_ip", db_home)
    if not pool_vip:
        logging.debug("VIP not configured")
        return
    if acquire_master_dlm_lock():
        if host_alive(pool_vip):
            if is_holding_vip(pool_vip):
                if is_master():
                    write_item(LOCAL_SERVER_DB, "is_master", True)
                    if take_crown:
                        take_crown = False
                        logging.info("It is master, VIP=%s" % pool_vip)
                        send_remaster_event()
                else:
                    take_crown = True
                    ifconfig_vip_down(pool_vip)
                    write_item(LOCAL_SERVER_DB, "is_master", False)
                    release_master_dlm_lock()
                    logging.debug("It is not master: VIP released")
            else:
                take_crown = True
                write_item(LOCAL_SERVER_DB, "is_master", False)
                release_master_dlm_lock()
                logging.debug("VIP in use: master lock released")
        else:
            take_crown = True
            try:
                setup_master_env()
                logging.debug("Master server elected: setup done")
                write_item(LOCAL_SERVER_DB, "is_master", True)
                send_remaster_event()
            finally:
                if not is_master():
                    release_master_dlm_lock()
    else:
        take_crown = True
        write_item(LOCAL_SERVER_DB, "is_master", False)


def remaster():
    lock = filelock.FileLock(CLUSTER_LOCK)
    lock.acquire(wait=60)
    try:
        d = dump_db(LOCAL_SERVER_DB)
        roles = d.get("roles")
        if roles and "master" in roles:
            if is_clustered():
                cluster_remaster()
            else:
                vip = d.get("pool_virtual_ip")
                ip = d.get("registered_ip")
                master_flag = d.get("is_master")
                if master_flag and vip and not is_holding_vip(vip):
                    if host_alive(vip):
                        logging.debug("VIP in use: conflict detected")
                    else:
                        ifconfig_vip_up(vip, ip)
                        logging.info("It is master, VIP=%s" % vip)
    finally:
        lock.release()


def teardown_master_server():
    if is_clustered():
        teardown_master_env()
    else:
        vip = read_item(LOCAL_SERVER_DB, "pool_virtual_ip")
        if vip:
            ifconfig_vip_down(vip)


def serve_forever():
    logging.info("REMASTER SERVER STARTED")
    config = get_config()
    remaster_interval = int(config.get("remaster", "interval"))
    while True:
        try:
            time.sleep(remaster_interval)
            remaster()
        except Exception, e:
            logging.exception("Error in remaster process: %s", e)
            time.sleep(SLEEP_ON_ERR)


if __name__ == '__main__':
    serve_forever()
