# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

from agent.remote.base import RemoteManagement
from agent.common import run_cmd, retry_cmd


class LinuxRemoteManagement(RemoteManagement):
    """
    Class for remote server management.
    """

    def get_services(self, version=None):
        return {
            'sys_shutdown': self.sys_shutdown,
            'sys_reboot': self.sys_reboot,
            'remote_power_off': self.remote_power_off,
            'remote_power_on': self.remote_power_on,
        }

    def sys_shutdown(self):
        """
        Shutdown the server.

        @raise IModuleException: on failure

        """
        run_cmd(["shutdown", "-P", "-h", "now", "shutdown requested by OVM"])

    def sys_reboot(self):
        """
        Reboot the server.

        @raise IModuleException: on failure

        """
        run_cmd(["shutdown", "-r", "now", "reboot requested by OVM"])

    def remote_power_off(self, controller_type, tgt_host, bmc_username="",
                         bmc_password=""):
        """
        Shutdown using ipmitool. For whatever reasons it's been decided
        that this can also shutdown another host so we need to take a
        hostname/ip as a paramater.

        @param controller_type: the mechanism to use for executing the
        power off request: C{'IPMI'} for IPMI
        @type  controller_type: C{str}
        
        @param tgt_host: for IPMI, an IP address where to send the power
        off request
        @type  tgt_host: C{str}
        
        @param bmc_username: a username for controlling the target host
        @type  bmc_username: C{str}
        
        @param bmc_password: a password for controlling the target host
        @type  bmc_password: C{str}

        @raise IModuleException: on failure

        """
        if controller_type != "IPMI":
            raise Exception("Controller type not supported: %s" %
                            controller_type)

        cmd = ["ipmitool", "-I", "lan", "-H", tgt_host]
        if bmc_username:
            cmd.extend(["-U", bmc_username])
        if bmc_password:
            cmd.extend(["-P", bmc_password])
        cmd.extend(["power", "off"])

        # It transpires that Dell BMC's are somewhat unreliable
        # (actually it may not just be exclusive to them but the result is
        # that we need to send this command a number of times to make sure it
        # was recieved)
        # leave in a small interval for the card to wake up and process the
        # command
        try:
            retry_cmd(cmd, interval=5, count=5)
        except:
            raise Exception("Remote power off failed: %s, %s" %
                    (controller_type, tgt_host))

    def remote_power_on(self, controller_type, tgt, arg1="", arg2=""):
        """
        Attempt to wakup machine over the lan using the ipmitool utility
        basically this is the same as wol/ether-wake etc. The number of
        arguments depend on the controller type, for example:
          - 4 args: IPMI, 1.2.3.4, fred, flintstone
          - 3 args: WOL, 11:22:33:44:55:66, eth1

        @param controller_type: the mechanism to use for executing the
        power on request: C{'IPMI'} for IPMI, C{'WOL'}
        for Wake-on-LAN (WOL)
        @type  controller_type: C{str}

        @param tgt: for IPMI, an IP address where to send the power
        on request; for WOL, a MAC address for the system to power on
        @type  tgt: C{str}

        @param arg1: for IPMI, a username for controlling the target host;
        for WOL, the network device to use for sending the WOL request
        @type  arg1: C{str}

        @param arg2: for IPMI, a password for controlling the target host;
        for WOL, this parameter is not used
        @type  arg2: C{str}
        
        @raise IModuleException: on failure

        """
        if controller_type not in ["IPMI", "WOL"]:
            raise Exception("Controller type not supported: %s" %
                            controller_type)

        # So why not call this directly from the manager side?
        # It turns out the licencing is somehow imcompatable
        # for bundling with the manager.

        if controller_type == "WOL":
            # Not necessary to do this, it's just for readability
            net_device = arg1
            sys_wol(tgt, net_device)
        else:
            # For readability...
            bmc_username = arg1
            bmc_password = arg2
            cmd = ["ipmitool", "-I", "lan", "-H", tgt]
            if bmc_username:
                cmd.extend(["-U", bmc_username])
            if bmc_password:
                cmd.extend(["-P", bmc_password])
            cmd.extend(["power", "on"])

            # It transpires that Dell BMC's are somewhat unreliable
            # (actually it may not just be exclusive to them but the result is
            # that we need to send this command a number of times to make sure
            # it was received)
            try:
                retry_cmd(cmd, interval=5, count=5)
            except:
                raise Exception("Remote power on failed: %s, %s" %
                        (controller_type, tgt))


def sys_wol(tgt_mac, net_device="eth0"):
    """
    Attempt to wake the host with ethernet packet(with optional message)
    This uses ether-wake so the target machine *MUST* have the network
    card in etherwake mode

    This is typically achieved with "ethtool -s ethX wol g" on the target host
    before shutdown. Also be aware that if standby power to the machine is
    lost, then this setting will usually be lost and the machine will not be
    able to be revived with this method. A few motherboards do allow the wol
    settings to be preserved after a powerloss. Basically this is a rather
    hit or miss affair.

    @param tgt_mac: a MAC address for the system to power on (MAC format:
    aa:bb:cc:dd:ee:ff)
    @type  tgt_mac: C{str}

    @param net_device: the network device to use for sending the WOL request
    (default: eth0)
    @type  net_device: C{str}

    """
    # It should just work with one packet but just for good measure
    # we'll send it out a few more times to be sure
    cmd = ["ether-wake", "-i", net_device, tgt_mac]
    retry_cmd(cmd, interval=1, count=5)
