# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import os
import codecs
import logging
from os.path import join, exists, islink, isfile, isdir
from agent.constants import LOCAL_REPO_DB, OVS_REPO_VERSION, OVS_REPO_HEAD, OVS_REPO_META
from agent.base import Agent
from agent.utils import filelock
from agent.common import run_cmd, old_run_cmd
from agent.db import read_item, write_item, dump_db, delete_item
from agent.host.linux import ovs_download_file
from agent.discover import SimpleDiscoveryXML
from agent.serverpool import get_owning_manager_uuid


class Repository(Agent):
    '''
    Class for repository management.
    '''

    def get_services(self, version=None):
        return {
            'discover_repositories': discover_repositories,
            'mount_repository_fs': mount_repository_fs,
            'unmount_repository_fs': unmount_repository_fs,
            'create_repository': create_repository,
            'add_repository': add_repository,
            'delete_repository': delete_repository,
            'get_repository_meta_data': get_repository_meta_data,
            'discover_repository_db': discover_repository_db,
            'edit_repository_db': edit_repository_db,
            'import_ISO': import_ISO,
            'import_virtual_disk': import_virtual_disk,
        }


def get_repository_meta_file(repo_mount_point):
    return join(OVS_REPO_HEAD, repo_mount_point, OVS_REPO_META)


def get_repository_meta_data(repo_mount_point):
    repo_meta_file = get_repository_meta_file(repo_mount_point)
    f = codecs.open(repo_meta_file, encoding='utf-8')
    data = {}
    for line in f:
        line = line.strip()
        if not line:
            continue
        if line.startswith('#'):
            continue
        k, v = line.split('=', 1)
        data[k] = v
    f.close()
    for entry in ['OVS_REPO_UUID', 'OVS_REPO_MGR_UUID', 'OVS_REPO_VERSION', 'OVS_REPO_ALIAS']:
        if entry not in data:
            raise Exception('Invalid repository %s meta: %s is missing', repo_mount_point, entry)
    return data


def set_repository_meta_data(repo_mount_point, meta):
    repo_meta_file = get_repository_meta_file(repo_mount_point)
    f = codecs.open(repo_meta_file, 'w', encoding='utf-8')
    for (key, value) in meta.items():
        f.write('%s=%s\n' % (key, value))
    f.close()
    # flush the cache
    run_cmd(['sync'])


def discover_meta(doc, mount_point):
    meta = get_repository_meta_data(mount_point)
    doc.add('Version', meta['OVS_REPO_VERSION'])
    doc.add('Manager_UUID', meta['OVS_REPO_MGR_UUID'])
    doc.add('Repository_UUID', meta['OVS_REPO_UUID'])
    doc.add('Repository_Alias', meta['OVS_REPO_ALIAS'])


def discover_dir(doc, mount_point, directory, entry):
    doc.push(directory)
    for name in os.listdir(join(mount_point, directory)):
        if name.startswith('.') or isfile(name):
            continue
        doc.push(entry, 'Name', name)
        subdir = join(mount_point, directory, name)
        for subname in os.listdir(subdir):
            if subname.startswith('.') or isdir(subname):
                continue
            doc.add('File', subname)
        doc.pop()
    doc.pop()


def discover_file(doc, mount_point, directory, entry):
    doc.push(directory)
    for name in os.listdir(join(mount_point, directory)):
        if name.startswith('.') or isdir(name):
            continue
        doc.add(entry, name)
    doc.pop()


def discover_assemblies(doc, mount_point):
    discover_dir(doc, mount_point, 'Assemblies', 'Assembly')


def discover_templates(doc, mount_point):
    discover_dir(doc, mount_point, 'Templates', 'Template')


def discover_virtual_machines(doc, mount_point):
    discover_dir(doc, mount_point, 'VirtualMachines', 'VirtualMachine')


def discover_virtual_disks(doc, mount_point):
    discover_file(doc, mount_point, 'VirtualDisks', 'Disk')


def discover_isos(doc, mount_point):
    discover_file(doc, mount_point, 'ISOs', 'ISO')


def discover_repositories(args):
    '''Discover information about the repositories and associated info.

    @param args: space separated repository mount points.
    @return     Returns the discovery data as an XML document.
                Raises an exception on failure.
    '''
    lock = filelock.FileLock('/var/run/ovs-agent/discover_repositories.lock')
    lock.acquire(wait=120)
    try:
        mountpoints = args.split()
        repo_mountpoints = os.listdir(OVS_REPO_HEAD)
        selected_mountpoints = list(set(mountpoints).intersection(set(repo_mountpoints)))
        doc = SimpleDiscoveryXML('Discover_Repositories_Result')
        doc.push('RepositoryList')
        for entry in selected_mountpoints:
            mount_point = join(OVS_REPO_HEAD, entry)
            if not os.path.ismount(mount_point):
                continue
            doc.push('Repository', 'Name', entry)
            discover_meta(doc, mount_point)
            discover_assemblies(doc, mount_point)
            discover_templates(doc, mount_point)
            discover_virtual_machines(doc, mount_point)
            discover_virtual_disks(doc, mount_point)
            discover_isos(doc, mount_point)
            doc.pop()
        doc.pop()
        return doc.finish()
    finally:
        lock.release()


def get_fs_type(fs_location):
    if fs_location.startswith('/dev/'):
        return 'ocfs2'
    else:
        return 'nfs'


def mount_repository_fs(fs_location, mount_point):
    repo_mount_point = join(OVS_REPO_HEAD, mount_point)
    if os.path.ismount(repo_mount_point):
        return
    if not exists(repo_mount_point):
        os.makedirs(repo_mount_point)
    filesystem = get_fs_type(fs_location)
    try:
        run_cmd(['mount', '-t', filesystem, fs_location, repo_mount_point])
    except Exception, err:
        try:
            os.rmdir(repo_mount_point)
        except Exception, e:
            logging.warn('Error remove repository mount point: %s', repo_mount_point)
        raise err


def unmount_repository_fs(mount_point):
    repo_mount_point = join(OVS_REPO_HEAD, mount_point)
    if not os.path.ismount(repo_mount_point):
        raise Exception('Repository file system not mounted')
    run_cmd(['umount', repo_mount_point])
    os.rmdir(repo_mount_point)


def create_repository(fs_location, mount_point, repo_uuid, repo_alias):
    if repo_uuid in dump_db(LOCAL_REPO_DB):
        raise Exception('Repository already exists')
    repo_mount_point = join(OVS_REPO_HEAD, mount_point)
    if not os.path.ismount(repo_mount_point):
        raise Exception('Repository not mounted')
    repo_meta_file = get_repository_meta_file(repo_mount_point)
    if exists(repo_meta_file):
        raise Exception('File system contains repository data')
    manager_uuid = get_owning_manager_uuid()
    subdirs = ['Assemblies', 'Templates', 'ISOs', 'VirtualDisks', 'VirtualMachines']
    try:
        for entry in subdirs:
            path = join(repo_mount_point, entry)
            if islink(path):
                os.remove(path)
            if not exists(path):
                os.mkdir(path)
    except Exception, e:
        for entry in subdirs:
            try:
                os.rmdir(entry)
            except:
                pass
        raise e
    meta = {'OVS_REPO_UUID': repo_uuid,
            'OVS_REPO_MGR_UUID': manager_uuid,
            'OVS_REPO_VERSION': OVS_REPO_VERSION,
            'OVS_REPO_ALIAS': repo_alias}
    set_repository_meta_data(repo_mount_point, meta)
    item = {'fs_location': fs_location,
            'mount_point': repo_mount_point,
            'alias': repo_alias,
            'version': OVS_REPO_VERSION,
            'manager_uuid': manager_uuid,
            'filesystem': get_fs_type(fs_location)}
    write_item(LOCAL_REPO_DB, repo_uuid, item)


def add_repository(fs_location, mount_point):
    repo_mount_point = join(OVS_REPO_HEAD, mount_point)
    if not os.path.ismount(repo_mount_point):
        raise Exception('Repository not mounted')
    meta = get_repository_meta_data(repo_mount_point)
    if meta['OVS_REPO_MGR_UUID'] not in ['NONE', get_owning_manager_uuid()]:
        raise Exception('Repository not owned by server manager.')
    if meta['OVS_REPO_VERSION'] != OVS_REPO_VERSION:
        raise Exception('Repository version is not correct.')
    item = {'fs_location': fs_location,
            'mount_point': repo_mount_point,
            'alias': meta['OVS_REPO_ALIAS'],
            'version': meta['OVS_REPO_VERSION'],
            'manager_uuid': meta['OVS_REPO_MGR_UUID'],
            'filesystem': get_fs_type(fs_location)}
    write_item(LOCAL_REPO_DB, str(meta['OVS_REPO_UUID']), item)


def delete_repository(repo_uuid, erase=False):
    if repo_uuid in dump_db(LOCAL_REPO_DB):
        if erase:
            repo_mount_point = read_item(LOCAL_REPO_DB, repo_uuid)['mount_point']
            if not os.path.ismount(repo_mount_point):
                raise Exception('Repository not mounted')
            meta = get_repository_meta_data(repo_mount_point)
            if meta['OVS_REPO_MGR_UUID'] not in ['NONE', get_owning_manager_uuid()]:
                raise Exception('Repository not owned by server manager')
            old_run_cmd('rm -f %s' % get_repository_meta_file(repo_mount_point))
            old_run_cmd('rm -fr %s/*' % repo_mount_point)
        delete_item(LOCAL_REPO_DB, repo_uuid)


def refresh_repository(repo_uuid, repo):
    try:
        if not os.path.ismount(repo['mount_point']):
            try:
                mount_repository_fs(repo['fs_location'], repo['mount_point'])
            except Exception, e:
                return 'Unmounted'
        subdirs = set(os.listdir(repo['mount_point']))
        repo_items =  set(['Assemblies', 'Templates', 'ISOs', 'VirtualDisks', 'VirtualMachines', '.ovsrepo'])
        if not subdirs & repo_items:
            unmount_repository_fs(repo['mount_point'])
            delete_item(LOCAL_REPO_DB, repo_uuid)
            return 'Empty'
        meta = get_repository_meta_data(repo['mount_point'])
        if meta['OVS_REPO_VERSION'] != OVS_REPO_VERSION:
            return 'Invalid'
        return 'Mounted'
    except Exception, err:
        logging.debug('Error getting repository %s status: %s', repo['mount_point'], err)
        return 'Invalid'


def discover_repository_db():
    lock = filelock.FileLock('/var/run/ovs-agent/discover_repository_db.lock')
    lock.acquire(wait=120)
    try:
        doc = SimpleDiscoveryXML('Discover_Repository_Db_Result')
        doc.push('RepositoryDbList')
        for repo_uuid, repo in dump_db(LOCAL_REPO_DB).items():
            doc.push('Repository', 'Uuid', repo_uuid)
            doc.add('Fs_location', repo['fs_location'])
            doc.add('Mount_point', repo['mount_point'])
            doc.add('Filesystem_type', repo['filesystem'])
            doc.add('Version', repo['version'])
            doc.add('Alias', repo['alias'])
            doc.add('Manager_uuid', repo['manager_uuid'])
            doc.add('Status', refresh_repository(repo_uuid, repo))
            doc.pop()
        doc.pop()
        return doc.finish()
    finally:
        lock.release()


def edit_repository_db(repo_uuid, db_changes):
    if repo_uuid not in dump_db(LOCAL_REPO_DB):
        raise Exception('Repository is not found in db.')
    repo = read_item(LOCAL_REPO_DB, repo_uuid)
    repo_mount_point = repo['mount_point']
    if not os.path.ismount(repo_mount_point):
        raise Exception('Repository is not mounted.')
    meta = get_repository_meta_data(repo_mount_point)
    if meta['OVS_REPO_MGR_UUID'] not in ['NONE', get_owning_manager_uuid()]:
        raise Exception('Repository is not owned by server manager.')
    if meta['OVS_REPO_VERSION'] != OVS_REPO_VERSION:
        raise Exception('Repository version is not correct.')
    if 'Version' in db_changes:
        meta['OVS_REPO_VERSION'] = db_changes['Version']
        repo['version'] = db_changes['Version']
    if 'Alias' in db_changes:
        meta['OVS_REPO_ALIAS'] = db_changes['Alias']
        repo['alias'] = db_changes['Alias']
    if 'Manager_uuid' in db_changes:
        meta['OVS_REPO_MGR_UUID'] = db_changes['Manager_uuid']
        repo['manager_uuid'] = db_changes['Manager_uuid']
    set_repository_meta_data(repo_mount_point, meta)


def import_file(repo_uuid, file_type, file_id, url, option, obj):
    if repo_uuid not in dump_db(LOCAL_REPO_DB):
        raise Exception('Repository is not found in db.')
    repo_mount_point = read_item(LOCAL_REPO_DB, repo_uuid)['mount_point']
    if not os.path.ismount(repo_mount_point):
        raise Exception('Repository is not mounted.')
    meta = get_repository_meta_data(repo_mount_point)
    if meta['OVS_REPO_MGR_UUID'] != get_owning_manager_uuid():
        raise Exception('Repository is not owned by server manager.')
    if meta['OVS_REPO_VERSION'] != OVS_REPO_VERSION:
        raise Exception('Repository version is not correct.')
    filename = join(repo_mount_point, file_type, file_id)
    ovs_download_file(url, filename, option, obj)


def import_virtual_disk(url, virtual_disk_id, repo_uuid, option=None):
    import_file(repo_uuid, 'VirtualDisks', virtual_disk_id, url, option, 'virtual_disk')


def import_ISO(url, iso_id, repo_uuid, option=None):
    import_file(repo_uuid, 'ISOs', iso_id, url, option, 'iso')


def mount_repositories():
    for repo in dump_db(LOCAL_REPO_DB).values():
        try:
            mount_repository_fs(repo['fs_location'], repo['mount_point'])
        except Exception, err:
            logging.warn('Error mount repository %s to %s: %s', repo['fs_location'], repo['mount_point'], err)
