# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import re
import os
import glob
import logging
import socket
import time
import agent.constants
from agent.constants import *
from agent.base import Agent
from agent.common import run_cmd, get_agt_url
from agent.db import read_item, write_item, update_db, truncate_db, get_cluster_db_home, dump_db
from agent.stats import write_config, reload_monitor
from agent.daemon import reload_notification_server, reload_ovm_consoled
from agent.discover import get_server_unique_id, SimpleDiscoveryXML
from agent.utils import filelock


VALID_ROLES = frozenset(["master", "utility", "xen"])
MASTER_ROLE = frozenset(["master"])


class ServerPool(Agent):
    """
    Class for OVS server pool management.
    """

    def get_services(self, version=None):
        return {
            'discover_server_pool': self.discover_server_pool,
            'take_ownership': self.take_ownership,
            'release_ownership': self.release_ownership,
            'create_server_pool': self.create_server_pool,
            'join_server_pool': self.join_server_pool,
            'destroy_server_pool': self.destroy_server_pool,
            'leave_server_pool': self.leave_server_pool,
            'set_server_pool_alias': self.set_server_pool_alias,
            'update_pool_virtual_ip': self.update_pool_virtual_ip,
            'update_server_roles': self.update_server_roles,
            'set_pool_member_ip_list': self.set_pool_member_ip_list,
        }

    def discover_server_pool(self):
        """
        Discover information about the server pool that contains this server.
        Returns an empty document if the server is unassociated.

        @return     If successful, returns the discovery data as an XML document
                    Raises an exception on failure.
        """
        lock = filelock.FileLock('/var/run/ovs-agent/discover_server_pool.lock')
        lock.acquire(wait=120)
        try:
            doc = SimpleDiscoveryXML("Discover_Server_Pool_Result")
            uuid = get_server_pool_uuid()
            if uuid:
                doc.push("Server_Pool")
                doc.add("Unique_Id", uuid)
                add_poolfs_info(doc)
                if is_clustered():
                    add_clustered_info(doc)
                else:
                    add_pooled_info(doc)
                doc.pop() # Server_Pool
            return doc.finish()
        finally:
            lock.release()

    def take_ownership(self, manager_uuid, manager_core_api_url):
        """
        Mark this server as being managed by a OVM Manager.

        @param      manager_uuid - UUID of the manager taking ownership
        @param      manager_core_api_url -
                    URL for making Core API calls to the manager
        @return     None.  Raises an exception on failure.
        """
        return take_ownership(manager_uuid, manager_core_api_url)

    def release_ownership(self, manager_uuid):
        """
        Release this server from being managed by an OVM Manager.

        @param      manager_uuid - UUID of the manager releasing ownership

        @return     None.  Raises an exception on failure.
        """
        return release_ownership(manager_uuid)

    def create_server_pool(self, pool_alias, pool_uuid, pool_virtual_ip,
                           node_number, server_hostname, server_ip, roles):
        """
        Create a server pool and add this server as its initial pool master.

        @param      pool_alias - human readable name for the pool
        @param      pool_uuid - globally unique id for the pool
        @param      pool_virtual_ip - server pool virtual IP
        @param      node_number -
                    node number in server pool and ocfs2 cluster
        @param      server_hostname -
                    hostname manager uses to talk to this server
        @param      server_ip -
                    ip that the manager uses to talk to this server
        @param      roles - text string with list of roles
        @return     None. Raises exception on failure
        """
        pool_add_server(pool_alias, pool_uuid, pool_virtual_ip,
            node_number, server_hostname, server_ip, roles, initial=True)

    def join_server_pool(self, pool_alias, pool_uuid, pool_virtual_ip,
                         node_number, server_hostname, server_ip, roles):
        """
        Add a server to an existing server pool.

        @param      pool_alias - human readable name for the pool
        @param      pool_uuid - globally unique id for the pool
        @param      pool_virtual_ip - server pool virtual IP
        @param      node_number -
                    node number in server pool and ocfs2 cluster
        @param      server_hostname -
                    hostname manager uses to talk to this server
        @param      server_ip -
                    ip that the manager uses to talk to this server
        @param      roles - text string with list of roles

        @return     None. Raises an exception on failure.
        """
        pool_add_server(pool_alias, pool_uuid, pool_virtual_ip,
            node_number, server_hostname, server_ip, roles, initial=False)

    def destroy_server_pool(self, pool_uuid):
        """
        Unjoin the last server (the master) from the server pool leaving
        no record of the pool.

        @param      pool_uuid -
                    globally unique ID of pool this server is in
        @return     None.  Raises an exception on failure.
        """
        pool_remove_server(pool_uuid, final=True)

    def leave_server_pool(self, pool_uuid):
        """
        Unjoin the server from the server pool.

        @param      pool_uuid -
                    globally unique ID of pool this server is in
        @return     None.  Raises an exception on failure
        """
        pool_remove_server(pool_uuid, final=False)

    def set_server_pool_alias(self, pool_alias):
        """
        Set a new human readable name for the server pool.

        @param      pool_alias - human readable name for the pool

        @return     None.  Raises an exception on failure.
        """
        if not pool_alias:
            raise ValueError("Null or missing parameter")
        state = get_membership_state()
        if state in (MEMBERSHIP_STATE_UNOWNED, MEMBERSHIP_STATE_OWNED):
            raise Exception("Server is not yet in a server pool")
        write_item(LOCAL_SERVER_DB, "pool_alias", pool_alias)
        if get_master_flag() and is_clustered():
            write_item("server_pool", "pool_alias", pool_alias, get_cluster_db_home())

    def update_pool_virtual_ip(self, new_pool_vip):
        """
        Update server pool virtual ip in database.

        @param      new_pool_vip - new server pool virtual IP

        @return     None.  Raises an exception on failure.
        """
        if not new_pool_vip:
            raise ValueError("Null or missing parameter")

        state = get_membership_state()
        if state in (MEMBERSHIP_STATE_UNOWNED, MEMBERSHIP_STATE_OWNED):
            raise Exception("Server is not yet in a server pool")

        write_item(LOCAL_SERVER_DB, "pool_virtual_ip", new_pool_vip)

        if get_master_flag() and is_clustered():
            write_item("server_pool", "pool_virtual_ip", new_pool_vip, get_cluster_db_home())

    def update_server_roles(self, new_roles):
        """
        Update the allowed roles for the server.

        @param      new_roles - role names, separated by commas
        @return     None.  Raises an exception on failure.
        """
        new_role_set = set([ e for e in new_roles.split(",") if e.strip() ])
        if new_role_set & VALID_ROLES != new_role_set:
            raise Exception("Invalid role(s): %s" % (new_role_set - VALID_ROLES))
        if get_master_flag() and MASTER_ROLE not in new_role_set:
            raise Exception("Cannot remove master role from the current master")
        write_item(LOCAL_SERVER_DB, "roles", new_role_set)

    def set_pool_member_ip_list(self, ip_list):
        """
        Set server pool members.

        @param      ip_list - ip list of pool members
        @return     None. Raises an exception on failure.
        """
        state = get_membership_state()
        if state in (MEMBERSHIP_STATE_UNOWNED, MEMBERSHIP_STATE_OWNED):
            raise Exception("Cannot set pool member ip list "
                            "in membership state: %s" % state)
        write_item(LOCAL_SERVER_DB, "pool_member_ip_list", ip_list)
        if is_clustered():
            write_item("server_pool", "pool_member_ip_list", ip_list, get_cluster_db_home())


def take_ownership(manager_uuid, manager_core_api_url):
    current_uuid = get_owning_manager_uuid()
    if current_uuid and current_uuid != manager_uuid:
        raise Exception("Server already owned by manager: %s" % current_uuid)
    logging.debug('manager_core_api_url: %s', manager_core_api_url)
    new_url = re.sub("0\.0\.0\.0", agent.constants.MANAGER_SRC_IP, manager_core_api_url)
    logging.debug('new manager_core_api_url: %s', new_url)
    manager_core_api_url = new_url
    manager_info = {
        "manager_uuid": manager_uuid,
        "manager_core_api_url": manager_core_api_url}
    update_db(LOCAL_SERVER_DB, manager_info)
    reload_notification_server()
    reload_ovm_consoled()


def release_ownership(manager_uuid):
    state = get_membership_state()
    if state != MEMBERSHIP_STATE_OWNED:
        raise Exception("Cannot release server in membership state: %s" % state)
    current_uuid = get_owning_manager_uuid()
    if not current_uuid:
        raise Exception("Server not currently owned")
    if current_uuid != manager_uuid:
        raise Exception("Server is owned by manager uuid: %s" % current_uuid)
    truncate_db(LOCAL_SERVER_DB)
    reload_notification_server()
    reload_ovm_consoled()


def pool_add_server(pool_alias, pool_uuid, pool_virtual_ip,
        node_number, server_hostname, server_ip, roles, initial=False):
    """
    Add this server as a pool member, possibly as the initial master

    @param      pool_alias - human readable name for the pool
    @param      pool_uuid - globally unique id for the pool
    @param      pool_virtual_ip - server pool virtual IP
    @param      node_number -
                node number in server pool and ocfs2 cluster
    @param      server_hostname -
                hostname manager uses to talk to this server
    @param      server_ip -
                ip that the manager uses to talk to this server
    @param      roles - text string with list of roles
    @param      initial -
                True if this is the initial pool member (and master)

    @return     None.  Raises an exception on failure.
    """
    role_set = set([ e for e in roles.split(",") if e.strip() ])
    if role_set & VALID_ROLES != role_set:
        raise Exception("Invalid role(s): %s" % (role_set - VALID_ROLES))

    if initial and len(MASTER_ROLE & role_set) == 0:
        raise Exception("Initial server must include master role")

    if not get_owning_manager_uuid():
        raise Exception("Server cannot be pooled: not owned by a manager")

    # Enforce the manager is using the server's actual (kernel) hostname
    # This will matter for cluster setup (and maybe in other places??)
    try:
        actual_hostname = socket.gethostname()
    except Exception, e:
        raise Exception("Unable to obtain hostname: %s" % e)

    if not actual_hostname:
        raise Exception("Server cannot be pooled: null hostname")

    if server_hostname != actual_hostname:
        raise Exception("Server's actual hostname is: %s" % actual_hostname)

    ip = socket.gethostbyname(server_hostname)
    if server_ip != ip:
        raise Exception("Invalid hostname resolution: %s -> %s"
                                        % (server_hostname, ip))
    uuid = get_server_pool_uuid()
    if uuid:
        raise Exception("Server already a member of pool: %s" % uuid)

    server_id = get_server_unique_id()
    if not server_id:
        raise Exception("Cannot get Server unique id.")

    # config qlayer-monitor
    write_config(pool_uuid, server_id)
    # start qlayer-monitor
    reload_monitor()

    d = {
        "pool_alias": pool_alias,
        "pool_uuid": pool_uuid,
        "roles": role_set,
        "registered_hostname": server_hostname,
        "registered_ip": server_ip,
        "pool_virtual_ip": pool_virtual_ip,
        "node_number": node_number,
        "is_master": initial}

    try:
        update_db(LOCAL_SERVER_DB, d)
    except Exception, e:
        raise Exception("Unable to update database: %s" % e)


def pool_remove_server(pool_uuid, final=False):
    """
    Unjoin the server from the server pool leaving no record that the
    server was in a pool

    @param      pool_uuid -
                globally unique ID of the pool this server is in
    @param      final -
                True if this is the final server (i.e. final master)

    @return     None.  Raises an exception on failure.
    """
    # It's up to the manager to know there are no other server in the pool
    if get_membership_state() != MEMBERSHIP_STATE_POOLED:
        raise Exception("Server must be in Pooled state")

    if not final and get_master_flag():
        raise Exception("Pool master cannot be removed from the server pool")

    if get_server_pool_uuid() != pool_uuid:
        raise Exception("Server not in specified pool")

    d = {
        "pool_alias": None,
        "pool_uuid": None,
        "roles": None,
        "registered_hostname": None,
        "registered_ip": None,
        "pool_virtual_ip": None,
        "node_number": None,
        "is_master": None,
        "pool_member_ip_list": None}

    try:
        update_db(LOCAL_SERVER_DB, d)
    except Exception, e:
        raise Exception("Unable to update database: %s" % e)


def get_owning_manager_uuid():
    return read_item(LOCAL_SERVER_DB, "manager_uuid")


def get_server_pool_uuid():
    return read_item(LOCAL_SERVER_DB, "pool_uuid")


def get_poolfs_uuid():
    return read_item(LOCAL_SERVER_DB, "poolfs_uuid")


def get_cluster_flag():
    return read_item(LOCAL_SERVER_DB, "clustered")


def get_master_flag():
    return read_item(LOCAL_SERVER_DB, "is_master")


def get_membership_state():
    """
    Return the current state of the server with respect to
    server pool and cluster membership.

    @return     One of: "Unowned", "Owned", "Pooled",
                "Heartbeat_Configured" and "Clustered".
    """
    if get_cluster_flag():
        return MEMBERSHIP_STATE_CLUSTERED

    if get_poolfs_uuid():
        return MEMBERSHIP_STATE_HB_CONFIGURED

    if get_server_pool_uuid():
        return MEMBERSHIP_STATE_POOLED

    if get_owning_manager_uuid():
        return MEMBERSHIP_STATE_OWNED

    return MEMBERSHIP_STATE_UNOWNED

def get_current_master_url():
    """
    Returns the URL used for making XMLRPC calls to the current pool
    master. Before the server is clustered, this comes from the local
    server database. Once we are clustered, return the cluster wide
    value that tracks pool master failover.

    @return     Current master's URL
    """
    state = get_membership_state()
    if state in (MEMBERSHIP_STATE_UNOWNED, MEMBERSHIP_STATE_OWNED):
        return ""

    if is_clustered():
        pool_virtual_ip = read_item("server_pool", "pool_virtual_ip", get_cluster_db_home())
    else:
        pool_virtual_ip = read_item(LOCAL_SERVER_DB, "pool_virtual_ip")

    return get_agt_url(pool_virtual_ip)


def get_pool_member_ip_list():
    """
    Get server pool member IP list.

    @return     IP list of server pool members.
    """
    ip_list = []
    state = get_membership_state()

    if state in (MEMBERSHIP_STATE_UNOWNED, MEMBERSHIP_STATE_OWNED):
        ip_list = []
    elif is_clustered():
        ip_list = read_item("server_pool", "pool_member_ip_list", get_cluster_db_home())
    else:
        ip_list = read_item(LOCAL_SERVER_DB, "pool_member_ip_list")

    # None -> []
    return ip_list or []

def get_ocfs2_cluster_name():
    """
    Get the ocfs2 cluster name in configfs.

    @return     Cluster name. Null string if not found.
    """
    ptn = os.path.join(CLUSTER_CONFIGFS, "*")
    lst = [ os.path.basename(e) for e in glob.glob(ptn) ]
    num = len(lst)

    if num == 0:
        return ""
    elif num == 1:
        return lst[0]
    else:
        raise Exception("Multiple clusters found: %s" % lst)


def get_cluster_state():
    """
    Get the current state of the cluster.

    @param      None

    @return     One of "Offline", "Online", "Heartbeating",
                "PoolFS_Mounted" and "DLM_Ready".
    """
    state = CLUSTER_STATE_OFFLINE
    cluster_name = get_ocfs2_cluster_name()

    if cluster_name:
        state = CLUSTER_STATE_ONLINE
        poolfs_uuid = get_poolfs_uuid()

        if poolfs_uuid:
            mount_point = os.path.join(POOLFS_MOUNT, poolfs_uuid)
            if os.path.ismount(mount_point):
                if os.path.ismount(DLM):
                    return CLUSTER_STATE_DLM_READY
                else:
                    return CLUSTER_STATE_POOLFS_MOUNTED

        output = run_cmd(["service", "o2cb", "status"])
        for line in output.splitlines():
            if line.strip() == "Checking O2CB heartbeat: Active":
                return CLUSTER_STATE_HEARTBEATING

    return state

def is_clustered():
    return (get_membership_state() == MEMBERSHIP_STATE_CLUSTERED and
            get_cluster_state() == CLUSTER_STATE_DLM_READY)

def add_clustered_info(doc):
    """
    Add XML for server pool members and virtual IP if clustered
    """
    d = dump_db("server_pool", get_cluster_db_home())
    pool_alias = d.get("pool_alias")
    master_vip = d.get("pool_virtual_ip")
    member_list = d.get("pool_member_ip_list")

    if pool_alias:
        doc.add("Pool_Alias", pool_alias)

    if master_vip:
        doc.add("Master_Virtual_Ip", master_vip)

    doc.push("Member_List")

    for ip in member_list:
        doc.push("Member")
        doc.add("Registered_IP", ip)
        doc.pop() # Member

    doc.pop() # Member_List

def add_pooled_info(doc):
    """
    Add XML for server pool members that we know about when not yet clustered
    """
    d = dump_db(LOCAL_SERVER_DB)
    pool_alias = d.get("pool_alias")
    virtual_ip = d.get("pool_virtual_ip")
    member_list = d.get("pool_member_ip_list")

    if pool_alias:
        doc.add("Pool_Alias", pool_alias)

    if virtual_ip:
        doc.add("Master_Virtual_Ip", virtual_ip)

    doc.push("Member_List")

    if member_list:
        for srv in member_list:
            doc.push("Member")
            doc.add("Registered_IP", srv)
            doc.pop() # Member

    doc.pop() # Member_List

def add_poolfs_info(doc):
    """Add XML for the pool filesystem if available."""
    d = dump_db(LOCAL_SERVER_DB)
    poolfs_type = d.get("poolfs_type")
    poolfs_target = d.get("poolfs_target")
    poolfs_uuid = d.get("poolfs_uuid")
    poolfs_nfsbase_uuid = d.get("poolfs_nfsbase_uuid")

    if poolfs_type:
        doc.push("Pool_Filesystem")
        doc.add("Pool_Filesystem_Type", poolfs_type)
        doc.add("Pool_Filesystem_Target", poolfs_target)
        doc.add("Pool_Filesystem_Uuid", poolfs_uuid)
        doc.add("Pool_Filesystem_Nfsbase_Uuid", poolfs_nfsbase_uuid)
        doc.pop() # Pool_Filesystem
