# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import copy
import logging
from agent.common import Common
from agent.storageplugin import StoragePlugin
from agent.network.linux_network import LinuxNetwork
from agent.remote.linux_remote import LinuxRemoteManagement
from agent.cluster.o2cb import ClusterO2CB
from agent.hypervisor.xenxm import Xen
from agent.poolfs.ocfs2 import PoolOCFS2
from agent.host.linux import Linux
from agent.serverpool import ServerPool
from agent.repository import Repository


TRACE_EXCEPTION = [
    'dispatch',
    'ovs_async_proc',
    'get_last_boot_time',
    'update_agent_password',
    'send_to_guest',
    'list_vms',
    'remote_power_on',
    'remote_power_off',
]

SCRUB_LIST = ['password', 'passwd']


def well_behavior(decorator):
    """This decorator can be used to turn simple functions
    into well-behaved decorators, so long as the decorators
    are fairly simple. If a decorator expects a function and
    returns a function (no descriptors), and if it doesn't
    modify function attributes or docstring, then it is
    eligible to use this. Simply apply @well_behavior to
    your decorator and it will automatically preserve the
    docstring and function attributes of functions to which
    it is applied."""
    def new_decorator(f):
        g = decorator(f)
        g.__name__ = f.__name__
        g.__doc__ = f.__doc__
        g.__dict__.update(f.__dict__)
        return g
    # Now a few lines needed to make well_behavior itself
    # be a well-behaved decorator.
    new_decorator.__name__ = decorator.__name__
    new_decorator.__doc__ = decorator.__doc__
    new_decorator.__dict__.update(decorator.__dict__)
    return new_decorator


def scrub_dict(d):
    for key in d:
        for word in SCRUB_LIST:
            if word in key:
                d[key] = '******'


def scrub_and_log(name, *args):
    temp_args = copy.deepcopy(args)
    for arg in temp_args:
        if type(arg) == dict:
            scrub_dict(arg)
    logging.debug("%s: %s" % (name, temp_args))


@well_behavior
def trace(func):
    def wrapper(*args):
        scrub_and_log(func.__name__, *args)
        result = func(*args)
        logging.debug("%s: call completed.", func.__name__)
        return result
    return wrapper


@well_behavior
def catch_error(func):
    def wrapper(*args):
        try:
            return func(*args)
        except Exception, e:
            logging.exception("catch_error: %s" % e)
            raise
    return wrapper


def expose(name):
    @well_behavior
    def wrapper(func):
        def f(*args):
            new_func = catch_error(func)
            if name not in TRACE_EXCEPTION:
                new_func = trace(new_func)
            return new_func(*args)
        f.__name__ = name
        return f
    return wrapper


def get_api(version=None):
    api = {}
    modules = [Common, Xen, Linux, LinuxNetwork, LinuxRemoteManagement,
               ClusterO2CB, PoolOCFS2, Repository, ServerPool,
               StoragePlugin]
    for mod in modules:
        services = mod().get_services(version)
        for (name, func) in services.items():
            api.update({name: expose(name)(func)})
    return api
