# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import os
import os.path
import shutil
import errno
import logging
import itertools
import simplejson
import yaml
import tempfile
import time
from agent.constants import *
from agent.utils import filelock
from agent.db import write_item, read_item
from agent.common import run_cmd
from agent.notification import notify_manager


def set_statistic_interval(interval):
    write_item(LOCAL_SERVER_DB, "stat_interval", interval)


def reload_monitor(base=MONITOR_BASE_DIR):
    lock = filelock.FileLock(MONITOR_LOCK)
    lock.acquire(wait=120)
    try:
        run_cmd([MONITOR, 'stop'])
        logging.debug("qlayer-monitor stopped")
        load_config(base)
        run_cmd([MONITOR, 'start'])
        logging.debug("qlayer-monitor started")
    finally:
        lock.release()


def load_config(base):
    # parse config
    fd = open(MONITOR_CFG, 'r')
    try:
        tmpdata = yaml.load(fd)
    finally:
        fd.close()
    stats_loc = tmpdata['consumers']['filesystem']['args']['stats_location']
    events_loc = tmpdata['consumers']['filesystem']['args']['events_location']

    # remove obsolete data
    if os.path.exists(stats_loc):
        shutil.rmtree(stats_loc, True)
    if os.path.exists(events_loc):
        shutil.rmtree(events_loc, True)

    # remove symlink
    if os.path.islink(MONITOR_BASE_DIR):
        os.remove(MONITOR_BASE_DIR)
    # link required
    if base != MONITOR_BASE_DIR:
        # remove obsolete data
        if os.path.exists(MONITOR_BASE_DIR):
            shutil.rmtree(MONITOR_BASE_DIR, True)
        # create new base
        if not os.path.exists(base):
            os.makedirs(base)
        # symlink
        os.symlink(base, MONITOR_BASE_DIR)

    # create stats and events locations
    if not os.path.exists(stats_loc):
        os.makedirs(stats_loc)
    if not os.path.exists(events_loc):
        os.makedirs(events_loc)


def write_config(pool, machine):
    tmpdata = {}
    # set to defaults if not present
    data = {'config':
               {'default_interval': MON_CFG_DFLT_INTERVAL,
                'plugin_path': MON_CFG_DFLT_PLUGIN_PATH,
                'reload_interval': MON_CFG_DFLT_RELOAD_INTERVAL,
                'default_grace': MON_CFG_DFLT_GRACE},
            'plugins': None,
            'consumers': {'filesystem': {'handler': MON_CFG_FS_HANDLER, 'args':
                {'events_location': MON_CFG_EVENTS_LOC % (pool, machine),
                 'stats_location': MON_CFG_STATS_LOC % (pool, machine) } }}}

    fd = open(MONITOR_CFG, 'r')
    try:
        try:
            tmpdata = yaml.load(fd)
        except yaml.YAMLError, exc:
            logging.warning("Invalid data in monitor config file, overwriting")
            pass
        except IOError, ioe:
            logging.warning("Unable to open monitor config file")
            pass
    finally:
        fd.close()

    try:
        if tmpdata['consumers']['filesystem']['args']['stats_location'] is not None and \
           tmpdata['consumers']['filesystem']['args']['events_location'] is not None and \
           tmpdata['consumers']['filesystem']['handler'] is not None:
            # config file has all needed keys
            return
    except KeyError:
        # merge parsed config and default config
        tmpdata.update(data)

    tmpfd,tmpname = tempfile.mkstemp(dir='/tmp')
    myfile = None
    try:
        try:
            myfile = os.fdopen(tmpfd, 'w')
        except:
            os.close(tmpfd)
            logging.exception('Unable to open temp file')
            raise

        try:
            yaml.dump(data, myfile, default_flow_style=False)
        except yaml.YAMLError, exc:
            logging.exception('Bad monitor config file format')
            raise RuntimeError('Bad monitor config file format')
    finally:
        if myfile:
            myfile.close()

    try:
        shutil.move(MONITOR_CFG, MONITOR_CFG + '.save')
        shutil.move(tmpname, MONITOR_CFG)
    except:
        logging.exception('Unable to move config file: %s', MONITOR_CFG)


def _process_stats():
    # Assume stat files are in $BASE/*/monitor/*/stats/
    # All $BASE/*/monitor folders, as (name, folder) tuples
    dirs1 = itertools.ifilter(lambda (_, d): os.path.isdir(d),
        ((name, os.path.join(MONITOR_BASE_DIR, name, 'monitor'))
            for name in os.listdir(MONITOR_BASE_DIR)))
    # All $BASE/*/monitor/*/stats folders, as (name1, name2, folder) tuples
    dirs2 = itertools.ifilter(lambda (_1, _2, d): os.path.isdir(d),
        ((name1, name2, os.path.join(dir1, name2, 'stats'))
            for (name1, dir1) in dirs1
            for name2 in os.listdir(dir1)))

    # All $BASE/*/monitor/*/stats/* files, as (name1, name2, file) tuples
    files = itertools.ifilter(lambda (_1, _2, f): os.path.isfile(f),
        ((name1, name2, os.path.join(dir2, name3))
            for (name1, name2, dir2) in dirs2
            for name3 in os.listdir(dir2)))

    # sort files based on mtime
    files = sorted(files, key=lambda x: os.path.getmtime(x[2]))

    objlist = []
    count = 0
    max = 256
    to_delete = []
    for name1, name2, file_ in files:
        fd = None
        try:
            fd = open(file_, 'rb')
            data = fd.read()
            fd.close()
        except Exception, e:
            if fd:
                fd.close()
            logging.debug('Error reading file %s: %', file_, e)
            continue

        try:
            # FIXME: always convert integer to long for core api binding limitation
            obj = simplejson.loads(data, parse_int=long)
            obj['pool'] = name1
            obj['machine'] = name2
            objlist.append(obj)
        except:
            logging.exception('Error during stat processing')

        to_delete.append(file_)
        count += 1
        if count >= max:
            break

    try:
        notify_manager('{STATS}', objlist)
    except Exception, e:
        # FIXME: should we log for stats (for every 20s)? No for now.
        # logging.debug("Error sending notification: %s", e)
        return

    # stats have been sent, delete stats files
    for file_ in to_delete:
        try:
            os.unlink(file_)
        except OSError, e:
            if e.errno != errno.ENOENT:
                logging.debug('Unexpected error in unlink %s: %s', file_, e)
        except Exception, e:
            logging.debug('Unexpected error in unlink %s: %s', file_, e)


def monitor_base_broken(clustered):
    if not os.path.isdir(MONITOR_BASE_DIR):
        return True
    if clustered:
        return not os.path.islink(MONITOR_BASE_DIR)
    else:
        return os.path.islink(MONITOR_BASE_DIR)


def fix_monitor_base(clustered):
    if clustered:
        poolfs_uuid = read_item(LOCAL_SERVER_DB, 'poolfs_uuid')
        mount_point = os.path.join(POOLFS_MOUNT, poolfs_uuid)
        monitor_base_dir_clustered = os.path.join(mount_point, 'qlayer')
        reload_monitor(monitor_base_dir_clustered)
    else:
        reload_monitor(MONITOR_BASE_DIR)
    logging.debug("Broken monitor base fixed")


def process_stats():
    lock = filelock.FileLock(MONITOR_LOCK)
    lock.acquire(wait=60)
    try:
        _process_stats()
    finally:
        lock.release()


def serve_forever():
    logging.info("STAT SERVER STARTED")

    from agent.serverpool import is_clustered, get_membership_state
    from agent.remaster import is_master

    while True:
        try:
            stat_interval = read_item(LOCAL_SERVER_DB, "stat_interval") or MONITOR_SERVER_INTERVAL
            time.sleep(stat_interval)
            if is_clustered():
                if monitor_base_broken(True):
                    fix_monitor_base(True)
                if is_master():
                    process_stats()
            else:
                if get_membership_state() not in (MEMBERSHIP_STATE_UNOWNED, MEMBERSHIP_STATE_OWNED):
                    if monitor_base_broken(False):
                        fix_monitor_base(False)
                    process_stats()
        except Exception, e:
            logging.exception("Error in stat process: %s", e)
            time.sleep(SLEEP_ON_ERR)


if __name__ == '__main__':
    serve_forever()
