# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import logging
import subprocess
import sys
import time
try:
    import cPickle as pickle
except:
    import pickle
from agent.base import Agent
from agent.notification import async_proc_prog
from agent.discover import SimpleDiscoveryXML
from agent.utils import filelock


class StoragePlugin(Agent):
    """
    Class for Oracle Storage Connect plugin.
    """

    def get_services(self, version=None):
        return {
            'storage_plugin_validate': storage_plugin_validate,
            'storage_plugin_getCapabilities': storage_plugin_getCapabilities,
            'storage_plugin_getStorageNames': storage_plugin_getStorageNames,
            'storage_plugin_getInfo': storage_plugin_getInfo,
            'storage_plugin_getAccessGroups': storage_plugin_getAccessGroups,
            'storage_plugin_createAccessGroups': storage_plugin_createAccessGroups,
            'storage_plugin_removeAccessGroups': storage_plugin_removeAccessGroups,
            'storage_plugin_renameAccessGroup': storage_plugin_renameAccessGroup,
            'storage_plugin_addToAccessGroup': storage_plugin_addToAccessGroup,
            'storage_plugin_removeFromAccessGroup': storage_plugin_removeFromAccessGroup,
            'storage_plugin_discover': storage_plugin_discover,
            'storage_plugin_start': storage_plugin_start,
            'storage_plugin_stop': storage_plugin_stop,
            'storage_plugin_refresh': storage_plugin_refresh,
            'storage_plugin_list': storage_plugin_list,
            'storage_plugin_updateSERecords': storage_plugin_updateSERecords,
            'storage_plugin_getStatus': storage_plugin_getStatus,
            'storage_plugin_online': storage_plugin_online,
            'storage_plugin_offline': storage_plugin_offline,
            'storage_plugin_create': storage_plugin_create,
            'storage_plugin_startPresent': storage_plugin_startPresent,
            'storage_plugin_stopPresent': storage_plugin_stopPresent,
            'storage_plugin_resize': storage_plugin_resize,
            'storage_plugin_remove': storage_plugin_remove,
            'storage_plugin_getCloneLimits': storage_plugin_getCloneLimits,
            'storage_plugin_isCloneable': storage_plugin_isCloneable,
            'storage_plugin_clone': storage_plugin_clone,
            'storage_plugin_splitClone': storage_plugin_splitClone,
            'storage_plugin_cloneFromSnap': storage_plugin_cloneFromSnap,
            'storage_plugin_getCurrentClones': storage_plugin_getCurrentClones,
            'storage_plugin_getSnapLimits': storage_plugin_getSnapLimits,
            'storage_plugin_isSnapable': storage_plugin_isSnapable,
            'storage_plugin_createSnap': storage_plugin_createSnap,
            'storage_plugin_createMultiSnap': storage_plugin_createMultiSnap,
            'storage_plugin_isRestorable': storage_plugin_isRestorable,
            'storage_plugin_snapRestore': storage_plugin_snapRestore,
            'storage_plugin_snapRemove': storage_plugin_snapRemove,
            'storage_plugin_getCurrentSnaps': storage_plugin_getCurrentSnaps,
            'storage_plugin_getQoSList': storage_plugin_getQoSList,
            'storage_plugin_setQoS': storage_plugin_setQoS,
            'storage_plugin_getStorageServerInfo': storage_plugin_getStorageServerInfo,
            'storage_plugin_getFileSystemInfo': storage_plugin_getFileSystemInfo,
            'storage_plugin_getFileInfo': storage_plugin_getFileInfo,
            'storage_plugin_listFileSystems': storage_plugin_listFileSystems,
            'storage_plugin_listMountPoints': storage_plugin_listMountPoints,
            'storage_plugin_mount': storage_plugin_mount,
            'storage_plugin_unmount': storage_plugin_unmount,
            'storage_plugin_createFileSystem': storage_plugin_createFileSystem,
            'storage_plugin_resizeFileSystem': storage_plugin_resizeFileSystem,
            'storage_plugin_destroyFileSystem': storage_plugin_destroyFileSystem,
            'storage_plugin_destroy': storage_plugin_destroy,
            'storage_plugin_getFileSystemCloneLimits': storage_plugin_getFileSystemCloneLimits,
            'storage_plugin_getFileSystemSnapLimits': storage_plugin_getFileSystemSnapLimits,
            'storage_plugin_deviceTeardown': storage_plugin_deviceTeardown,
            'storage_plugin_deviceResize': storage_plugin_deviceResize,
            'storage_plugin_deviceSizeRefresh': storage_plugin_deviceSizeRefresh,
            'storage_iSCSI_logoutTarget': storage_iSCSI_logoutTarget,
            'storage_iSCSI_deletePortal': storage_iSCSI_deletePortal,
            'discover_storage_plugins': discover_storage_plugins,
        }


def storage_plugin(impl_name, *params):
    if "/opt/storage-connect" not in sys.path: sys.path.append("/opt/storage-connect")
    from PluginManager import PluginManager

    # Get my callers function name from the stack to be able to figure out
    # which plugin function is called via XMLRPC
    caller_name = sys._getframe(1).f_code.co_name

    # Get my function name which is also the prefix used in the exposed
    # XMLRPC functions
    my_prefix = '%s_' % sys._getframe().f_code.co_name

    logging.info("%s(%s)" % (caller_name, impl_name))

    plugin_mgr = PluginManager(logging.getLogger())
    plugin_class = plugin_mgr.getPlugin(impl_name)
    index = caller_name.find(my_prefix)
    caller_name = caller_name[index :]
    method_name = caller_name.replace(my_prefix, '', 1)

    plugin_method = getattr(plugin_class, method_name)
    try:
        update_cache(plugin_class, *params)
        result = plugin_method(*params)
        if method_name in ['create', 'clone', 'resize', 'createFiles', 'cloneFiles', 'resizeFiles']:
            async_method = getattr(plugin_class, 'getAsyncProgress')
            result = watch_async_progress(async_method, method_name, result, *params)
    except Exception, exc:
        logging.info("Exception: " % exc)
        raise
    return_cache(plugin_class, *params)
    # XMLRPC is stupid and cannot send 64bit longs over the wire so we
    # make them all Strings...
    result = process_results(result)
    return result


def update_cache(plugin_class, ss_record, *params):

    if ss_record and ss_record.has_key('uuid') and ss_record.has_key('cache'):
        plugin_cache = getattr(plugin_class, 'cache')
        if ss_record['cache']:
            cache = pickle.loads(ss_record['cache'])
        else:
            cache = ''
        plugin_cache.write(ss_record['uuid'], cache)


def return_cache(plugin_class, ss_record, *params):
    plugin_cache = getattr(plugin_class, 'cache')
    if ss_record and ss_record.has_key('uuid'):
        cache = plugin_cache.read(ss_record['uuid'])
        if cache:
            if cache == '':
                send_cache = {ss_record['uuid']: ''}
            else:
                send_cache = {ss_record['uuid']:pickle.dumps(cache, 0)}


def watch_async_progress(async_method, method_name, result_record, ss_record, *params):
    if type(result_record) == dict and result_record.has_key('async_progress'):
        status = {'OPERATION': 'storage_' + method_name,
                  'PROGRESS': 0}
        while result_record['async_progress'] != None:
            status['PROGRESS'] = result_record['async_progress']
            async_proc_prog(status)
            time.sleep(2)
            # It is assumed that there will be a valid async_method
            # if there is an async_progress key
            result_record = async_method(ss_record, result_record)
    if result_record.has_key('async_progress'):
        del result_record['async_progress']
    if result_record.has_key('async_handle'):
        del result_record['async_handle']
    return result_record


def process_results(param):
    if type(param) == list:
        for p in param:
            p = process_results(p)
    elif type(param) == dict:
        for p in param.keys():
            param[p] = process_results(param[p])
    elif type(param) == int or type(param) == long:
        param = str(param)
    return param


def storage_plugin_validate(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_getCapabilities(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_getStorageNames(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_getInfo(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_getAccessGroups(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_createAccessGroups(impl_name,*params):
    return storage_plugin(impl_name, *params)

def storage_plugin_removeAccessGroups(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_renameAccessGroup(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_addToAccessGroup(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_removeFromAccessGroup(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_discover(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_start(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_stop(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_refresh(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_list(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_updateSERecords(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_getStatus(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_online(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_offline(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_create(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_startPresent(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_stopPresent(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_resize(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_remove(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_getCloneLimits(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_isCloneable(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_clone(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_splitClone(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_cloneFromSnap(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_getCurrentClones(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_getSnapLimits(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_isSnapable(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_createSnap(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_createMultiSnap(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_isRestorable(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_snapRestore(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_snapRemove(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_getCurrentSnaps(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_getQoSList(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_setQoS(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_getStorageServerInfo(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_getFileSystemInfo(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_getFileInfo(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_listFileSystems(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_listMountPoints(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_mount(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_unmount(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_createFileSystem(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_resizeFileSystem(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_destroyFileSystem(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_destroy(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_getFileSystemCloneLimits(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_getFileSystemSnapLimits(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_deviceTeardown(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_deviceResize(impl_name, *params):
    return storage_plugin(impl_name, *params)

def storage_plugin_deviceSizeRefresh(impl_name, *params):
    return storage_plugin(impl_name, *params)


def storage_iSCSI_logoutTarget(target, portal):
    logging.debug("ENTER: storage_iSCSI_logoutTarget()")
    iscsi_p = subprocess.Popen(["/sbin/iscsiadm",
                                "--mode", "node",
                                "--targetname=%s" % target,
                                "--portal=%s" % portal,
                                "--logout"],
                               stdout = subprocess.PIPE,
                               stderr = subprocess.PIPE)
    (p_out, p_err) = iscsi_p.communicate()
    if iscsi_p.returncode != 0:
        logging.error("Target logout failed: %s" % p_err)
        raise Exception("Target logout failed: %s" % p_err)
    logging.debug("EXIT: storage_iSCSI_logoutTarget() - Success")


def storage_iSCSI_deletePortal(portal):
    logging.debug("ENTER: storage_iSCSI_deletePortal()")
    iscsi_p = subprocess.Popen(["/sbin/iscsiadm",
                                "--mode", "node",
                                "--portal=%s" % portal,
                                "--op=delete"],
                               stdout = subprocess.PIPE,
                               stderr = subprocess.PIPE)
    (p_out, p_err) = iscsi_p.communicate()
    if iscsi_p.returncode != 0:
        logging.error("Portal delete failed: %s" % p_err)
        raise Exception("Portal delete failed: %s" % p_err)
    logging.debug("EXIT: storage_iSCSI_deletePortal() - Success")


def discover_storage_plugins():
    """
    Discover the installed Storage Connect plugins on this OVM Server.

    @return: None. Raises an exception on failure.
    """

    if "/opt/storage-connect" not in sys.path: sys.path.append("/opt/storage-connect")
    import OSCPlugin
    from PluginManager import PluginManager

    class DiscoverPlugins(object):
        """
        Discover the installed plugins.
        """
        def __init__(self):
            """
            Constructor for the Storage Connect Plugin Discovery.
            """
            self.__plugin_mgr__ = PluginManager(logging.getLogger())
            self.__plugin_names__ = None

        def getInfoOnPlugins(self):
            """
            Obtain information on all the plugins installed in the system.

            @return: A dict with the information on all the discovered plugins.
            """
            self.__plugin_names__ = self.__plugin_mgr__.discoverPlugins(ignore_failed = True)
            info = {}
            for plugin_name in self.__plugin_names__:
                p_info = self.__plugin_mgr__.getPluginInfo(plugin_name)
                abilities = p_info['plugin_ability']
                for ability in abilities.keys():
                    abilities[ability] = OSCPlugin.ABILITY_TYPES.getID(abilities[ability])
                info[plugin_name] = p_info
            return info

        def genDiscoverResult(self, doc):
            """
            Generate the discover request result XML for the plugins.

            @param:  XML Document to insert information about the plugins into.

            @return: None. Raises an exception on error.
            """
            if not self.__plugin_names__ or len(self.__plugin_names__) <= 0:
                self.__plugin_names__ = self.__plugin_mgr__.discoverPlugins(ignore_failed = True)
                if not self.__plugin_names__ or len(self.__plugin_names__) <= 0:
                    # We expect at least one plugin (the generic ones) to be discovered
                    raise Exception("Unable to obtain Storage Connect plugin list")
            doc.push("storage_plugin_info_list")
            for impl_name in self.__plugin_names__:
                try:
                    plugin_info = self.__plugin_mgr__.getPluginInfo(impl_name)
                except:
                    # We just ignore any plugin that gives any kind of error
                    continue
                doc.push("storage_plugin_info", "plugin_impl_name", impl_name)
                for info_key in plugin_info.keys():
                    if info_key != "plugin_ability":
                        if type(plugin_info[info_key]) == str:
                            doc.add(info_key, plugin_info[info_key])
                        elif type(plugin_info[info_key]) == list:
                            doc.add(info_key, ",".join(plugin_info[info_key]))
                        else:
                            doc.add(info_key, str(plugin_info[info_key]))
                abilities = plugin_info.get("plugin_ability", None)
                if abilities:
                    doc.push("abilities")
                    for ability in abilities.keys():
                        if ability == "max_access_entries":
                            doc.add(ability, str(abilities[ability]))
                        elif ability == "backing_device_type":
                            abilities[ability] = OSCPlugin.BACKING_DEVICE_TYPES.getID(abilities[ability])
                            doc.add(ability, abilities[ability])
                        else:
                            abilities[ability] = OSCPlugin.ABILITY_TYPES.getID(abilities[ability])
                            doc.add(ability, abilities[ability])
                    doc.pop() # abilities
                doc.pop() # storage_plugin_info
            doc.pop() # storage_plugin_info_list

    lock = filelock.FileLock('/var/run/ovs-agent/discover_storage_plugins.lock')
    lock.acquire(wait=120)
    try:
        doc = SimpleDiscoveryXML("Discover_Storage_Plugins_Result")
        plugin_discovery = DiscoverPlugins()
        plugin_discovery.genDiscoverResult(doc)
        return doc.finish()
    finally:
        lock.release()
