# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import os
from os.path import exists, isfile
from conf import ConfShellVar


SYSCONFDEVICEDIR = "/etc/sysconfig/network-scripts/"


class NetworkError(Exception):
    pass


class ConfNetwork(ConfShellVar):
    def __init__(self, filename):
        if not os.access(self.filename, os.R_OK):
            self.oldmode = 0644
        else:
            status = os.stat(self.filename)
            self.oldmode = status[0]
        ConfShellVar.__init__(self, filename)

    def write(self):
        self.chmod(self.oldmode)
        ConfShellVar.write(self)

    def get(self, key, default=None):
        if key in self.vars:
            return self[key]
        return default

    def update(self, dict=None, **kwargs):
        if dict is None:
            pass
        else:
            for k, v in dict.items():
                self[k] = v
        if len(kwargs):
            self.vars.update(kwargs)

    def remove(self):
        if exists(self.filename):
            os.unlink(self.filename)

    def clear(self):
        for key in self.keys():
            del self[key]
        self.lines = []


class ConfDevice(ConfNetwork):
    def __init__(self, name):
        self.filename = SYSCONFDEVICEDIR + "ifcfg-" + name
        ConfNetwork.__init__(self, self.filename)


class ConfRoute(ConfNetwork):
    def __init__(self, name):
        self.filename = SYSCONFDEVICEDIR + "route-" + name
        ConfNetwork.__init__(self, self.filename)


def getConfiguredDevices():
    # ignored files: ifcfg- | *~ | *.bak | *.orig | *.rpmnew | *.rpmorig | *.rpmsave | *-range
    devlist = []
    for entry in os.listdir(SYSCONFDEVICEDIR):
        if (len(entry) > 6) and \
           entry.startswith("ifcfg-") and \
           not entry.endswith("~") and \
           not entry.endswith(".bak") and \
           not entry.endswith(".orig") and \
           not entry.endswith(".rpmnew") and \
           not entry.endswith(".rpmorig") and \
           not entry.endswith(".rpmsave") and \
           not entry.endswith("-range") and \
           isfile(SYSCONFDEVICEDIR + entry) and \
           os.access(SYSCONFDEVICEDIR + entry, os.R_OK):
            devlist.append(entry[6:])
    return devlist


def getConfig(device):
    conf = ConfDevice(device)
    return conf.vars


def getBondSlavesFromConfig(device):
    devlist = []
    for dev in getConfiguredDevices():
        conf = ConfDevice(dev)
        if conf.get("MASTER") == device:
            devlist.append(dev)
    return devlist


def getBondMastersFromConfig():
    devlist = []
    for dev in getConfiguredDevices():
        conf = ConfDevice(dev)
        if "MASTER" in conf.keys():
            devlist.append(conf['MASTER'])
    return devlist


def getBridgePortsFromConfig(device):
    devlist = []
    for dev in getConfiguredDevices():
        conf = ConfDevice(dev)
        if conf.get("BRIDGE") == device:
            devlist.append(dev)
    return devlist
