# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import os
import sys
import signal
import socket
import logging
import agent.constants
from SocketServer import ForkingMixIn
from agent.utils import xmlrpclib
from agent.utils.SimpleXMLRPCServer import SimpleXMLRPCServer, SimpleXMLRPCRequestHandler, SimpleXMLRPCDispatcher
from agent.constants import *
from agent.config import get_config
from agent.common import run_cmd, get_api_version
from agent.password import encrypt_password, read_password, update_password
from agent.daemon import reload_xmlrpc_server
from agent.host.linux import get_ovs_info
from agent.serverpool import get_pool_member_ip_list
from agent.service import get_api


PASSWORD = None


def get_passwd():
    global PASSWORD
    if PASSWORD is None:
        PASSWORD = read_password()
    return PASSWORD


def sig_handler(sig, frame):
    global PASSWORD
    if sig == signal.SIGUSR1:
        PASSWORD = None


def auth(username, passwd):
    password_dict = get_passwd()
    method, stored_passwd = password_dict.get(username, ('CRYPT', '!'))
    if stored_passwd != encrypt_password(passwd, method, stored_passwd):
        raise Exception('Authorization Failed')
    if method == 'SHA':
        logging.warn('Replacing SHA password for user: %s', username)
        update_password(username, passwd)
        reload_xmlrpc_server()


def auth_ip(ip):
    if ip != "127.0.0.1" and ip not in get_pool_member_ip_list():
        raise Exception('Authorization failed: server not in the server pool: %s' % ip)


class AgentXMLRPCDispatcher(SimpleXMLRPCDispatcher):
    def _marshaled_dispatch(self, data, dispatch_method=None, path=None):
        params, method = xmlrpclib.loads(data)
        try:
            if dispatch_method is not None:
                response = dispatch_method(method, params)
            else:
                response = self._dispatch(method, params)
            # wrap response in a singleton tuple
            response = (response,)
            response = xmlrpclib.dumps(response, methodresponse=1, allow_none=1)
        except xmlrpclib.Fault, fault:
            response = xmlrpclib.dumps(fault, allow_none=1)
        except:
            # report exception back to server
            response = xmlrpclib.dumps(
                xmlrpclib.Fault(1, "%s:%s" % (sys.exc_type, sys.exc_value)),
                allow_none=1)
        return response


class AgentXMLRPCRequestHandler(SimpleXMLRPCRequestHandler):

    rpc_paths = ['/', '/RPC2']

    def is_rpc_path_valid(self):
        if self.rpc_paths:
            return self.path in self.rpc_paths
        else:
            # If .rpc_paths is empty, just assume all paths are legal
            return True

    def setup(self):
        config = get_config()
        if config.get("security", "ssl").lower().startswith("enable"):
            self.connection = self.request
            self.rfile = socket._fileobject(self.request, "rb", self.rbufsize)
            self.wfile = socket._fileobject(self.request, "wb", self.wbufsize)
        else:
            SimpleXMLRPCRequestHandler.setup(self)

    def do_POST(self):
        # Check that the path is legal
        if not self.is_rpc_path_valid():
            self.send_error(404, "No such page: %s" % self.path)
            return

        try:
            src_ip = self.client_address[0]
            agent.constants.MANAGER_SRC_IP = src_ip
            if "authorization" in self.headers:
                username, password = self.headers["authorization"].split(" ")[1].decode("base64").split(":")
                auth(username, password)
            else:
                auth_ip(src_ip)
        except:
            errmsg = "Unauthorized access attempt from %s!" % (self.client_address,)
            logging.exception(errmsg)
            self.send_error(403, errmsg)
            return

        config = get_config()
        if config.get("security", "ssl").lower().startswith("enable"):
            try:
                # get arguments
                data = self.rfile.read(int(self.headers["content-length"]))
                # In previous versions of SimpleXMLRPCServer, _dispatch
                # could be overridden in this class, instead of in
                # SimpleXMLRPCDispatcher. To maintain backwards compatibility,
                # check to see if a subclass implements _dispatch and dispatch
                # using that method if present.
                response = self.server._marshaled_dispatch(
                        data, getattr(self, '_dispatch', None), self.path
                    )
            except: # This should only happen if the module is buggy
                # internal error, report as HTTP server error
                self.send_response(500)
                self.end_headers()
            else:
                # got a valid XML RPC response
                self.send_response(200)
                self.send_header("Content-type", "text/xml")
                self.send_header("Content-length", str(len(response)))
                self.end_headers()
                self.wfile.write(response)

                # shut down the connection
                self.wfile.flush()
                self.connection.shutdown()
        else:
            try:
                # get arguments
                data = self.rfile.read(int(self.headers["content-length"]))
                # In previous versions of SimpleXMLRPCServer, _dispatch
                # could be overridden in this class, instead of in
                # SimpleXMLRPCDispatcher. To maintain backwards compatibility,
                # check to see if a subclass implements _dispatch and dispatch
                # using that method if present.
                response = self.server._marshaled_dispatch(
                        data, getattr(self, '_dispatch', None), self.path
                    )
            except: # This should only happen if the module is buggy
                # internal error, report as HTTP server error
                self.send_response(500)
                self.end_headers()
            else:
                # got a valid XML RPC response
                self.send_response(200)
                self.send_header("Content-type", "text/xml")
                self.send_header("Content-length", str(len(response)))
                self.end_headers()
                self.wfile.write(response)

                # shut down the connection
                self.wfile.flush()
                self.connection.shutdown(1)

    def log_message(self, format, *args):
        logging.info(format % args)


class OVSAgentServer(ForkingMixIn, SimpleXMLRPCServer):

    allow_reuse_address = True
    max_children = 500 # default 40 in ForkingMixin

    def __init__(self, addr, requestHandler=AgentXMLRPCRequestHandler, logRequests=1):
        SimpleXMLRPCServer.__init__(self, addr, requestHandler, logRequests)
        config = get_config()
        if config.get("security", "ssl").lower().startswith("enable"):
            from OpenSSL import SSL
            KEYFILE = os.path.join(CONFIG_DIR, "cert/key.pem")
            CERTFILE = os.path.join(CONFIG_DIR, "cert/certificate.pem")
            ctx = SSL.Context(SSL.SSLv23_METHOD)
            ctx.set_options(SSL.OP_NO_SSLv2)
            ctx.use_privatekey_file(KEYFILE)
            ctx.use_certificate_file(CERTFILE)
            self.socket = SSL.Connection(ctx, socket.socket(self.address_family, self.socket_type))
            self.server_bind()
            self.server_activate()
        self.dispatchers = {}
        requestHandler.rpc_paths = ['/', '/RPC2']
        api_version = get_api_version()
        for version in api_version:
            api = get_api(version)
            dispatcher = AgentXMLRPCDispatcher()
            for (api, func) in api.items():
                dispatcher.register_function(func, api)
            rpc_path = '/api/%s' % version
            self.dispatchers[rpc_path] = dispatcher
            requestHandler.rpc_paths.append(rpc_path)

        latest_api_rpc_path = '/api/%s' % api_version[-1]
        self.dispatchers['/'] = self.dispatchers[latest_api_rpc_path]
        self.dispatchers['/RPC2'] = self.dispatchers[latest_api_rpc_path]

    def _marshaled_dispatch(self, data, dispatch_method=None, path=None):
        try:
            response = self.dispatchers[path]._marshaled_dispatch(
               data, dispatch_method, path)
        except:
            # report low level exception back to server
            # (each dispatcher should have handled their own
            # exceptions)
            response = xmlrpclib.dumps(
                xmlrpclib.Fault(1, "%s:%s" % (sys.exc_type, sys.exc_value)),
                allow_none=1)
        return response

    def verify_request(self, request, client_address):
        """
        overwrite original verify_request to implement IP based access control
        """
        #--------------------<> internal function <>----
        def match(rule, ip):
            #check if ip is in the range
            try:
                #e.g.: 219.142.73.80
                if ip == rule: return True
                #e.g.: 219.142.73.*
                if rule.endswith("*") and ip.startswith(rule[:-1]): return True
                #e.g.: 219.142.73.0/20
                vip, bits = rule.split("/")
                vip_str = "".join(["%02x" % int(x) for x in vip.split(".")])
                ip_str = "".join(["%02x" % int(x) for x in ip.split(".")])
                q, r = divmod(int(bits), 4)
                if vip_str[:q] == ip_str[:q]:
                    if int(vip_str[q], 16) >> (4 - r) == int(ip_str[q], 16) >> (4 - r):
                        return True
            except: #in case there are errors in rules
                pass
        #--------------------<> main processing <>----
        #check ip access control rules, sequence: allow=>deny
        (ip, port) = client_address

        # always allow connect from localhost
        if ip == '127.0.0.1': return True

        config = get_config()
        #allow rules
        allow = False
        allow_rules = config.get("security", "allow").split(",")
        for rule in allow_rules:
            if match(rule.strip(), ip):
                allow = True
                break
        if not allow: return False

        #deny rules
        deny = False
        deny_rules = config.get("security", "deny").split(",")
        for rule in deny_rules:
            if match(rule.strip(), ip):
                deny = True
                break
        if deny: return False

        #nothing wrong, finally
        return True


def serve_forever():
    signal.signal(signal.SIGUSR1, sig_handler)
    config = get_config()
    server = OVSAgentServer(('', int(config.get("agent", "port"))),
                            AgentXMLRPCRequestHandler, logRequests=False)
    logging.info("Oracle VM Agent XMLRPC Server started.")
    hostname = ""
    hostip = ""
    try:
        hostname = socket.gethostname()
        hostip = socket.gethostbyname(hostname)
    except Exception, err:
        logging.warn("Error getting hostname and IP: %s", err)
    logging.info("Oracle VM Server version: %s, hostname: %s, ip: %s",
                 get_ovs_info(), hostname, hostip)
    server.serve_forever()


if __name__ == '__main__':
    serve_forever()
